/* $Id: crires_extract.c,v 1.67 2012-09-20 12:06:34 yjung Exp $
 *
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2012-09-20 12:06:34 $
 * $Revision: 1.67 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
   								Includes
 -----------------------------------------------------------------------------*/

#include <math.h>
#include <float.h>
#include <cpl.h>

#include "irplib_spectrum.h"

#include "crires_extract.h"
#include "crires_dfs.h"
#include "crires_utils.h"

/*-----------------------------------------------------------------------------
   								Includes
 -----------------------------------------------------------------------------*/

#define CRIRES_PIXSCALE     0.086

/*-----------------------------------------------------------------------------
   								Static
 -----------------------------------------------------------------------------*/

static cpl_image * crires_extract_create_profile(const cpl_image *,
        crires_illum_period, int, int, int, int, int, int *, int *) ;
static int crires_extract_detect(const cpl_image *, crires_illum_period, 
        int, int, int) ;
static cpl_vector * crires_extract_bgd_rms(const cpl_image *, const cpl_image *,
        double, int, double, cpl_image **) ;
static cpl_image * crires_extract_profile_rect(int, int, int, int, int *, 
        cpl_vector **);
static cpl_image * crires_extract_profile_opt(cpl_image *, cpl_vector *, int, 
        int, double, int, cpl_vector **) ;
static cpl_vector * crires_extract_opt_correction(cpl_image *, cpl_image *) ;
static cpl_vector * crires_extract_compute_error(const cpl_image *, 
        const cpl_vector *, const cpl_image *, double, int) ;
static cpl_vector * crires_extract_compute_spectrum(const cpl_image *,
        const cpl_image *) ;
static cpl_vector * crires_extract_normalise_profile(cpl_image *) ;
static int crires_extract_reject_sides(cpl_vector *, int, int) ;
static double crires_extract_fwhm_med(const cpl_image *) ;

/*----------------------------------------------------------------------------*/
/**
 * @defgroup crires_extract     Spectrum extraction routines
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Extract the spectrum
  @param    spec_image      the image with the horizontal spectrum
  @param    contrib         the contribution map or NULL
  @param    optimal         flag to use the optimal extraction
  @param    box_size        the horizontal size of the box for the profile
  @param    spec_hsize      the spectrum half size (used in non-optimal)
  @param    kappa           the kappa value for the background identification
  @param    closing_hs      Half size of the kernel for the closing on the bgmap
  @param    clean_rate      Min rate of bgd pixels per line allowed
  @param    rej_left        the nb of values set to 0 on the left
  @param    rej_right       the nb of values set to 0 on the right
  @param    spec_starty     Start of the zone where the spectrum is
  @param    spec_stopy      Stop of the zone where the spectrum is
  @param    chip            the chip number
  @param    tot_ndit        ndit*nexp*dit
  @param    period          The illumination period
  @param    spec_pos        the computed brightest spectrum position
  @param    spec_wrec       the computed spectrum width RECT
  @param    spec_wopt       the computed spectrum width OPT
  @param    spec_optmed     the median of the optimal spectrum
  @param    spec_s2n        the median of the signal to noise
  @param    profile         the computed profile image
  @param    bg_map          the computed background map or NULL
  @return   the extracted table or NULL in error case

  box_size must be in [3, CRIRES_SPEC_BOX_SIZE_LIMIT]
  kappa must be higher than 0

  The return table must be deallocated with cpl_table_delete()
 */
/*----------------------------------------------------------------------------*/
cpl_table * crires_extract_spectrum(
        const cpl_image     *   spec_image,
        const cpl_image     *   contrib,
        int                     box_size,
        int                     spec_hsize,
        double                  kappa,
        int                     closing_hs,
        double                  clean_rate,
        int                     rej_left,
        int                     rej_right,
        int                     spec_starty,
        int                     spec_stopy,
        int                     chip,
        double                  tot_ndit,
        crires_illum_period     period,
        int                 *   spec_pos,
        int                 *   spec_wrec,
        int                 *   spec_wopt,
        double              *   spec_optmed,
        double              *   spec_s2n,
        cpl_image           **  profile,
        cpl_image           **  bg_map)
{
    cpl_table           *   out ;
    cpl_image           *   prof_init ;
    cpl_image           *   bgmap ;
    int                     pos_loc, ly ;
    int                     spec_wopt_loc ;
    int                     spec_wrec_loc ;
    double                  spec_optmed_loc ;
    double                  spec_s2n_loc ;
    cpl_vector          *   bgd_rms ;
    cpl_image           *   prof_rect ;
    cpl_image           *   prof_opt ;
    cpl_vector          *   correction_opt ;
    cpl_vector          *   weight_rect ;
    cpl_vector          *   weight_opt ;
    cpl_vector          *   error_rect ;
    cpl_vector          *   error_opt ;
    cpl_vector          *   spec_rect ;
    cpl_vector          *   spec_opt ;
    double              *   perror_opt ;
    int                     i ;
    
    /* Test entries */
    if (spec_image == NULL) return NULL ;
    if (period == CRIRES_ILLUM_UNKNOWN) return NULL ;
    if (period == CRIRES_ILLUM_LOWER_HALF) return NULL ;
    if (period == CRIRES_ILLUM_UPPER_HALF) return NULL ;
    if (box_size < 3 || box_size > CRIRES_SPEC_BOX_SIZE_LIMIT) return NULL ;
    if (kappa < 0) return NULL ;
    if (chip < 1 || chip > CRIRES_NB_DETECTORS) return NULL ;

    /* Initialise */
    if (profile != NULL) *profile = NULL ;
    if (bg_map != NULL) *bg_map = NULL ;
    ly = -1 ;
 
    /* Compute the RMS of the background */
    cpl_msg_info(__func__, "Background RMS computation") ;
    cpl_msg_indent_more() ;
    if ((bgd_rms = crires_extract_bgd_rms(spec_image, contrib, kappa,
                    closing_hs, clean_rate, &bgmap))==NULL) {
        cpl_msg_error(__func__, "Cannot compute the background RMS") ;
        cpl_msg_indent_less() ;
        return NULL ;
    }
    cpl_msg_indent_less() ;

    /* Create the spectrum profile */
    cpl_msg_info(__func__, "Initial profile computation") ;
    if ((prof_init = crires_extract_create_profile(spec_image, period,
                    box_size, spec_starty, spec_stopy, chip, spec_hsize,
                    &pos_loc, &spec_wopt_loc)) == NULL) {
        cpl_msg_error(__func__, "Cannot compute the initial profile") ;
        cpl_vector_delete(bgd_rms) ;
        cpl_image_delete(bgmap) ;
        return NULL ;
    }
    cpl_msg_info(__func__, "Spectrum at y=%d", pos_loc) ;

    /* Get the correction for the position */
    if (chip == 1) ly = crires_get_detector_ly1(period) ;
    if (chip == 2) ly = crires_get_detector_ly2(period) ;
    if (chip == 3) ly = crires_get_detector_ly3(period) ;
    if (chip == 4) ly = crires_get_detector_ly4(period) ;
    /* If ly not found, use the bottom of the image */
    if (ly < 0) ly = 1 ;

    /* Compute the rectangular profile */
    cpl_msg_info(__func__, "Rectangular profile computation") ;
    if ((prof_rect = crires_extract_profile_rect(pos_loc-ly+1, spec_hsize, 
                    cpl_image_get_size_x(spec_image),
                    cpl_image_get_size_y(spec_image),
                    &spec_wrec_loc,
                    &weight_rect)) == NULL) {
        cpl_msg_error(__func__, "Cannot create the rectangular profile") ;
        cpl_image_delete(prof_init) ;
        cpl_vector_delete(bgd_rms) ;
        cpl_image_delete(bgmap) ;
        return NULL ;
    }
    
    /* Compute the optimal profile */
    cpl_msg_info(__func__, "Optimal profile computation") ;
    if ((prof_opt = crires_extract_profile_opt(prof_init, bgd_rms, chip,
                    pos_loc-ly+1, tot_ndit, box_size, &weight_opt)) == NULL) {
        cpl_msg_error(__func__, "Cannot create the optimal profile") ;
        cpl_image_delete(prof_init) ;
        cpl_vector_delete(bgd_rms) ;
        cpl_image_delete(prof_rect) ;
        cpl_vector_delete(weight_rect) ;
        cpl_image_delete(bgmap) ;
        return NULL ;
    }
    
    /* Compute the Correction factor to apply for the otimal spectrum */
     cpl_msg_info(__func__, "Optimal correction factor computation") ;
    if ((correction_opt = crires_extract_opt_correction(prof_init, 
                    prof_opt)) == NULL) {
        cpl_msg_error(__func__, "Cannot compute the correction factor") ;
        cpl_image_delete(prof_init) ;
        cpl_vector_delete(bgd_rms) ;
        cpl_image_delete(prof_rect) ;
        cpl_image_delete(prof_opt) ;
        cpl_vector_delete(weight_rect) ;
        cpl_vector_delete(weight_opt) ;
        cpl_image_delete(bgmap) ;
        return NULL ;
    }
    cpl_image_delete(prof_init) ;

    /* Compute the error RECT */
    cpl_msg_info(__func__, "Error RECT computation") ;
    if ((error_rect = crires_extract_compute_error(spec_image, bgd_rms,
                    prof_rect, tot_ndit, chip))==NULL){
        cpl_msg_error(__func__, "Cannot compute the error RECT") ;
        cpl_vector_delete(bgd_rms) ;
        cpl_image_delete(prof_rect) ;
        cpl_vector_delete(weight_rect) ;
        cpl_image_delete(prof_opt) ;
        cpl_vector_delete(weight_opt) ;
        cpl_vector_delete(correction_opt) ;
        cpl_image_delete(bgmap) ;
        return NULL ;
    }
    cpl_vector_multiply_scalar(error_rect, spec_wrec_loc) ;

    /* Compute the error OPT */
    cpl_msg_info(__func__, "Error OPT computation") ;
    if ((error_opt = crires_extract_compute_error(spec_image, bgd_rms,
                    prof_opt, tot_ndit, chip))==NULL){
        cpl_msg_error(__func__, "Cannot compute the error OPT") ;
        cpl_vector_delete(bgd_rms) ;
        cpl_image_delete(prof_rect) ;
        cpl_vector_delete(weight_rect) ;
        cpl_image_delete(prof_opt) ;
        cpl_vector_delete(weight_opt) ;
        cpl_vector_delete(error_rect) ;
        cpl_vector_delete(correction_opt) ;
        cpl_image_delete(bgmap) ;
        return NULL ;
    }
    /* Apply the correction factor */
    cpl_vector_divide(error_opt, correction_opt) ;

    /* Compute the spectrum RECT */
    cpl_msg_info(__func__, "Spectrum RECT computation") ;
    if ((spec_rect = crires_extract_compute_spectrum(spec_image, 
                    prof_rect))==NULL){
        cpl_msg_error(__func__, "Cannot compute the spectrum RECT") ;
        cpl_image_delete(prof_rect) ;
        cpl_vector_delete(weight_rect) ;
        cpl_image_delete(prof_opt) ;
        cpl_vector_delete(weight_opt) ;
        cpl_vector_delete(error_rect) ;
        cpl_vector_delete(error_opt) ;
        cpl_vector_delete(bgd_rms) ;
        cpl_vector_delete(correction_opt) ;
        cpl_image_delete(bgmap) ;
        return NULL ;
    }
    cpl_vector_multiply_scalar(spec_rect, spec_wrec_loc) ;

    /* Compute the spectrum OPT */
    cpl_msg_info(__func__, "Spectrum OPT computation") ;
    if ((spec_opt = crires_extract_compute_spectrum(spec_image, 
                    prof_opt))==NULL){
        cpl_msg_error(__func__, "Cannot compute the spectrum OPT") ;
        cpl_image_delete(prof_rect) ;
        cpl_vector_delete(weight_rect) ;
        cpl_image_delete(prof_opt) ;
        cpl_vector_delete(weight_opt) ;
        cpl_vector_delete(error_rect) ;
        cpl_vector_delete(error_opt) ;
        cpl_vector_delete(spec_rect) ;
        cpl_vector_delete(bgd_rms) ;
        cpl_vector_delete(correction_opt) ;
        cpl_image_delete(bgmap) ;
        return NULL ;
    }
    cpl_image_delete(prof_rect) ;

    /* Apply the correction factor */
    cpl_vector_divide(spec_opt, correction_opt) ;
    cpl_vector_delete(correction_opt) ;

    /* Reject the pixels on the edges */
    crires_extract_reject_sides(spec_rect, rej_left, rej_right) ;
    crires_extract_reject_sides(spec_opt, rej_left, rej_right) ;

    /* Create the output table */
    if ((out = crires_extract_gen_tab(spec_rect, spec_opt, error_rect,
                    error_opt, weight_rect, weight_opt, bgd_rms)) == NULL) {
        cpl_msg_error(__func__, "Cannot create the extracted table") ;
        cpl_image_delete(prof_opt) ;
        cpl_vector_delete(weight_rect) ;
        cpl_vector_delete(weight_opt) ;
        cpl_vector_delete(error_rect) ;
        cpl_vector_delete(error_opt) ;
        cpl_vector_delete(spec_rect) ;
        cpl_vector_delete(spec_opt) ;
        cpl_vector_delete(bgd_rms) ;
        cpl_image_delete(bgmap) ;
        return NULL ;
    }
    cpl_vector_delete(weight_rect) ;
    cpl_vector_delete(weight_opt) ;
    cpl_vector_delete(error_rect) ;
    cpl_vector_delete(spec_rect) ;
    cpl_vector_delete(bgd_rms) ;
 
    /* Statistics on error_opt and spec_opt */
    spec_optmed_loc = cpl_vector_get_median_const(spec_opt) ;
    perror_opt = cpl_vector_get_data(error_opt) ;
    for (i=0 ; i<cpl_vector_get_size(error_opt) ; i++) {
        if (fabs(perror_opt[i]) < 1e-5) perror_opt[i] = 1.0 ;
    }
    cpl_vector_divide(spec_opt, error_opt) ; 
    cpl_vector_delete(error_opt) ;
    spec_s2n_loc = cpl_vector_get_median_const(spec_opt) ;
    cpl_vector_delete(spec_opt) ;

    /* Return  */
    if (profile != NULL)    *profile = prof_opt ;
    else                    cpl_image_delete(prof_opt) ;
    if (bg_map != NULL)     *bg_map = bgmap ;
    else                    cpl_image_delete(bgmap) ;
    if (spec_pos != NULL)   *spec_pos = pos_loc ;
    if (spec_wrec != NULL)  *spec_wrec = spec_wrec_loc ;
    if (spec_wopt != NULL)  *spec_wopt = spec_wopt_loc ;
    if (spec_optmed!=NULL)  *spec_optmed = spec_optmed_loc ; 
    if (spec_s2n!=NULL)     *spec_s2n = spec_s2n_loc ; 
    return  out ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Extract the lines
  @param    spec_image      the image with the horizontal spectrum
  @param    start_y         the lower part of the zone to collapse
  @param    stop_y          the upper part of the zone to collapse
  @param    rej_left        the nb of values set to 0 on the left
  @param    rej_right       the nb of values set to 0 on the right
  @return   the extracted spectrum or NULL in error case

  The return vector has to be deallocated with cpl_vector_delete().
 */
/*----------------------------------------------------------------------------*/
cpl_vector * crires_extract_lines(
        const cpl_image     *   spec_image,
        int                     start_y,
        int                     stop_y,
        int                     rej_left,
        int                     rej_right)
{
    cpl_image           *   collapsed ;
    cpl_vector          *   spec ;
    
    /* Test entries */
    if (spec_image == NULL) return NULL ;

    /* Initialise */
    if (start_y < 1 || start_y > cpl_image_get_size_y(spec_image))   
        return NULL ;
    if (stop_y > cpl_image_get_size_y(spec_image) || stop_y < 1)  
        return NULL ;
    if (start_y > stop_y) 
        return NULL ;

    /* Collapse the image */
    if ((collapsed = cpl_image_collapse_window_create(
                    spec_image, 
                    1, 
                    start_y, 
                    cpl_image_get_size_x(spec_image),
                    stop_y, 
                    0)) == NULL) {
        cpl_msg_error(__func__, "Cannot collapse the image") ;
        return NULL ;
    }
    spec = cpl_vector_new_from_image_row(collapsed, 1) ;
    cpl_image_delete(collapsed) ;

    /* Reject the pixels on the edges */
    crires_extract_reject_sides(spec, rej_left, rej_right) ;
    
    return spec ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the median FWHM of the combined and profile as QC
  @param    comb            the combined image
  @param    prof            the profile
  @param    fwhm_comb_pix   the fwhm median from the combined image (pix)
  @param    fwhm_comb_as    the fwhm median from the combined image (arcsec)
  @param    fwhm_prof_pix   the fwhm median from the profile (pix)
  @param    fwhm_prof_as    the fwhm median from the profile (arcsec)
  @param    fwhm_diff       fwhm_prof_pix - fwhm_comb_pix

  @return   int 0 if ok, or -1 in error case
 */
/*----------------------------------------------------------------------------*/
int crires_extract_qc_fwhm(
        const cpl_image     *   comb,
        const cpl_image     *   prof,
        double              *   fwhm_comb_pix,
        double              *   fwhm_comb_as,
        double              *   fwhm_prof_pix,
        double              *   fwhm_prof_as,
        double              *   fwhm_diff)
{

    cpl_image       *   mask_im ;
    float           *   pmask_im ;
    int                 nx, ny ;
    double              fwhm_comb, fwhm_prof, pixscale ; 
    int                 i, j ;

    /* Test entries */
    if (comb == NULL || prof == NULL) return -1 ;
    nx = cpl_image_get_size_x(comb) ;
    ny = cpl_image_get_size_y(comb) ;
    if (fwhm_comb_pix == NULL || 
            fwhm_comb_as == NULL || 
            fwhm_prof_pix == NULL || 
            fwhm_prof_as == NULL || 
            fwhm_diff== NULL) return -1 ;
    if (nx != cpl_image_get_size_x(prof)) return -1 ;
    if (ny != cpl_image_get_size_y(prof)) return -1 ;

    /* Create the mask from the profile */
    mask_im = cpl_image_duplicate(prof) ;
    pmask_im = cpl_image_get_data_float(mask_im) ;
    for (i=0 ; i<nx ; i++) {
        for (j=0 ; j<ny ; j++) {
            if (pmask_im[i+j*nx]>1e-3)  pmask_im[i+j*nx] = 1.0 ;
            else                        pmask_im[i+j*nx] = 0.0 ;
        }
    }

    /* Apply the mask */
    if (cpl_image_multiply(mask_im, comb) != CPL_ERROR_NONE) {
        cpl_image_delete(mask_im) ;
        return -1 ;
    }

    /* Compute the fwhm from the combined image */
    fwhm_comb = crires_extract_fwhm_med(mask_im) ;
    cpl_image_delete(mask_im) ;
    if (fwhm_comb < 0) {
        cpl_error_reset() ;
        return -1 ;   
    }

    /* Compute the fwhm from the profile */
    fwhm_prof = crires_extract_fwhm_med(prof) ;
    if (fwhm_prof < 0) {
        cpl_error_reset() ;
        return -1 ;   
    }

    /* Check if ok */
    if (fwhm_comb < 0.0 || fwhm_prof < 0.0) return -1 ;

    /* Get the pixscale  */
    pixscale = CRIRES_PIXSCALE ;

    /* Return */
    *fwhm_comb_pix = fwhm_comb ;
    *fwhm_comb_as = fwhm_comb * pixscale ;
    *fwhm_prof_pix = fwhm_prof ;
    *fwhm_prof_as = fwhm_prof * pixscale ;
    *fwhm_diff = fwhm_prof - fwhm_comb ;
    return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the spatial source fwhm
  @param    ima     the input image
  @return   the bivector containing the center and FWHM values 
            or NULL in error case

  For each column of the image, the positive beam is fitted with a
  gaussian, and the center and FWHM are computed.

  The returned bivector must be deallocated with cpl_bivector_delete()
 */
/*----------------------------------------------------------------------------*/
cpl_bivector * crires_extract_spatial_fwhm(const cpl_image * ima) 
{
    int                     nx, ny ;
    cpl_image           *   lab_im ;
    int                 *   plab_im ;
    const float         *   pima ;
    cpl_apertures       *   aperts ;
    int                     bot, top ;
    cpl_bivector        *   centers_fwhms ;
    double              *   pfwhms ;
    double              *   pcenters ;
    double                  center, fwhm ;
    cpl_image           *   extracted ;
    double                  stdev ;
    int                     ind, x ;
    cpl_vector          *   vec_x ;
    cpl_vector          *   vec_y ;
    double                  x0, sigma, norm, offset ;
    int                     found ;
    cpl_size                cenx, ceny ;
    int                     i, j ;
    
    /* Test entries */
    if (ima == NULL) return NULL ;

    /* Initialise */
    nx = cpl_image_get_size_x(ima) ;
    ny = cpl_image_get_size_y(ima) ;
    pima = cpl_image_get_data_float_const(ima) ;
    
    /* Create a label image */
    lab_im = cpl_image_new(nx, ny, CPL_TYPE_INT) ;
    plab_im = cpl_image_get_data_int(lab_im) ;

    /* Detect for each column what is above 2*sigmas */
    ind = 0 ;
    for (i=0 ; i<nx ; i++) {
        /* Compute stdev */
        stdev = cpl_image_get_stdev_window(ima, i+1, 1, i+1, ny) ;
        if (stdev > 0.0) { 
            ind++ ;
            found=0 ;
            /* Apply the threshold */
            for (j=0 ; j<ny ; j++) {
                if (pima[i+j*nx]>2*stdev) { 
                    plab_im[i+j*nx] = ind ;
                    found = 1 ;
                }
            }
            if (found==0) ind-- ;
        }
    }

   /* Compute the stats on the objects */
    if ((aperts = cpl_apertures_new_from_image(ima, lab_im)) == NULL) {
        cpl_msg_error(__func__, "Cannot compute the stats on the bgd zone");
        cpl_image_delete(lab_im) ;
        cpl_error_reset() ;
        return NULL ;
    }
    cpl_image_delete(lab_im) ;

    /* Create the output vector */
    centers_fwhms = cpl_bivector_new(nx) ;
    cpl_vector_fill(cpl_bivector_get_x(centers_fwhms), 0.0) ;
    cpl_vector_fill(cpl_bivector_get_y(centers_fwhms), 0.0) ;
    pcenters = cpl_bivector_get_x_data(centers_fwhms) ;
    pfwhms = cpl_bivector_get_y_data(centers_fwhms) ;
    
    /* Loop on the apertures */
    for (i=0 ; i<cpl_apertures_get_size(aperts) ; i++) {
        /* Get the borders of the objects */
        bot = cpl_apertures_get_bottom(aperts, i+1) ;
        top = cpl_apertures_get_top(aperts, i+1) ;
        x = (int)cpl_apertures_get_pos_x(aperts, i+1) ;

        /* Extract the image */
        if ((extracted = cpl_image_extract(ima, x, bot, x, top))==NULL) {
            cpl_msg_error(__func__, "Cannot extract the image");
            cpl_bivector_delete(centers_fwhms) ;
            cpl_apertures_delete(aperts) ;
            return NULL ;
        }

        /* Put the image in a vector */
        vec_y = cpl_vector_new_from_image_column(extracted, 1) ;
        cpl_image_delete(extracted) ;
        vec_x = cpl_vector_duplicate(vec_y) ;
        for (j=0 ; j<cpl_vector_get_size(vec_x) ; j++) {
            cpl_vector_set(vec_x, j, (double)j+1) ;
        }
        
        /* There should not be any 0 in the extracted */
        if (cpl_vector_get_min(vec_y) <= 0) {
            center = 0.0 ;
            fwhm = 0.0 ;
        } else {
            /* Compute the FWHMS */
            if (cpl_vector_fit_gaussian(vec_x, NULL, vec_y, NULL,
                        CPL_FIT_ALL, &x0, &sigma, &norm, &offset,
                        NULL, NULL, NULL) != CPL_ERROR_NONE) {
                cpl_image_get_maxpos_window(ima, x, 1, x, ny, &cenx, &ceny);
                center = (double)ceny ;
                fwhm = 0 ;
                cpl_error_reset() ;
            } else {
                center = x0 + bot - 1 ; ;
                fwhm = 2*sigma*sqrt(2*log(2)) ;
            }
        }
        cpl_vector_delete(vec_x) ;
        cpl_vector_delete(vec_y) ;
        
        /* Store the result */
        pcenters[x-1] = center ;
        pfwhms[x-1] = fwhm ;
    }
    cpl_apertures_delete(aperts) ;
    return centers_fwhms ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Generate the extracted table
  @param    ext_rect        the extracted spectrum RECT
  @param    ext_opt         the extracted spectrum OPT or NULL
  @param    err_rect        the error RECT or NULL
  @param    err_opt         the error OPT or NULL
  @param    weight_rect     the weights RECT or NULL
  @param    weight_opt      the weights OPT or NULL
  @param    bgd_rms         the background noide or NULL
  @return   the extracted table or NULL in error case

  The return table has to be deallocated with cpl_table_delete().
 */
/*----------------------------------------------------------------------------*/
cpl_table * crires_extract_gen_tab(
        const cpl_vector    *   ext_rect,
        const cpl_vector    *   ext_opt,
        const cpl_vector    *   err_rect,
        const cpl_vector    *   err_opt,
        const cpl_vector    *   weight_rect,
        const cpl_vector    *   weight_opt,
        const cpl_vector    *   bgd_rms)
{
    cpl_table       *   tab ;
    double              extr, exto, errr, erro, wnr, wno, bgn ;
    int                 tab_size ;
    int                 i ;

    /* Check entries */
    if (ext_rect == NULL) return NULL ;

    /* Create the table */
    tab_size = cpl_vector_get_size(ext_rect) ;
    tab = cpl_table_new(tab_size) ;
    cpl_table_new_column(tab, CRIRES_COL_EXTRACT_INT_RECT, CPL_TYPE_DOUBLE) ;
    cpl_table_new_column(tab, CRIRES_COL_EXTRACT_INT_OPT, CPL_TYPE_DOUBLE) ;
    cpl_table_new_column(tab, CRIRES_COL_EXTRACT_ERR_RECT, CPL_TYPE_DOUBLE) ;
    cpl_table_new_column(tab, CRIRES_COL_EXTRACT_ERR_OPT, CPL_TYPE_DOUBLE) ;
    cpl_table_new_column(tab, CRIRES_COL_EXTRACT_WN_RECT, CPL_TYPE_DOUBLE) ;
    cpl_table_new_column(tab, CRIRES_COL_EXTRACT_WN_OPT, CPL_TYPE_DOUBLE) ;
    cpl_table_new_column(tab, CRIRES_COL_EXTRACT_BG_NOISE, CPL_TYPE_DOUBLE) ;

    for (i=0 ; i<tab_size ; i++) {
        extr = cpl_vector_get(ext_rect, i) ;
        if (ext_opt != NULL)        exto = cpl_vector_get(ext_opt, i) ;
        else                        exto = 0.0 ;
        if (err_rect != NULL)       errr = cpl_vector_get(err_rect, i) ;
        else                        errr = 0.0 ;
        if (err_opt != NULL)        erro = cpl_vector_get(err_opt, i) ;
        else                        erro = 0.0 ;
        if (weight_rect != NULL)    wnr = cpl_vector_get(weight_rect, i) ;
        else                        wnr = 0.0 ;
        if (weight_opt != NULL)     wno = cpl_vector_get(weight_opt, i) ;
        else                        wno = 0.0 ;
        if (bgd_rms != NULL)        bgn = cpl_vector_get(bgd_rms, i) ;
        else                        bgn = 0.0 ;
        cpl_table_set_double(tab, CRIRES_COL_EXTRACT_INT_RECT, i, extr) ;
        cpl_table_set_double(tab, CRIRES_COL_EXTRACT_INT_OPT, i, exto) ;
        cpl_table_set_double(tab, CRIRES_COL_EXTRACT_ERR_RECT, i, errr) ;
        cpl_table_set_double(tab, CRIRES_COL_EXTRACT_ERR_OPT, i, erro) ;
        cpl_table_set_double(tab, CRIRES_COL_EXTRACT_WN_RECT, i, wnr) ;
        cpl_table_set_double(tab, CRIRES_COL_EXTRACT_WN_OPT, i, wno) ;
        cpl_table_set_double(tab, CRIRES_COL_EXTRACT_BG_NOISE, i, bgn) ;
    }
    return tab ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Create the spectrum profile
  @param    lamp_spec       The lamp spectrum
  @param    wl_sol          wavelength-pixel relation
  @param    display         Flag to display
  @param    qc_lines_flux   The median of the lines flux
  @param    qc_fwhm         The median if the fwhms
  @param    qc_rpower       The median of the resolving power
  @return   0 if ok -1 if error
 */
/*----------------------------------------------------------------------------*/
int crires_extract_lines_properties(
        const cpl_vector        *   lamp_spec,
        const cpl_polynomial    *   wl_sol,
        int                         display,
        double                  *   qc_lines_flux,
        double                  *   qc_fwhm,
        double                  *   qc_rpower)
{
    cpl_vector      *   bright_lines ;
    cpl_vector      *   fwhms ;
    cpl_vector      *   areas ;
    cpl_vector      *   rpower ;
    double          *   prpower ;
    double              wl, disp, fwhm_val, line_pos ;
    int                 i ;

    /* Test entries */
    if (lamp_spec == NULL) return -1 ;
    if (wl_sol == NULL) return -1 ;
    if (qc_lines_flux == NULL || qc_fwhm == NULL || qc_rpower == NULL) 
        return -1 ;

    /* Detect the brightest lines */
    cpl_msg_info(__func__, "Detect the brightest lines") ;
    cpl_msg_indent_more() ;
    if ((bright_lines = irplib_spectrum_detect_peaks(lamp_spec, 2, 2.0,
                    display, &fwhms, &areas)) == NULL) {
        cpl_msg_error(__func__, "Cannot detect lines");
        cpl_msg_indent_less() ;
        return -1 ;
    }
    cpl_msg_indent_less() ;

    /* Resolving power */
    rpower = cpl_vector_duplicate(bright_lines) ;
    prpower = cpl_vector_get_data(rpower) ;
    for (i=0 ; i<cpl_vector_get_size(rpower) ; i++) {
        line_pos = cpl_vector_get(bright_lines, i) ;
        wl = cpl_polynomial_eval_1d(wl_sol, line_pos, NULL) ;
        if (line_pos == 0)
            disp = cpl_polynomial_eval_1d(wl_sol, 1, NULL) - wl ;
        else
            disp = wl - cpl_polynomial_eval_1d(wl_sol, line_pos-1.0, NULL) ;
        fwhm_val = cpl_vector_get(fwhms, i) ;
        if (fwhm_val > 0)   prpower[i] = wl / (disp * fwhm_val) ;
        else                prpower[i] = 0.0 ;
    }
    cpl_vector_delete(bright_lines) ;

    /* Return */
    *qc_lines_flux = cpl_vector_get_median_const(areas) ;
    *qc_fwhm = cpl_vector_get_median_const(fwhms) ;
    *qc_rpower = cpl_vector_get_median_const(rpower) ;
    cpl_vector_delete(rpower) ;
    cpl_vector_delete(fwhms) ;
    cpl_vector_delete(areas) ;
    return 0 ;
}

/**@}*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Create the spectrum profile
  @param    ima         the ima with the horizontal spectrum
  @param    period          The illumination period
  @param    size        the horizontal size of the running window
  @param    spec_starty Start of the zone where the spectrum is
  @param    spec_stopy  Stop of the zone where the spectrum is
  @param    chip        the chip (1 to CRIRES_NB_DETECTORS)
  @param    spec_pos    the spectrum position or NULL
  @param    spec_w      computed spectrum width or NULL
  @return   the profile or NULL in error case

  The input image is expected to be FLOAT.
  
  Apply a running computation along X: the profile pixels are equal to
  the median of the box size x 1 around the pixel
  The spectrum is detected, and everything away of the spectrum is set
  to 0.
  Everything negative is then set to 0.
  The returned profile must be deallocated with cpl_image_delete().

  The passed size must be in [3, CRIRES_SPEC_BOX_SIZE_LIMIT].
  The created profile is a FLOAT image.
 */
/*----------------------------------------------------------------------------*/
static cpl_image * crires_extract_create_profile(
        const cpl_image     *   ima,
        crires_illum_period     period,
        int                     size,
        int                     spec_starty,
        int                     spec_stopy,
        int                     chip,
        int                     spec_hsize,
        int                 *   spec_pos,
        int                 *   spec_w)
{
    int                     nx, ny, hsize ;
    cpl_image           *   imad ;
    double              *   pimad ;
    cpl_image           *   profile ;
    float               *   pprofile ;
    cpl_vector          *   vec_med ;
    int                     vec_start, vec_size, ly ;
    int                     spec_pos_loc, low, high ;
    int                     i, j ;
    
    /* Test entries */
    if (ima == NULL) return NULL ;
    if (period == CRIRES_ILLUM_UNKNOWN) return NULL ;
    if (period == CRIRES_ILLUM_LOWER_HALF) return NULL ;
    if (period == CRIRES_ILLUM_UPPER_HALF) return NULL ;
    if (size < 3 || size > CRIRES_SPEC_BOX_SIZE_LIMIT) return NULL ;
    if (chip < 1 || chip > CRIRES_NB_DETECTORS) return NULL ;

    /* Initialise */
    nx = cpl_image_get_size_x(ima) ;
    ny = cpl_image_get_size_y(ima) ;
    hsize = (int)(size/2.0) ;
    ly = -1 ;

    /* Convert the input image in double */
    imad = cpl_image_cast(ima, CPL_TYPE_DOUBLE) ;
    pimad = cpl_image_get_data_double(imad) ;

    /* Copy the input image */
    profile = cpl_image_new(nx, ny, CPL_TYPE_FLOAT) ;
    pprofile = cpl_image_get_data_float(profile) ;
   
    /* Loop on the image and compute the median around the pixels -> profile  */
    for (i=0 ; i<nx ; i++) {
        if (i<hsize)                vec_start = 0 ;
        else                        vec_start = i-hsize ;
        if (vec_start + size >= nx) vec_size = nx - vec_start ;
        else                        vec_size = size ;
        for (j=0 ; j<ny ; j++) {
            vec_med = cpl_vector_wrap(vec_size, &(pimad[vec_start + j*nx])) ;
            pprofile[i+j*nx] = (float)cpl_vector_get_median_const(vec_med) ;
            cpl_vector_unwrap(vec_med) ;
        }
    }
    cpl_image_delete(imad) ;
    
    /* Spectrum detection */
    cpl_msg_info(__func__, "Spectrum position detection") ;
    if ((spec_pos_loc = crires_extract_detect(profile, period,
                    spec_starty, spec_stopy, chip)) < 0) {
        cpl_msg_error(__func__, "Cannot detect any spectrum") ;
        cpl_image_delete(profile) ;
        return NULL ;
    }
    if (spec_pos != NULL) *spec_pos = spec_pos_loc ; 

    /* Get the correction for the position */
    if (chip == 1) ly = crires_get_detector_ly1(period) ;
    if (chip == 2) ly = crires_get_detector_ly2(period) ;
    if (chip == 3) ly = crires_get_detector_ly3(period) ;
    if (chip == 4) ly = crires_get_detector_ly4(period) ;
    /* If ly not found, use the bottom of the image */
    if (ly < 0) ly = 1 ;

    spec_pos_loc -= ly - 1 ;

    /* Bounds */
    low = spec_pos_loc - spec_hsize ;
    high = spec_pos_loc + spec_hsize ;

    /* Check the bounds */
    if (low < 1) low = 1 ;
    if (high > ny) high = ny ;
    if (low > high) {
        cpl_image_delete(profile) ;
        return NULL ;
    }
    
    /* Clean the profile */
    for (i=0 ; i<nx ; i++) {
        for (j=0 ; j<low-1 ; j++) pprofile[i+j*nx] = 0.0 ;
        for (j=high ; j<ny ; j++) pprofile[i+j*nx] = 0.0 ;
    }

    /* Threshold the negative values */
    cpl_image_threshold(profile, 0.0, DBL_MAX, 0.0, 0.0) ; 

    /* Return  */
    if (spec_w != NULL) *spec_w = 2*spec_hsize+1 ;
    return profile ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Detection of the positive spectrum
  @param    profile     the computed spectrum profile
  @param    period          The illumination period
  @param    spec_starty Start of the zone where the spectrum is
  @param    spec_stopy  Stop of the zone where the spectrum is
  @param    chip        the chip (1 to CRIRES_NB_DETECTORS)
  @return   the spectrum position or -1 in error case

  The profile is horizontally collapsed, and the maximum gives the
  positive spectrum position. 
  The horizontal collapse excludes the right and left 10%

  Vertical 100 pixels width bands are collapsed horizontally. The
  spectrum is detected from these bands. The median is kept.
 */
/*----------------------------------------------------------------------------*/
static int crires_extract_detect(
        const cpl_image     *   profile,
        crires_illum_period     period,
        int                     spec_starty,
        int                     spec_stopy,
        int                     chip)
{
    cpl_image           *   collapsed ;
    cpl_image           *   filtered ;
    cpl_mask            *   kernel ;
    int                     nx, ny, llx, lly, urx, ury ;
    double                  max ;
    cpl_size                max_x, max_y ;
    int                     final_pos ;
    int                     nbands, band_width, ly ;
    cpl_vector          *   pos ;
    int                     i ;
    
    /* Test entries */
    if (profile == NULL) return -1 ;
    if (period == CRIRES_ILLUM_UNKNOWN) return -1 ;
    if (period == CRIRES_ILLUM_LOWER_HALF) return -1 ;
    if (period == CRIRES_ILLUM_UPPER_HALF) return -1 ;
    if (chip < 1 || chip > CRIRES_NB_DETECTORS) return -1 ;

    /* Initialize */
    ly = -1 ;
    nx = cpl_image_get_size_x(profile) ;
    ny = cpl_image_get_size_y(profile) ;
    nbands = 8 ;
    band_width = 100 ;
    
    /* Re-adjust the position */
    if (chip == 1) ly = crires_get_detector_ly1(period) ;
    if (chip == 2) ly = crires_get_detector_ly2(period) ;
    if (chip == 3) ly = crires_get_detector_ly3(period) ;
    if (chip == 4) ly = crires_get_detector_ly4(period) ;
    /* If ly not found, use the bottom of the image */
    if (ly < 0) ly = 1 ;

    /* Restrict search zone if specified */
    if (spec_starty > 0) {
        lly = spec_starty - (ly - 1) ;
    } else {
        lly = 1 ;
    }
    if (spec_stopy > 0) {
        ury = spec_stopy - (ly - 1) ;
    } else {
        ury = ny ;
    }

    /* Verify search zone */
    if (lly > ury || lly<1 || ury>ny) {
        cpl_msg_error(__func__, "Invalid search zone specification <> %d - %d",
                ly, ly + ny - 1) ;
        return -1 ;
    }
    
    /* Median vertical filtering 1x3 */
    kernel = cpl_mask_new(1, 3) ;
    cpl_mask_not(kernel) ;
    filtered = cpl_image_new(
            cpl_image_get_size_x(profile),
            cpl_image_get_size_y(profile),
            cpl_image_get_type(profile)) ;
    if (cpl_image_filter_mask(filtered, profile, kernel,
                CPL_FILTER_MEDIAN, CPL_BORDER_FILTER) != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot filter the profile") ;
        cpl_mask_delete(kernel) ;
        cpl_image_delete(filtered) ;
        return -1 ;
    }
    cpl_mask_delete(kernel) ;
 
    /* Create positions vector */
    pos = cpl_vector_new(nbands) ;

    /* Loop on the bands */
    for (i=0 ; i<nbands ; i++) {
        llx = band_width * (i+1) ; 
        urx = llx + band_width ;

        /* Collapse the profile */
        if ((collapsed = cpl_image_collapse_window_create(filtered, llx, lly,
                        urx, ury, 1)) == NULL) {
            cpl_msg_error(__func__, "Cannot collapse the spectrum") ;
            cpl_image_delete(filtered) ;
            cpl_vector_delete(pos) ;
            return -1 ;
        } 

        /* Get the max and min infos */
        max = cpl_image_get_max(collapsed) ;
        cpl_image_get_maxpos(collapsed, &max_x, &max_y) ;
        /* Collapse is done on a ZONE - max_y is the position in the image */
        max_y += lly - 1 ;
        cpl_image_delete(collapsed) ;

        /* Fill pos */
        cpl_vector_set(pos, i, (double)max_y) ;
        
        /* All negative - no spectrum */
        if (max<0) {
            cpl_image_delete(filtered) ;
            cpl_vector_delete(pos) ;
            return -1 ;
        }
    }
    cpl_image_delete(filtered) ;

    /* Get the position */
    final_pos = cpl_vector_get_median_const(pos) ;
    cpl_vector_delete(pos) ;

    return final_pos + ly - 1 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the background noise with an iterative sigma clipping 
  @param    profile     the profile
  @param    contrib     the contribution map or NULL
  @param    kappa       the kappa for bgd identification
  @param    closing_hs  Half size of the kernel for the closing on the bgmap
  @param    clean_rate  Min rate of bgd pixels per line allowed
  @param    bg_map      the returned background map or NULL
  @return   the computed sigmas vector or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_vector * crires_extract_bgd_rms(
        const cpl_image     *   profile,
        const cpl_image     *   contrib,
        double                  kappa,
        int                     closing_hs,
        double                  clean_rate,
        cpl_image           **  bg_map)
{
    int                 nx, ny, bpm_nx, bpm_ny ;
    const float     *   pprofile ;
    cpl_mask        *   kernel ;
    cpl_mask        *   mask ;
    cpl_image       *   noise ;
    float           *   pnoise ;
    cpl_vector      *   rms ;
    double          *   prms ;
    cpl_vector      *   bgd ;
    double          *   pbgd ;
    cpl_vector      *   low_thresh ;
    cpl_vector      *   high_thresh ;
    double              max_val ;
    cpl_mask        *   bpm ;
    cpl_mask        *   bpm_start ;
    cpl_binary      *   pbpm ;
    const int           niterations = 5 ;
    double              sqsum, sum ;
    int                 npix, pos, nbgd ;
    double              lt, ht ;
    int                 i, j, k ;

    /* Check entries */
    if (profile == NULL) return NULL ;
    
    /* Initialise */
    nx = cpl_image_get_size_x(profile) ;
    ny = cpl_image_get_size_y(profile) ;
    pprofile = cpl_image_get_data_float_const(profile) ;

    /* Median vertical filtering 1x5 */
    mask = cpl_mask_new(1, 5) ;
    cpl_mask_not(mask) ;
    noise = cpl_image_new(
            cpl_image_get_size_x(profile),
            cpl_image_get_size_y(profile),
            cpl_image_get_type(profile)) ;
    if (cpl_image_filter_mask(noise, profile, mask,
                CPL_FILTER_MEDIAN, CPL_BORDER_FILTER) != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot filter the profile") ;
        cpl_mask_delete(mask) ;
        cpl_image_delete(noise) ;
        return NULL ;
    }
    cpl_mask_delete(mask) ;
    pnoise = cpl_image_get_data_float(noise) ;

    /* (Inverted) Noise image */
    cpl_image_subtract(noise, profile);

    /* Create the vectors */
    rms = cpl_vector_new(nx) ;
    prms = cpl_vector_get_data(rms) ;
    bgd = cpl_vector_new(nx) ;
    pbgd = cpl_vector_get_data(bgd) ;

    /* Thresholds */
    low_thresh = cpl_vector_new(nx) ;
    high_thresh = cpl_vector_new(nx) ;

    /* Use the contribution map if provided */
    if (contrib != NULL) {
        /* Threshold to keep only the maximum */
        max_val = cpl_image_get_max(contrib) ;
        bpm=cpl_mask_threshold_image_create(contrib, max_val-0.5, max_val+0.5) ;
        cpl_mask_not(bpm) ;
    } else bpm = cpl_mask_new(nx, ny) ;
    pbpm = cpl_mask_get_data(bpm) ;

    /* Keep the bpm of the contrib */
    bpm_start = cpl_mask_duplicate(bpm) ;
 
    /* Iterations */
    for (i=0 ; i<niterations ; i++) {
        /* Get mean and stdev */
        for (j=0 ; j<nx ; j++) {
            sqsum = sum = 0.0 ;
            npix = 0 ;
            for (k=0 ; k<ny ; k++) {
                pos = j+k*nx ;
                if (pbpm[pos] == CPL_BINARY_0) {
                    sqsum += (double)(pnoise[pos] * pnoise[pos]) ;
                    sum += (double)pnoise[pos] ;
                    npix++ ;
                }
            }
            if (npix <= 1) {
                prms[j] = 0 ;
                pbgd[j] = 0 ;
            } else {
                prms[j] = sqrt((sqsum-(sum*sum/npix))/(npix-1)) ;
                pbgd[j] = sum / npix ;
            }
        }
        
        /* Set the thresholds */
        cpl_vector_copy(low_thresh, rms) ;
        cpl_vector_multiply_scalar(low_thresh, kappa*(-1)) ;
        cpl_vector_add(low_thresh, bgd) ;
        cpl_vector_copy(high_thresh, rms) ;
        cpl_vector_multiply_scalar(high_thresh, kappa) ;
        cpl_vector_add(high_thresh, bgd) ;

        /* Identify where mean-sigma*stdev < noise < mean+sigma*stdev */
        for (j=0 ; j<nx ; j++) {
            lt = cpl_vector_get(low_thresh, j) ;
            ht = cpl_vector_get(high_thresh, j) ;
            for (k=0 ; k<ny ; k++) {
                if (pprofile[j+k*nx] > ht || pprofile[j+k*nx] < lt) 
                    pbpm[j+k*nx] = CPL_BINARY_1 ;
                else pbpm[j+k*nx] = CPL_BINARY_0 ;
            }
        }
        cpl_mask_or(bpm, bpm_start) ;
    }
    cpl_mask_delete(bpm_start) ;
    cpl_vector_delete(low_thresh) ;
    cpl_vector_delete(high_thresh) ;
    cpl_vector_delete(bgd) ;
    cpl_image_delete(noise) ;

    /* Invert the bpm to flag the background */
    cpl_mask_not(bpm) ;

    /* Closing on the background map */
    if (closing_hs > 0) {
        kernel = cpl_mask_new(2*closing_hs+1, 2*closing_hs+1) ;
        cpl_mask_not(kernel) ;
        (void)cpl_mask_filter(bpm, bpm, kernel, CPL_FILTER_CLOSING, 
                CPL_BORDER_ZERO);
        cpl_mask_delete(kernel) ;
    }

    /* Horizontal cleaning */
    if (clean_rate > 0) {
        pbpm = cpl_mask_get_data(bpm) ;
        bpm_nx = cpl_mask_get_size_x(bpm) ;
        bpm_ny = cpl_mask_get_size_y(bpm) ;
        for (j=0 ; j<bpm_ny ; j++) {
            nbgd = 0 ;
            for (i=0 ; i<bpm_nx ; i++) {
                if (pbpm[i+j*bpm_nx] == CPL_BINARY_1) nbgd ++ ;
            }
            if (nbgd < (int)(clean_rate*bpm_nx)) {
                for (i=0 ; i<bpm_nx ; i++) {
                    pbpm[i+j*bpm_nx] = CPL_BINARY_0 ;
                }
            }
        }
    }

    /* Compute the stats on the background of the input image */
    for (i=0 ; i<nx ; i++) {
        sqsum = sum = 0.0 ;
        npix = 0 ;
        for (j=0 ; j<ny ; j++) {
            pos = i+j*nx ;
            if (pbpm[pos] == CPL_BINARY_1) {
                sqsum += (double)(pprofile[pos] * pprofile[pos]) ;
                sum += (double)pprofile[pos] ;
                npix++ ;
            }
        }
        if (npix <= 1)  prms[i] = 0 ;
        else            prms[i] = sqrt((sqsum-(sum*sum/npix))/(npix-1)) ;
    }

    /* Return the background map if requested */
    if (bg_map != NULL)     *bg_map = cpl_image_new_from_mask(bpm) ;
    cpl_mask_delete(bpm);
    return rms ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the rectangular profile
  @param    pos         spectrum position
  @param    hsize       spectrum half-size
  @param    lx          x size of the profile to generate
  @param    ly          y size of the profile to generate
  @param    spec_w      computed spec width
  @param    weight      computed normalisation factor
  @return   the profile or NULL in error case

  The returned image must be deallocated with cpl_image_delete()
 */
/*----------------------------------------------------------------------------*/
static cpl_image * crires_extract_profile_rect(
        int             pos,
        int             hsize,
        int             lx,
        int             ly,
        int         *   spec_w,
        cpl_vector  **  weight)
{
    cpl_image   *   profile ;
    float       *   pprofile ;
    int             low, high ;
    double          norm ;
    cpl_vector  *   weight_loc ;
    int             i, j ;

    /* Test entries */
    if (weight == NULL) return NULL ;
    if (hsize < 0) return NULL ;
    if (spec_w == NULL) return NULL ;
    
    /* Bounds */
    low = pos - hsize ;
    high = pos + hsize ;
    if (low < 1) low = 1 ;
    if (high > ly) high = ly ;
    if (low>high) return NULL ;

    /* Norm */
    norm = high - low + 1 ;
    
    /* Create profile */
    profile = cpl_image_new(lx, ly, CPL_TYPE_FLOAT) ;
    pprofile = cpl_image_get_data_float(profile) ;

    /* Fill the profile */
    for (i=0 ; i<lx ; i++) {
        for (j=low-1 ; j<high ; j++) {
            pprofile[i+j*lx] = 1.0/norm ;
        }
    }

    /* Create the weight */
    weight_loc = cpl_vector_new(lx) ;
    cpl_vector_fill(weight_loc, norm) ;
    
    /* Return  */
    *weight = weight_loc ;
    *spec_w = 2 * hsize + 1 ;
    return profile ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the optimal profile
  @param    prof_init   the initial profile
  @param    bgd_rms     the background RMS
  @param    chip        the chip number
  @param    pos         spectrum position
  @param    tot_ndit    ndit * nexp * dit
  @param    size        median filter width
  @param    weight      computed normalisation factor
  @return   the profile or NULL in error case

  - pix = pix / ( (pix/(conv*tot_ndit)) + size*bgd_rms^2)

  where conv is the gain that depend on the chip
        tot_ndit is DIT * NDIT * NEXP
        bgd_rms is the background noise
        size is the median filter width
        
  - Normalise and return the normalisation factors
  
  The returned image must be deallocated with cpl_image_delete()
 */
/*----------------------------------------------------------------------------*/
static cpl_image * crires_extract_profile_opt(
        cpl_image           *   prof_init,
        cpl_vector          *   bgd_rms,
        int                     chip,
        int                     pos,
        double                  tot_ndit,
        int                     size,
        cpl_vector          **  weight)
{
    cpl_image   *   profile ;
    float       *   pprofile ;
    double          val, conv, corr_fact ;
    int             nx, ny ;
    cpl_vector  *   weight_loc ;
    cpl_vector  *   bgd_rms_filt ;
    int             i, j ;

    /* Test entries */
    if (weight == NULL) return NULL ;
    if (tot_ndit <= 0) return NULL ;
    if (prof_init == NULL) return NULL ;
    if (size <= 0) return NULL ;
    nx = cpl_image_get_size_x(prof_init) ;
    ny = cpl_image_get_size_y(prof_init) ;
    if (pos < 1 || pos > ny) return NULL ;
    if (chip < 1 || chip > CRIRES_NB_DETECTORS) return NULL ;
    
    /* Initialise */
    if (chip == 1)      conv = CRIRES_GAIN_CHIP1 ;
    else if (chip == 2) conv = CRIRES_GAIN_CHIP2 ; 
    else if (chip == 3) conv = CRIRES_GAIN_CHIP3 ; 
    else if (chip == 4) conv = CRIRES_GAIN_CHIP4 ; 
    else return NULL ;

    /* Create profile */
    profile = cpl_image_duplicate(prof_init) ;
    pprofile = cpl_image_get_data_float(profile) ;

    /* Smoothe the background noise */
    bgd_rms_filt = cpl_vector_filter_median_create(bgd_rms, 15) ;
    
    /* Compute new profile with the background rms */
    for (i=0 ; i<nx ; i++) {
        val = cpl_vector_get(bgd_rms_filt, i) ;
        corr_fact = val * val * size ;
        for (j=0 ; j<ny ; j++) {
            if (((pprofile[i+j*nx]/(conv * tot_ndit)) + corr_fact) <= 0.0)
                val = 0.0 ;
            else 
                val = pprofile[i+j*nx] / ((pprofile[i+j*nx]/(conv * tot_ndit)) +
                        corr_fact) ;
            pprofile[i+j*nx] = val ;
        }
    }
    cpl_vector_delete(bgd_rms_filt) ;
    
    /* Normalise and compute the weight */
    if ((weight_loc = crires_extract_normalise_profile(profile)) == NULL) {
        cpl_msg_error(__func__, "Cannot normlise the weights") ;
        cpl_image_delete(profile) ;
        return NULL ;
    }
    
    /* Return  */
    *weight = weight_loc ;
    return profile ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the correction factor to apply to the optimal extraction
  @param    prof_init   the initial profile
  @param    prof_opt    the optimal profile
  @return   the correction or NULL in error case

  The returned image must be deallocated with cpl_vector_delete()
 */
/*----------------------------------------------------------------------------*/
static cpl_vector * crires_extract_opt_correction(
        cpl_image           *   prof_init,
        cpl_image           *   prof_opt)
{
    cpl_image       *   normalised ;
    cpl_vector      *   weights ;
    cpl_vector      *   correction ;
    double          *   pcorrection ;
    int                 i ;

    /* Test entries */
    if (prof_init == NULL) return NULL ;
    if (prof_opt == NULL) return NULL ;

    /* Normalise the profile */
    normalised = cpl_image_duplicate(prof_init) ;
    if ((weights = crires_extract_normalise_profile(normalised)) == NULL) {
        cpl_msg_error(__func__, "Cannot normalise the profile") ;
        cpl_image_delete(normalised) ;
        return NULL ;
    }
    cpl_vector_delete(weights) ;

    /* Extract the normalised profile */
    if ((correction=crires_extract_compute_spectrum(normalised,
                    prof_opt)) == NULL) {
        cpl_msg_error(__func__, "Cannot extract the normalised spectrum") ;
        cpl_image_delete(normalised) ;
        return NULL ;
    }
    cpl_image_delete(normalised) ;

    /* Set to 1 the values that are 0 */
    pcorrection = cpl_vector_get_data(correction) ;
    for (i=0 ; i<cpl_vector_get_size(correction) ; i++) {
        if (pcorrection[i] < 1e-5) pcorrection[i] = 1.0 ;
    }

    /* Free and return */
    return correction ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the error
  @param    image       the spectrum image
  @param    bgd_rms     the background RMS
  @param    profile     the profile
  @param    tot_ndit    ndit * nexp * dit
  @param    chip        the chip number (1 to 4)
  @return   the error or NULL in error case

  The error sig_tot is:

  sig_tot^2 = sig_r^2 + sig_g^2

  sig_r = bgd_rms * sqrt(sum(w^2)) 
  
  sig_g = sqrt(sum(w^2 * s / (conv*tot_ndit)))

  where w is the normalised weight (from profile)
        s is the signal (from image)    
        conv is a factor that depends on the chip
        tot_ndit is DIT * NDIT * NEXP
        
  The returned vector must be deallocated with cpl_vector_delete()
 */
/*----------------------------------------------------------------------------*/
static cpl_vector * crires_extract_compute_error(
        const cpl_image     *   image,
        const cpl_vector    *   bgd_rms,
        const cpl_image     *   profile,
        double                  tot_ndit,
        int                     chip)
{
    cpl_vector  *   error_r ;
    double      *   perror_r ;
    cpl_vector  *   error_g ;
    double          conv ;
    cpl_image   *   prof_sq ;
    cpl_image   *   prof_sq_sig ;
    cpl_image   *   collapsed ;
    int             i ;

    /* Test entries */
    if (image == NULL) return NULL ;
    if (bgd_rms == NULL) return NULL ;
    if (profile == NULL) return NULL ;
    if (chip < 1 || chip > CRIRES_NB_DETECTORS) return NULL ;

    /* Initialise */
    if (chip == 1)      conv = CRIRES_GAIN_CHIP1 ;
    else if (chip == 2) conv = CRIRES_GAIN_CHIP2 ; 
    else if (chip == 3) conv = CRIRES_GAIN_CHIP3 ; 
    else if (chip == 4) conv = CRIRES_GAIN_CHIP4 ; 
    else return NULL ;

    /* Compute profile * profile */
    prof_sq = cpl_image_duplicate(profile) ;
    cpl_image_multiply(prof_sq, profile) ;

    /* Compute profile * profile * image / conv */
    prof_sq_sig = cpl_image_multiply_create(prof_sq, image) ;
    cpl_image_divide_scalar(prof_sq_sig, conv*tot_ndit) ;

    /* Create the error vectors */
    collapsed = cpl_image_collapse_create(prof_sq, 0) ;
    cpl_image_delete(prof_sq) ;
    error_r = cpl_vector_new_from_image_row(collapsed, 1) ;
    cpl_image_delete(collapsed) ;
    collapsed = cpl_image_collapse_create(prof_sq_sig, 0) ;
    cpl_image_delete(prof_sq_sig) ;
    error_g = cpl_vector_new_from_image_row(collapsed, 1) ;
    cpl_image_delete(collapsed) ;

    /* Compute error_r */
    cpl_vector_multiply(error_r, bgd_rms) ;
    cpl_vector_multiply(error_r, bgd_rms) ;

    /* Compute error */
    cpl_vector_add(error_r, error_g) ;
    cpl_vector_delete(error_g) ;
    
    /* Set the negative values to 0 */
    perror_r = cpl_vector_get_data(error_r) ;
    for (i=0 ; i<cpl_vector_get_size(error_r) ; i++) {
        if (perror_r[i] < 0) perror_r[i] = 0 ;
    }

    /* Take the Sq root */
    cpl_vector_sqrt(error_r) ;
    
    /* Return  */
    return error_r ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Extract the spectrum
  @param    spec_image      the image with the horizontal spectrum
  @param    norm_profile    the normalised profile
  @return   the extracted spectrum or NULL in error case

  The spectrum is computed by multiplying the image with the profile,
  and by collapsing it in the y direction.

  The returned vector must be deallocated with cpl_vector_delete()
 */
/*----------------------------------------------------------------------------*/
static cpl_vector * crires_extract_compute_spectrum(
        const cpl_image     *   spec_image,
        const cpl_image     *   norm_profile) 
{
    cpl_image           *   product ;
    cpl_image           *   collapsed ;
    cpl_vector          *   spec ;
    
   /* Test entries */
    if (spec_image == NULL) return NULL ;
    if (norm_profile == NULL) return NULL ;

    /* Multiply the 2 images */
    if ((product = cpl_image_multiply_create(spec_image, norm_profile))==NULL){
        cpl_msg_error(__func__, "Cannot multiply 2 images");
        return NULL ;
    }

    /* Compute the sum of the columns */
    if ((collapsed = cpl_image_collapse_create(product, 0)) == NULL) {
        cpl_msg_error(__func__, "Cannot collapse the product");
        cpl_image_delete(product) ;
        return NULL ;
    }
    cpl_image_delete(product) ;

    /* Put the result in a vector */
    if ((spec = cpl_vector_new_from_image_row(collapsed, 1)) == NULL) {
        cpl_msg_error(__func__, "Cannot get the vector");
        cpl_image_delete(collapsed) ;
        return NULL ;
    }
    cpl_image_delete(collapsed) ;

    return spec ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Normalise the profile
  @param    profile     the input profile (modified)
  @return   the weight values or NULL in error case

  The returned vector must be de allocated with cpl_vector_delete()
 */
/*----------------------------------------------------------------------------*/
static cpl_vector * crires_extract_normalise_profile(cpl_image * profile)
{
    cpl_vector          *   weight ;
    double              *   pweight ;
    float               *   pprofile ;
    int                     nx, ny ;
    int                     i, j ;
    
    /* Test entries */
    if (profile == NULL) return NULL; 

    /* Initialize */
    nx = cpl_image_get_size_x(profile) ;
    ny = cpl_image_get_size_y(profile) ;
    pprofile = cpl_image_get_data_float(profile) ;

    /* Crestate the output vector */
    weight = cpl_vector_new(nx) ;
    pweight = cpl_vector_get_data(weight) ;

    /* Loop on the columns */
    for (i=0 ; i<nx ; i++) {
        /* Normalise the profile column by column */
        pweight[i]=cpl_image_get_flux_window(profile, i+1, 1, i+1, ny) ;
        if (pweight[i] > 0) 
            for (j=0 ; j<ny ; j++) pprofile[i+j*nx] /= (float)pweight[i] ;
    }
    return weight ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Set to 0 the sides of the spectrum
  @param    spec        the spectrum
  @param    left        the nb of samples on the left
  @param    right       the nb of samples on the right
  @return   0 if ok, -1 in error case
 */
/*----------------------------------------------------------------------------*/
static int crires_extract_reject_sides(
        cpl_vector      *   spec,
        int                 left,
        int                 right)
{
    int         nsamples ;
    double  *   pspec ;
    int         i ;
    
    /* Test entries */
    if (spec == NULL) return -1 ;

    /* Initialize */
    nsamples = cpl_vector_get_size(spec) ;

    /* Test entries */
    if (left > nsamples || right > nsamples) return -1 ;

    /* Set the sides to 0 */
    pspec = cpl_vector_get_data(spec) ;
    for (i=0 ; i<left ; i++) pspec[i] = 0 ;
    for (i=0 ; i<right ; i++) pspec[nsamples-i-1] = 0 ;
    
    return 0 ;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    Get the fwhms of an horizontal spectrum and returns the median 
  @param    ima         the image
  @return   the fwhm median or -1.0 in error case
 */
/*----------------------------------------------------------------------------*/
static double crires_extract_fwhm_med(const cpl_image * ima)
{
    cpl_bivector    *   fwhms ;
    double              fwhm_med ;
    cpl_vector      *   fwhms_nonzero ;
    int                 n_nonzero ;
    const double    *   pfwhms ;
    int                 i ;

    /* Test entries */
    if (ima == NULL) return -1 ;

    /* Initialize */
    n_nonzero = 0 ;

    /* Compute fwhms */
    if ((fwhms = crires_extract_spatial_fwhm(ima)) == NULL) return -1.0 ;

    /* Get the median of the non-zero fwhms */
    pfwhms = cpl_bivector_get_y_data_const(fwhms) ;

    /* Count the nonzero values */
    for (i=0 ; i<cpl_bivector_get_size(fwhms) ; i++) {
        if (pfwhms[i] > 1e-3) n_nonzero++ ;
    }
    if (n_nonzero == 0) return -1.0 ;

    /* Create the nonzero values vector */
    fwhms_nonzero = cpl_vector_new(n_nonzero) ;
    n_nonzero = 0 ;
    for (i=0 ; i<cpl_bivector_get_size(fwhms) ; i++) {
        if (pfwhms[i] > 1e-3) {
            cpl_vector_set(fwhms_nonzero, n_nonzero, pfwhms[i]) ;
            n_nonzero++ ;
        }
    }
    cpl_bivector_delete(fwhms) ;

    /* Compute the median */
    fwhm_med = cpl_vector_get_median(fwhms_nonzero) ;
    cpl_vector_delete(fwhms_nonzero) ;

    return fwhm_med ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Merge 4 spectra
  @param    spec1       the spectrum nb 1
  @param    spec2       the spectrum nb 2
  @param    spec3       the spectrum nb 3
  @param    spec4       the spectrum nb 4
  @param    gap         gap between the spectra
  @return   the merged spectrum or NULL
 */
/*----------------------------------------------------------------------------*/
cpl_vector * crires_merge_spectra(
        const cpl_vector    *   spec1,
        const cpl_vector    *   spec2,
        const cpl_vector    *   spec3,
        const cpl_vector    *   spec4,
        int                     gap)
{
    cpl_vector      *   out ;
    double          *   pout ;
    const double    *   pin ;
    int                 out_size, size1, size2, size3, size4 ;
    int                 i ;
    
    /* Test entries */
    if (spec1 == NULL) return NULL ;
    if (spec2 == NULL) return NULL ;
    if (spec3 == NULL) return NULL ;
    if (spec4 == NULL) return NULL ;
    if (gap < 0) return NULL ;

    /* Initialise */
    size1 = cpl_vector_get_size(spec1) ;
    size2 = cpl_vector_get_size(spec2) ;
    size3 = cpl_vector_get_size(spec3) ;
    size4 = cpl_vector_get_size(spec4) ;
    out_size = size1 + size2 + size3 + size4 + 3*gap ;

    /* Create output vector */
    out = cpl_vector_new(out_size) ;
    cpl_vector_fill(out, 0.0) ;
    pout = cpl_vector_get_data(out) ;
    
    /* First vector */
    pin = cpl_vector_get_data_const(spec1) ;
    for (i=0 ; i<size1 ; i++) pout[i] = pin[i] ;
    /* Second vector */
    pin = cpl_vector_get_data_const(spec2) ;
    for (i=0 ; i<size2 ; i++) pout[size1+gap+i] = pin[i] ;
    /* Third vector */
    pin = cpl_vector_get_data_const(spec3) ;
    for (i=0 ; i<size3 ; i++) pout[size1+size2+2*gap+i] = pin[i] ;
    /* Fourth vector */
    pin = cpl_vector_get_data_const(spec4) ;
    for (i=0 ; i<size4 ; i++) pout[size1+size2+size3+3*gap+i] = pin[i] ;

    /* Return */
    return out ;
}


