/* $Id: crires_model_refining.c,v 1.36 2012-09-19 14:10:27 yjung Exp $
 *
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2012-09-19 14:10:27 $
 * $Revision: 1.36 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include <math.h>
#include <string.h>

#include "crires_model_refining.h"
#include "crires_utils.h"
#include "crires_dfs.h"
#include "crires_model_kernel.h"
#include "crires_model_io.h"
#include "crires_model_qmetric.h"
#include "crires_pfits.h"
#include "crires_combine.h"
/*----------------------------------------------------------------------------*/
/**
 * @defgroup crires_model_refining    Model Refining routines
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief   Extract the spectrum along a given shape
  @param    in          the input image
  @param    period
  @param    trace       the spectrum trace
  @param    bpm         the bad pixels map
  @param    flat        the flat field
  @param    detlin      detlin coeffs
  @param    chip    the chip (1 to CRIRES_NB_DETECTORS)
  @return   The extracted spectrum or NULL in error case

  The input image must be float
 */
/*----------------------------------------------------------------------------*/
cpl_vector * crires_model_refining_extract(
        const cpl_image     *   in,
        crires_illum_period     period,
        cpl_vector    *   trace,
        const char          *   bpm,
        const char          *   flat,
        const char          *   detlin,
        int                     chip)
{
    cpl_imagelist   *       ilist_tmp ;
    cpl_image       *       in_loc ;
    cpl_image       *       unfiltered ;
    cpl_image       *       filtered ;
    cpl_mask        *       mask;
    cpl_vector      *       extracted ;
    double          *       pextracted ;
    double    *       ptrace ;
    double                  ly ;
    double tot, moment;
    int h=5;
    int                     nx, pos, ypos ;
    int                     box_hsize ;
    int                     i,j ;
    int minx,maxx;
    double cont;

    /* Test entries */
    if (in == NULL) return NULL ;
    if (period == CRIRES_ILLUM_UNKNOWN) return NULL ;
    if (trace == NULL) return NULL ;
    if (chip < 1 || chip > CRIRES_NB_DETECTORS) return NULL ;
    ly = -1 ;

    /* Initialise */
    if (chip == 1) ly = crires_get_detector_ly1(period) ;
    if (chip == 2) ly = crires_get_detector_ly2(period) ;
    if (chip == 3) ly = crires_get_detector_ly3(period) ;
    if (chip == 4) ly = crires_get_detector_ly4(period) ;
    /* If ly not found, use the bottom of the image */
    if (ly < 0) ly = 1 ;

    box_hsize = 3 ; /*actual size is 2n+1*/
    nx = cpl_image_get_size_x(in) ;
    if (cpl_vector_get_size(trace)-1 != nx) return NULL ;
    /* The -1 is because trace now has an extra entry holding the
       entrance slit posn in mm*/

    /* Duplicate the input image */
    in_loc = cpl_image_duplicate(in) ;

    /* Apply the calibrations */
    ilist_tmp = cpl_imagelist_new() ;
    cpl_imagelist_set(ilist_tmp, in_loc, 0) ; 
    if (crires_calib_chip_list(ilist_tmp, period, chip, flat, NULL, bpm, 
                detlin, 0.0, -1, -1, -1)) {
        cpl_msg_error(__func__, "Cannot apply the calibrations") ;
        cpl_imagelist_delete(ilist_tmp) ;
        return NULL ;
    }
    in_loc = cpl_imagelist_unset(ilist_tmp, 0) ;
    cpl_imagelist_delete(ilist_tmp) ;

    unfiltered = cpl_image_duplicate(in_loc) ;

    /* Filter the bad pixels */
    if (bpm == NULL) {
       /* Median vertical filtering 1x5 */
        mask = cpl_mask_new(1, 5) ;
        cpl_mask_not(mask) ;
        filtered = cpl_image_new(
                cpl_image_get_size_x(in_loc),
                cpl_image_get_size_y(in_loc),
                cpl_image_get_type(in_loc)) ;
        if (cpl_image_filter_mask(filtered, in_loc, mask,
                    CPL_FILTER_MEDIAN, CPL_BORDER_COPY) != CPL_ERROR_NONE) {
            cpl_msg_error(__func__, "Cannot filter the image") ;
            cpl_mask_delete(mask) ;
            cpl_image_delete(filtered) ;
            return NULL ;
        }
        cpl_mask_delete(mask) ;
    } else {
        filtered = cpl_image_duplicate(in_loc) ;
    }
    cpl_image_delete(in_loc) ;

    /* Check */
    if (filtered == NULL) {
        cpl_msg_error(__func__, "Cannot filter the image") ;
        return NULL ;
    }

    /* Create the vector */
    extracted = cpl_vector_new(nx) ;

    /* Extract the spectrum */
    pextracted = cpl_vector_get_data(extracted) ;
    ptrace = cpl_vector_get_data(trace) ;
    for (i=0 ; i<nx ; i++) {
      tot=0.0;
      moment=0;
      ypos = (int)(ptrace[i] - (ly-1)) ;
      /*Now modify the trace so that each element contains no longer the model
	predicated ypos, but instead the barycenter of flux in a box 2h+1 high
	centered on the model predicted ypos and 3 pix wide (current column
	and direct neighbours. This value will only be meaningful where there
	is a feature. The trace will next be used to get the y centroids of
	the matched features in the match function*/
      /* 07.07.11pb Updated to modify the trace before the extraction so that
	 the extracted flux is centred on the feature, this somewhat improves
	 the S/N*/
      if (i>50) {
	minx=i-50;
      }
      else {
	minx=1;
      }
      if (i<973) {
	maxx=i+50;
      }
      else {
	maxx=1023;
      }
      for (j=-h; j<=h; j++) {
	if (i>0 && i<1023) {
	  cont=cpl_image_get_median_window(unfiltered, i-1,ypos+j, i+1, ypos+j)-
	       cpl_image_get_median_window(unfiltered,minx, ypos-30,maxx, ypos-20);
	  if (cont<0.0) cont=0.0;
	  tot+=cont;
	  moment+=cont*(float)(j);
	  //printf("%d %d %d %lf %lf %lf %lf %lf\n",i,ypos,j,cpl_image_get_median_window(unfiltered, i-1,ypos+j, i+1, ypos+j),tot,cpl_image_get_median_window(unfiltered,minx, ypos-30,maxx, ypos-20),cont,cont*(float)(j));
	}
      } 
      if (tot>0.0) ptrace[i]=(ly-1)+(float)(ypos)+moment/tot;
      ypos = (int)(ptrace[i] - (ly-1)) ;
      if (ypos == 0) {
	pextracted[i] = 0.0 ;
      } else {
	pos = i + ypos * nx ;
	pextracted[i] = cpl_image_get_mean_window(filtered, i+1,
						  ypos-box_hsize,
						  i+1, ypos+box_hsize) ;
      } 
      //printf("%d %lf %d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf \n", i, pextracted[i], ypos+(int)(ly), ptrace[i], moment, tot, moment/tot,cpl_image_get(filtered,i,ypos+(int)(ly),&whatever),cpl_image_get(filtered,i,ypos+(int)(ly)+1,&whatever),cpl_image_get(filtered,i,ypos+(int)(ly)+2,&whatever),cpl_image_get(filtered,i,ypos+(int)(ly)+3,&whatever),cpl_image_get(filtered,i,ypos+(int)(ly)+4,&whatever),cpl_image_get(filtered,i,ypos+(int)(ly)+5,&whatever),cpl_image_get(filtered,i,ypos+(int)(ly)+6,&whatever),cpl_image_get(filtered,i,ypos+(int)(ly)+7,&whatever),cpl_image_get(filtered,i,ypos+(int)(ly)-1,&whatever),cpl_image_get(filtered,i,ypos+(int)(ly)-2,&whatever),cpl_image_get(filtered,i,ypos+(int)(ly)-3,&whatever));
    }
    cpl_image_delete(unfiltered) ;
    cpl_image_delete(filtered) ;
    return extracted ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Make a confusion free line list for encoder values in header 
  @param    cat             the lines catalog
  @param    search_box_sz   the search box size
  @param    file            sci/cal file with header values
  @param    cfg             model config file
  @return   cust_lines      customised line list

returned ptr should be deallocated with:
crires_model_free2Darray(ret_ptr,6)

 */
/*----------------------------------------------------------------------------*/
double** crires_model_refining_custom_lines(
        const char          *   cat,
        int                     search_box_sz,
        const char          *   file,
        const char          *   cfg,
	const char          *   preset)
{
  double** cust_lines;
  double** all_lines;
  DOUBLE** znse_ref;
  cpl_table*   lines_tab ;
  double lambda, lambda_nm, inout, lam_min, lam_max;
  double pos1, lamstep, half_prof_wid, chipx_current;
  double p_dif, sinc_arg, blaze;
  double centroid, featflux;
  double order_fac;
  double inten, flux_scale, spotsize, prec, weight, peak;
  double prob, prob2,gauss1,gauss2,es_y_init;
  int morder,morder_it, morder_cnt,lines_number,ii,iii,jj,ll,photons;
  int gap_flag,flag;
  int buffer,all_tot;
  double* trace_flux=NULL; /*trace array */
  double** ccd=NULL; /*ccd array */
  struct crires crires_model;
  struct crires* p_crires; /*pointer to crires_model */
  p_crires=&crires_model;

  if ((cust_lines=crires_model_alloc2Darray(6,500))==NULL) {
    cpl_msg_error(__func__, "Cannot allocate 2D array");
    return NULL;
  }

  half_prof_wid=8.0;
  buffer=1+search_box_sz/4;
  prec=0.1;
  spotsize=0.15;
  flux_scale=2500.0; /*could set this according to exposure time*/

  if ((morder=crires_model_setup(file,cfg,-1,p_crires))
      ==-1) {
    cpl_msg_error(__func__, "Failing model setup") ;
    return NULL ;
  }
  p_crires->slitw*=1.2;

    if ((znse_ref=crires_model_alloc2Darray(8,7))==NULL) {
      cpl_msg_error(__func__, "Cannot allocate 2D array");
      return NULL;
    };
  /*open and read refractive index data file */
  if (crires_model_ref_ind_read(znse_ref,p_crires->temper)==-1) {
    cpl_msg_error(__func__, "error setting ref index at T= %lf", p_crires->temper);
    return NULL ;
  }
  if (crires_model_init(p_crires)!=0) {
    cpl_msg_error(__func__, "Failed to initialise model" );
    return NULL ;
  }

    if ((ccd=crires_model_alloc2Darray(ASIZE, BSIZE))==NULL) {
      cpl_msg_error(__func__, "Cannot allocate ccd array");
      return NULL;
    }

    for (ll=0;ll<500;ll+=1) {
      cust_lines[0][ll]=0.0;
      cust_lines[1][ll]=0.0;
      cust_lines[2][ll]=0.0;
      cust_lines[3][ll]=0.0;
      cust_lines[4][ll]=0.0;
      cust_lines[5][ll]=0.0;
    }

    if ((all_lines=crires_model_alloc2Darray(6,500))==NULL) {
      cpl_msg_error(__func__, "Cannot allocate 2D array");
      return NULL;
    }
    for (ll=0;ll<500;ll+=1) {
      all_lines[0][ll]=0.0;
      all_lines[1][ll]=0.0;
      all_lines[2][ll]=0.0;
      all_lines[3][ll]=0.0;
      all_lines[4][ll]=0.0;
      all_lines[5][ll]=0.0;
    }

    if (preset!=NULL) {
      morder=crires_model_preset_angles((char*)preset,p_crires);
    }

    p_crires->es_y=0.0;
    es_y_init=0.0;
    p_crires->es_x=0.0;
    morder_it=1;

    p_crires->slit_on=0;


    for (ii=0;ii<ASIZE;ii+=1) {
      for (jj=0;jj<BSIZE;jj+=1) {
	ccd[ii][jj]=0.0;
      }
    }
 
    /* Read the input FITS lines list */
    lines_tab = cpl_table_load(cat, 1, 0) ;
    lines_number = cpl_table_get_nrow(lines_tab) ;
    jj=0;
    /* Loop on the lines and  call the stis model routine */
    for (ll=0 ; ll<lines_number ; ll++) {
      lambda_nm=cpl_table_get_double(lines_tab, CRIRES_COL_WAVELENGTH, ll, NULL);
      inten = cpl_table_get_double(lines_tab, CRIRES_COL_EMISSION, ll, NULL);
      lambda=lambda_nm*1e-6; /*lambda in mm */
      /*Iterate over orders immediately above and below that chosen to
	allow for light from other orders arriving on detectors*/
      for (morder_cnt=morder-morder_it;
	   morder_cnt<=morder+morder_it;
	   morder_cnt+=1) {

	/*Blaze equations to compute wavelength ranges of orders*/
	inout=2.0*atan(0.5*p_crires->cam_dis/p_crires->fk);

	lam_min=(sin(p_crires->mug-(inout/2.0))+sin(p_crires->mug+(inout/2.0)))/
	  ((p_crires->sg)*((double)(morder_cnt)+2.0));
	lam_max=(sin(p_crires->mug-(inout/2.0))+sin(p_crires->mug+(inout/2.0)))/
	  ((p_crires->sg)*((double)(morder_cnt)-2.0));
	if (lambda>lam_min && lambda<lam_max) {
	  /*set centre of entrance slit as default*/
	  p_crires->es_x=0.0;
	  p_crires->es_y=es_y_init;
	  /*call model once for this entrance slit position. This will be
	    the result stored in the main line list. It is also used to
	    establish which lines fall on one of the detector arrays and
	    should be computed for the whole slit*/
	  crires_model_eval(lambda,morder_cnt,znse_ref, p_crires);
	  crires_model_detpix(p_crires);
	  if (morder_cnt==morder || strncmp(cfg,"search_cfg",10)==0) {
	    //printf("A %d %lf %lf %lf %lf %lf %d %d %d \n",morder_cnt,lambda*1000.0,p_crires->xpospix,p_crires->ypospix,p_crires->xdet,p_crires->ydet,p_crires->chippix[0],p_crires->chippix[1],p_crires->chippix[2]);
	    //if ( p_crires->onslit==1 &&
	    if (p_crires->chippix[0]>=0 && p_crires->chippix[0]<4) {
	      //printf("B %d %lf %lf %lf %lf %lf %d %d %d \n",morder_cnt,lambda*1000000.0,p_crires->xpospix,p_crires->ypospix,p_crires->xdet,p_crires->ydet,p_crires->chippix[0],p_crires->chippix[1],p_crires->chippix[2]);
	      all_lines[0][jj]=lambda_nm;
	      all_lines[1][jj]=(double)(morder_cnt);
	      all_lines[2][jj]=inten;
	      all_lines[3][jj]=(double)(p_crires->chippix[0]);
	      all_lines[4][jj]=p_crires->xpospix;
	      all_lines[5][jj]=p_crires->ypospix;
	      jj++;
	    }
	  }
	  /*if on one of the detector arrays, simulate photons from the
	    rest of the slit. NOTE 4 is allowed, as this
	    just means that the central ray falls on one of
	    the gaps between detectors*/
	  if (p_crires->chippix[0]>-1 && p_crires->chippix[0]<5) {// || p_crires->chippix[0]==7 || p_crires->chippix[0]==8) {
	    /*iterate over photons in spectral line*/

	    p_dif=(sin(p_crires->grat_alpha-(-p_crires->mug-p_crires->axis_mug))+
		   sin(p_crires->grat_beta-(-p_crires->mug-p_crires->axis_mug)));
	    if (p_crires->grat_alpha>p_crires->grat_beta) {
	      sinc_arg=(CPL_MATH_PI/(lambda*p_crires->sg))*
		(cos(p_crires->grat_alpha)/
		 cos(p_crires->grat_alpha-(-p_crires->mug-p_crires->axis_mug)))*p_dif;
	      blaze=((sin(sinc_arg))/sinc_arg)*((sin(sinc_arg))/sinc_arg);
	    }
	    else {
	      sinc_arg=(CPL_MATH_PI/(lambda*p_crires->sg))*(cos(p_crires->grat_beta)/cos(p_crires->grat_alpha-(-p_crires->mug-p_crires->axis_mug)))*p_dif;
	      blaze=(cos(p_crires->grat_beta)/cos(p_crires->grat_alpha))*(cos(p_crires->grat_beta)/cos(p_crires->grat_alpha))*((sin(sinc_arg))/sinc_arg)*((sin(sinc_arg))/sinc_arg);
	    }
	    for (photons=1; photons<(int)(inten*flux_scale); photons+=1) {
	      /*call random number generator twice and generate two
		gaussian distributions with mean=0, sigma=1.0 => FWHM=2.35*/
	      prob=rand()/2147483648.0;
	      prob2=rand()/2147483648.0;
	      gauss1=sqrt(-2.0*log(1-prob))*cos(2*CPL_MATH_PI*prob2);
	      gauss2=sqrt(-2.0*log(1-prob))*sin(2*CPL_MATH_PI*prob2);
	      flag=0;

	      p_crires->es_x=spotsize*gauss1;
	      p_crires->es_y=es_y_init+spotsize*gauss2;
	      /*Set flag if photon is not on slit (pos was drawn from a
		gaussian, it could be several sigma from the centre)*/
	      if (fabs(p_crires->es_x)>p_crires->es_s) {
		flag=1;
	      }
	      /*Set flag if photon is not on slit (pos was drawn from a
		gaussian, it could be several sigma from the centre)*/
	      if (fabs(2.0*p_crires->es_y)>p_crires->es_w) {
		flag=1;
	      }
	      p_crires->chippix[0]=-2;
	      /* Send to the model the wavelength, order, temperature,
		 sellmeier refractive index coefs and a pointer to the structure
		 Get back the chip no. and x,y position)*/
	      crires_model_eval(lambda,morder_cnt,znse_ref, p_crires);
	      crires_model_detpix(p_crires);
	      if (morder_cnt==morder) {
		order_fac=1.0;
	      }
	      else {
		order_fac=0.2;
	      }
		
	      if (p_crires->chippix[0]>-1 && p_crires->chippix[0]<4 && flag!=1) {
		/*ccd is an array with one location for each pixel in all four
		  chips plus the immaginary pixels that would fill the gaps
		  between them*/
		/*add one to value at x,y co-ord and shift according to which chip*/
		ccd[p_crires->chippix[0]*(int)(chipxpix+chipxspace)+p_crires->chippix[1]][p_crires->chippix[2]]+=order_fac;
	      }
	    }
	  }
	}
      }
    }
    all_tot=jj;
    if ((trace_flux=crires_model_alloc1Darray(ASIZE))==NULL){
      cpl_msg_error(__func__, "Cannot allocate array");
      return NULL;
    }
    p_crires->slitw/=1.2;
    cpl_table_delete(lines_tab);


    /*We now need to extract a 1D spectrum from the detector array that
      we just simulated. We do this by stepping along the locus of the
      central slit position on the array and interpolating the flux at
      the centre of each column (we use an extraction window that is 
      half_prof_wid pixels high)*/

    ll=0;
    /* We can turn the slit off here since all we are using the model for is getting the locus*/
    p_crires->slit_on=0;

    morder_it=0;

    p_crires->es_y=es_y_init;
    p_crires->es_x=0.0;
    for (morder_cnt=morder-morder_it;
	 morder_cnt<=morder+morder_it;
	 morder_cnt+=1) {
      /*Blaze equations to compute wavelength ranges of orders*/
      inout=2.0*atan(0.5*p_crires->cam_dis/p_crires->fk);

      lam_min=(sin(p_crires->mug-(inout/2.0))+sin(p_crires->mug+(inout/2.0)))/
	((p_crires->sg)*((double)(morder_cnt)+1.0));
      lam_max=(sin(p_crires->mug-(inout/2.0))+sin(p_crires->mug+(inout/2.0)))/
	((p_crires->sg)*((double)(morder_cnt)-1.0));
      /*get lambda in the middle*/
      lambda=(lam_max+lam_min)/2.0;
      /*guess step size*/
      lamstep=(lam_max-lam_min)/(40000.0);
      /*compute detector posn for lambda*/
      if (crires_model_init(p_crires)!=0) {
	cpl_msg_error(__func__, "Failed to initialise model" );
	cpl_free(trace_flux);
	if ((crires_model_free2Darray(znse_ref,8))!=0) {
	  cpl_msg_error(__func__, "Cannot free 2D array");
	  return NULL;
	}
	return NULL;
      }
      for (ii=0;ii<4096;ii++){
	trace_flux[ii]=0.0;
      }
      crires_model_eval(lambda,morder_cnt,znse_ref, p_crires);
      pos1=p_crires->xdet/p_crires->pix;
      /*compute detector posn for lambda+lamstep*/
      crires_model_eval(lambda+lamstep,morder_cnt,znse_ref, p_crires);
      /*refine lamstep*/
      lamstep=lamstep/(2.5*((p_crires->xdet/p_crires->pix)-pos1));
      chipx_current=0;
      /*step along locus*/
      for (lambda=lam_min; lambda<lam_max; lambda+=lamstep) {
	lambda_nm=lambda*1000000.0;
	crires_model_eval(lambda,morder_cnt,znse_ref, p_crires);
	crires_model_detpix(p_crires);
	if (p_crires->chippix[0]>-1 && p_crires->chippix[0]<4) {
	  /*reset chipx_current when starting new chip */
	  if (chipx_current>p_crires->chippix[1]+1) {
	    chipx_current=p_crires->chippix[1];
	  }
	  /*trace is 1d, 4096 long, i.e. chips are concatenated*/
	  trace_flux[p_crires->chippix[0]*(int)(chipxpix+chipxspace)+
		     p_crires->chippix[1]]=0.0;
	  for (ii=-half_prof_wid;ii<half_prof_wid+1;ii+=1) {
	    trace_flux[p_crires->chippix[0]*(int)(chipxpix+chipxspace)+
		       p_crires->chippix[1]]+=
	      ccd[p_crires->chippix[0]*(int)(chipxpix+chipxspace)+
		  p_crires->chippix[1]][p_crires->chippix[2]+ii];
	  }
	  chipx_current=p_crires->chippix[1]+1;
	}
      }

      /*We now need to check if the extracted centroids (from the
	trace_flux spectrum) match the expected centroids (in the
	all_lines structure)*/

      for (jj=0 ; jj<all_tot ; jj++) {
	if (morder_cnt==(int)(all_lines[1][jj])) {
	  lambda_nm=all_lines[0][jj];
	  centroid=0.0;
	  featflux=0.0;
	  peak=0.0;
	  for (ii=search_box_sz+buffer; ii>=-(search_box_sz+buffer);ii-=1) {
	    iii=(int)(all_lines[4][jj]+0.5)+ii;
	    if (iii>0 && iii<1024) {
	      if (trace_flux[iii+(int)(all_lines[3][jj]*1024.0)]>peak) {
		peak=trace_flux[iii+(int)(all_lines[3][jj]*1024.0)];
	      }
	    }
	  }
	  gap_flag=0;
	  for (ii=0; ii<=search_box_sz+buffer;ii+=1) {
	    iii=(int)(all_lines[4][jj]+0.5)+ii;
	    if (iii>0 && iii<1024) {
	      if (gap_flag==0 && trace_flux[iii+(int)(all_lines[3][jj]*1024.0)]<0.1*peak) {
		peak=trace_flux[iii+(int)(all_lines[3][jj]*1024.0)];
		gap_flag=1;
	      }
	      if (gap_flag==1 && trace_flux[iii+(int)(all_lines[3][jj]*1024.0)]>peak) {
		gap_flag=2;
	      }
	      if (gap_flag==1 && trace_flux[iii+(int)(all_lines[3][jj]*1024.0)]<peak) {
		peak=trace_flux[iii+(int)(all_lines[3][jj]*1024.0)];
	      }
	      if (gap_flag!=2) {
		weight=(float)(ii);
	      }
	      else {
		weight=(float)(ii*(search_box_sz+buffer-abs(ii)))/(float)(search_box_sz+buffer);
	      }
	      centroid+=trace_flux[iii+(int)(all_lines[3][jj]*1024.0)]*weight;
	      featflux+=trace_flux[iii+(int)(all_lines[3][jj]*1024.0)];
	      //printf("%d %d %d %lf %lf \n",jj, ii, iii, trace_flux[iii+(int)(all_lines[3][jj]*1024.0)],weight);
	    }
	  }
	  peak=0.0;
	  for (ii=search_box_sz+buffer; ii>=-(search_box_sz+buffer);ii-=1) {
	    iii=(int)(all_lines[4][jj]+0.5)+ii;
	    if (iii>0 && iii<1024) {
	      if (trace_flux[iii+(int)(all_lines[3][jj]*1024.0)]>peak) {
		peak=trace_flux[iii+(int)(all_lines[3][jj]*1024.0)];
	      }
	    }
	  }
	  gap_flag=0;
	  for (ii=-1; ii>=-(search_box_sz+buffer);ii-=1) {
	    iii=(int)(all_lines[4][jj]+0.5)+ii;
	    if (iii>0 && iii<1024) {
	      if (gap_flag==0 && trace_flux[iii+(int)(all_lines[3][jj]*1024.0)]<0.1*peak) {
		peak=trace_flux[iii+(int)(all_lines[3][jj]*1024.0)];
		gap_flag=1;
	      }
	      if (gap_flag==1 && trace_flux[iii+(int)(all_lines[3][jj]*1024.0)]>peak) {
		gap_flag=2;
	      }
	      if (gap_flag==1 && trace_flux[iii+(int)(all_lines[3][jj]*1024.0)]<peak) {
		peak=trace_flux[iii+(int)(all_lines[3][jj]*1024.0)];
	      }
	      if (gap_flag!=2) {
		weight=(float)(ii);
	      }
	      else {
		weight=(float)(ii*(search_box_sz+buffer-abs(ii)))/(float)(search_box_sz+buffer);
	      }
	      centroid+=trace_flux[iii+(int)(all_lines[3][jj]*1024.0)]*weight;
	      featflux+=trace_flux[iii+(int)(all_lines[3][jj]*1024.0)];
	      //printf("%d %d %d %lf %lf \n",jj, ii, iii, trace_flux[iii+(int)(all_lines[3][jj]*1024.0)],weight);
	    }
	  }
	  if (featflux>0.0) {
	    centroid=centroid/featflux;
	    centroid+=(int)(all_lines[4][jj]+0.5);
	  }
	  /* Now select only those whose centroids are not distorted by more than prec*/
	  if (fabs(all_lines[4][jj]-centroid-0.5)<prec) {
	    cust_lines[2][ll]=all_lines[2][jj];
	    //	    printf("selected %d %lf %d %d %lf %lf %lf %lf %lf \n",morder_cnt, all_lines[0][jj], (int)(all_lines[1][jj]), (int)(all_lines[3][jj]), all_lines[4][jj],centroid+0.5,all_lines[4][jj]-centroid-0.5, all_lines[5][jj], all_lines[2][jj]);
	  }
	  /*unselected lines get a -ve intensity but are kept in the structure*/
	  else {
	    cust_lines[2][ll]=-all_lines[2][jj];
	    //	    printf("NOT selected %d %lf %d %d %lf %lf %lf %lf %lf \n",morder_cnt, all_lines[0][jj], (int)(all_lines[1][jj]), (int)(all_lines[3][jj]), all_lines[4][jj],centroid+0.5,all_lines[4][jj]-centroid-0.5, all_lines[5][jj], all_lines[2][jj]);
	  }
	  cust_lines[0][ll]=all_lines[0][jj];
	  cust_lines[1][ll]=all_lines[1][jj];
	  cust_lines[3][ll]=all_lines[3][jj];
	  cust_lines[4][ll]=all_lines[4][jj];
	  cust_lines[5][ll]=all_lines[5][jj];
	  ll++;
	}
      }
    }
    cpl_free(trace_flux);

    if ((crires_model_free2Darray(ccd,ASIZE))!=0) {
      cpl_msg_error(__func__, "Cannot free 2D array");
      return NULL;
    }
    if ((crires_model_free2Darray(all_lines,6))!=0) {
      cpl_msg_error(__func__, "Cannot free 2D array");
      return NULL;
    }

  if ((crires_model_free2Darray(znse_ref,8))!=0) {
    cpl_msg_error(__func__, "Cannot free 2D array");
    return NULL;
  }
  return cust_lines;
}

/* Updates the cust_lines list to have the correct predicted
   detector x,y for the current fibre*/
double** crires_model_refining_fibrefix(
	double**                temp_lines,
        const char*             file,
        const char*             cfg,
	int                     ph)
{
  DOUBLE** znse_ref;
  int morder,ii;
  struct crires crires_model;
  struct crires* p_crires; /*pointer to crires_model */
  p_crires=&crires_model;

  if ((morder=crires_model_setup(file,cfg,-1,p_crires))==-1) {
    cpl_msg_error(__func__, "Failing model setup") ;
    return NULL ;
  }
  if ((znse_ref=crires_model_alloc2Darray(8,7))==NULL) {
    cpl_msg_error(__func__, "Cannot allocate 2D array");
    return NULL;
  };
  /*open and read refractive index data file */
  if (crires_model_ref_ind_read(znse_ref,p_crires->temper)==-1) {
    cpl_msg_error(__func__, "error setting ref index at T= %lf", p_crires->temper);
    if ((crires_model_free2Darray(znse_ref,8))!=0) {
      cpl_msg_error(__func__, "Cannot free 2D array");
      return NULL;
    }
    return NULL ;
  }
  p_crires->es_x=p_crires->slit[ph];
  p_crires->slit_on=0;

  if (crires_model_init(p_crires)!=0) {
    cpl_msg_error(__func__, "Failed to initialise model" );
    if ((crires_model_free2Darray(znse_ref,8))!=0) {
      cpl_msg_error(__func__, "Cannot free 2D array");
      return NULL;
    }
    return NULL ;
  }

  for (ii=0;ii<50;ii++) {
    if (temp_lines[0][ii]) {
      crires_model_eval(temp_lines[0][ii]/1000000.0,(int)(temp_lines[1][ii]),znse_ref,p_crires);
      crires_model_detpix(p_crires);
      temp_lines[4][ii]=p_crires->xpospix;
      temp_lines[5][ii]=p_crires->ypospix;
    }
  }

  if ((crires_model_free2Darray(znse_ref,8))!=0) {
    cpl_msg_error(__func__, "Cannot free 2D array");
    return NULL;
  }
  return temp_lines;

}
/*----------------------------------------------------------------------------*/
/**
  @brief    Match the detected lines with the catalog
  @param    bright          the bright lines
  @param    line_list      the customised line list
  @param    search_box_sz   the search box size
  @param    min_matches_nb  the monimum number of matches
  @param    file            sci/cal file with header values
  @return   temp_coord      2D array containing the information required to
                            build msp_coord or NULL in error case

 */
/*----------------------------------------------------------------------------*/
double** crires_model_refining_match(
        const cpl_vector**      bright,
        const cpl_vector    **  traces,
        double**                line_list,
        int                     search_box_sz,
        int                     min_matches_nb,
        const char          *   file)
{
    double** temp_coord;
    cpl_vector**   wl;
    cpl_vector**   ord;
    int ii,jj,kk,gone,iigone,jjgone,line_no,match_cnt,detected_size_tot;
    int use_det,chip,grat_enc,prism_enc;
    double offset, offset_best,diff, meandiff, meandiff_min, meandiff_max, score, slitw;
    int detected_size[4];
    double alldiffs[500];
    cpl_propertylist    *   plist ;

    /* Test entries */
    if (bright == NULL) return NULL ;
    if (line_list == NULL) return NULL ;

    if ((plist=cpl_propertylist_load(file, 0)) == NULL) {
      cpl_msg_error(__func__, "getting header from file %s", file);
      cpl_propertylist_delete(plist) ;
      return NULL ;
    }
    grat_enc=crires_pfits_get_grat_enc(plist);
    prism_enc=crires_pfits_get_prism_enc(plist);
    slitw=crires_pfits_get_slit2_pos(plist);
    cpl_propertylist_delete(plist) ;
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
      cpl_msg_error(__func__, "Cannot read keywords from the header") ;
      return NULL ;  
    }

    if ((wl=cpl_malloc(CRIRES_NB_DETECTORS*sizeof(cpl_vector*)))==NULL) {
      cpl_msg_error(__func__, "Cannot allocate wl array");
      return NULL;
    }
    if ((ord=cpl_malloc(CRIRES_NB_DETECTORS*sizeof(cpl_vector*)))==NULL) {
      cpl_msg_error(__func__, "Cannot allocate ord array");
      return NULL;
    }

    detected_size_tot=0;
    for (chip=0;chip<CRIRES_NB_DETECTORS;chip++) {
      detected_size_tot+=cpl_vector_get_size(bright[chip]);
      detected_size[chip]=cpl_vector_get_size(bright[chip]);
      wl[chip]=cpl_vector_new(cpl_vector_get_size(bright[chip]));
      cpl_vector_fill(wl[chip], 0.0) ;
      ord[chip]=cpl_vector_new(cpl_vector_get_size(bright[chip]));
      cpl_vector_fill(ord[chip], 0.0) ;
    }

    line_no=0;
    offset_best=-9999.0;
    meandiff_min=1000.0;
    meandiff_max=0.0;
    for (offset=-3.0*search_box_sz; offset<=3.0*search_box_sz; offset+=0.1) {
      line_no=0;
      for (ii=0; ii<detected_size_tot;ii++){
	alldiffs[ii]=-1.0;
      }
      jj=0;
      while (line_list[0][line_no]!=0.0) {
	for (chip=0;chip<CRIRES_NB_DETECTORS;chip++) {
	  if ((int)(line_list[3][line_no])==chip) {
	    gone=0;
	    jjgone=0;
	    if (detected_size[chip]<=5) {
	      use_det=detected_size[chip];
	    }
	    else if (detected_size[chip]>=8) {
	      use_det=detected_size[chip]-3;
	    }
	    else {
	      use_det=5;
	    }
	    for (ii=0;ii<use_det;ii+=1) {
	      diff=fabs(cpl_vector_get(bright[chip],ii)-line_list[4][line_no]+offset);
	      if (diff<search_box_sz/4.0) {
		/*now compute a +ve score instead of diff*/
		score=(search_box_sz/4.0-diff);
		/* very small residuals only reflect chance alignment, make them all 0.2 */
		  //		if (diff<0.2) diff=0.2;
		/*check that this list feature has not already been matched*/
		if (gone==0) {
		  alldiffs[jj]=score;
		  jjgone=jj;
		  jj++;
		}
		gone++;
	      }
	    }
	    /*(Don't) remove doubles*/
/* 	    if (gone>1) { */
/* 	      alldiffs[jjgone]=-1.0; */
/* 	    } */
	  }
	}
	line_no+=1;
      }
      meandiff=0.0;
      match_cnt=0;
      for (ii=0;ii<detected_size_tot;ii++) {
	if (alldiffs[ii]>=0.0) {
	  /*meandiff is now really total score*/
	  meandiff+=alldiffs[ii];//*(float)((detected_size/3)/((detected_size/3)-ii));
	  match_cnt+=1;
	  //printf("%lf %d %lf %lf %lf %d %d\n", offset, ii,alldiffs[ii],meandiff,meandiff/((float)(match_cnt)*(float)(match_cnt)),match_cnt,detected_size_tot);
	}
      }
      if (meandiff>meandiff_max) {
	offset_best=offset;
	meandiff_max=meandiff;
      }
    }
    if (offset_best==-9999.0){
      cpl_msg_warning(__func__, "Couldn't x-correlate");
    }
    else {
      cpl_msg_info(__func__, "Linear x-corr offset: %lf", offset_best) ;
    }

    /*It should never be possible that a detected feature is matched
      to >1 list feature since the separation between listed features
      should be set larger than search_box_sz*/
    line_no=0;
    match_cnt=0;
    while (line_list[0][line_no]!=0.0) {
      if (line_list[2][line_no]>0.0) {
	gone=0;
	for (chip=0;chip<CRIRES_NB_DETECTORS;chip++) {
	  if (line_list[3][line_no]==chip) {
	    iigone=0;
	    if (detected_size[chip]<=5) {
	      use_det=detected_size[chip];
	    }
	    else if (detected_size[chip]>=8) {
	      use_det=detected_size[chip]-3;
	    }
	    else {
	      use_det=5;
	    }
	    for (ii=0;ii<use_det;ii+=1) {
	      diff=fabs(cpl_vector_get(bright[chip],ii)-line_list[4][line_no]+offset_best);
	      if (diff<search_box_sz/2.0) {
	      //printf("%lf %lf %lf %lf %lf\n",diff,line_list[0][line_no],line_list[3][line_no],cpl_vector_get(bright[chip],ii),line_list[4][line_no]);
	  /*check that this list feature has not already been matched*/
		if (gone==0) {
		  cpl_vector_set(wl[chip],ii,line_list[0][line_no]);
		  cpl_vector_set(ord[chip],ii,line_list[1][line_no]);
		  iigone=ii;
		  match_cnt+=1;
		}
		gone++;
	      }
	    }
	  /*remove doubles*/
	    if (gone>1) {
	      cpl_vector_set(wl[chip],iigone,0.0);
	      cpl_vector_set(ord[chip],iigone,0.0);
	      match_cnt-=1;
	    }
	  }
	}
      }
      line_no+=1;
    }

    if (match_cnt<min_matches_nb) {
       cpl_msg_warning(__func__, "Not enough matches identified");
       for (ii=0 ; ii<CRIRES_NB_DETECTORS ; ii++) cpl_vector_delete(wl[ii]) ;
       cpl_free(wl) ;
       for (ii=0 ; ii<CRIRES_NB_DETECTORS ; ii++) cpl_vector_delete(ord[ii]) ;
       cpl_free(ord) ;
       return NULL;
    }

  if ((temp_coord=crires_model_alloc2Darray(11,100))==NULL) {
    cpl_msg_error(__func__, "Cannot allocate 2D array");
    for (ii=0 ; ii<CRIRES_NB_DETECTORS ; ii++) cpl_vector_delete(wl[ii]) ;
    cpl_free(wl) ;
    for (ii=0 ; ii<CRIRES_NB_DETECTORS ; ii++) cpl_vector_delete(ord[ii]) ;
    cpl_free(ord) ;
    return NULL;
  }
  for (kk=0;kk<100;kk+=1) {
    temp_coord[0][kk]=0.0;
    temp_coord[1][kk]=0.0;
    temp_coord[2][kk]=0.0;
    temp_coord[3][kk]=0.0;
    temp_coord[4][kk]=0.0;
    temp_coord[5][kk]=0.0;
    temp_coord[6][kk]=0.0;
    temp_coord[7][kk]=0.0;
    temp_coord[8][kk]=0.0;
    temp_coord[9][kk]=0.0;
    temp_coord[10][kk]=0.0;
  }

  kk=0;
  for (chip=0;chip<CRIRES_NB_DETECTORS;chip+=1){
    for (jj=0;jj<cpl_vector_get_size(bright[chip]);jj+=1){
      if (cpl_vector_get(wl[chip],jj)>0.0) {
	temp_coord[0][kk]=0.000001*cpl_vector_get(wl[chip],jj);/*wave*/
	temp_coord[1][kk]=cpl_vector_get(bright[chip],jj);/*x*/
	temp_coord[2][kk]=100.0; /*flux*//*no flux weighting in the optimisation*/
	temp_coord[3][kk]=(float)(chip);/*chip*/
	temp_coord[4][kk]=cpl_vector_get(ord[chip],jj);/*mode*/
	temp_coord[5][kk]=cpl_vector_get(traces[chip],(int)(temp_coord[1][kk]));/*y*/
	temp_coord[6][kk]=(float)(prism_enc);/*penc*/
	temp_coord[7][kk]=(float)(grat_enc);/*genc*/
	temp_coord[8][kk]=0.0; /*slit_pos*/
	temp_coord[9][kk]=(float)(jj);/*counter*/
	temp_coord[10][kk]=slitw;/*slit width from header*/
	kk++;
      }
    }
  }
  temp_coord[0][kk]=0.0;

  for (ii=0 ; ii<CRIRES_NB_DETECTORS ; ii++) cpl_vector_delete(wl[ii]) ;
  cpl_free(wl) ;
  for (ii=0 ; ii<CRIRES_NB_DETECTORS ; ii++) cpl_vector_delete(ord[ii]) ;
  cpl_free(ord) ;

  return temp_coord;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Use annealing to refine the model
  @param    msp_coord      Measured co-ordinates
  @param    msp_coord_size Number of measured co-ordinates
  @param    traces         the CRIRES_NB_DETECTORS spectra traces
  @param    file           Input file to get infos from (order ...)
  @param    cfg            Input model configuration file
  @return   The table containing the new model parameters

  wl[j] and pos[j] must have the same size.
  They can be both NULL if no lines were detected on the chip j+1.

 */
/*----------------------------------------------------------------------------*/
cpl_table * crires_model_refining_anneal(coord* msp_coord,
					 int msp_coord_size,
					 double y_pos_mm,
					 const char          *   file,
					 const char          *   cfg)
{
    int adim,morder;
    cpl_propertylist    *   plist ;
    cpl_table   *   tab ;
    //int             grat_enc, prism_enc ;
    int             ii, jj, s5_size;
    int 	  ipc1, igc1, isg, ifk, icx[4];
    int onchip[4]={0,0,0,0};
    ann_all_par all_par[300], *p_all_par;
    double abest[300];
    double amin[300];
    double amax[300];
    int aname[300];
    DOUBLE ** znse_ref;
    struct crires crires_model;
    struct crires *p_crires; /*pointer to crires_model */
    /* Test entries */
    if (file == NULL) return NULL ;
    if (cfg == NULL) return NULL ;

    p_crires=&crires_model;
    p_all_par=&all_par[0];

    if ((adim=crires_model_readfits(abest,
				    amin,
				    amax,
				    aname,
				    cfg,
				    p_crires,
				    p_all_par))==-1) {
      cpl_msg_error(__func__, "Failed to read config FITS file" );
      return NULL;
    }

    if ((morder=crires_model_setup(file,cfg,-1,p_crires))==-1) {
      cpl_msg_error(__func__, "Failing model setup") ;
      return NULL ;
    }

    /*still need to get the encoder values (setup gets them but
      doesn't pass them back as usually only mup1 and mug are needed */
    /* Read the header informations */
    if ((plist=cpl_propertylist_load(file, 0)) == NULL) {
      cpl_msg_error(__func__, "getting header from file %s", file);
      cpl_propertylist_delete(plist) ;
      return NULL ;
    }
    cpl_propertylist_delete(plist) ;
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
      cpl_msg_error(__func__, "Cannot read keywords from the header") ;
      return NULL ;  
    }

    if ((znse_ref=crires_model_alloc2Darray(8,7))==NULL) {
      cpl_msg_error(__func__, "Cannot allocate 2D array");
      return NULL;
    }
    /*open and read refractive index data file */
    if (crires_model_ref_ind_read(znse_ref,p_crires->temper)==-1) {
      cpl_msg_error(__func__, "error setting ref index at T= %lf", p_crires->temper);
      return NULL ;
    }

    /*Make slit[0] equal to the entrance slit posn
      in mm determined earlier with crires_model_profile and stored at the
      end of the traces vector and then, below, set all entries to be at the
      middle slit position*/
    p_crires->slit[0]=y_pos_mm;

    if (adim==0) {
      /*input file has no parameters flagged for optimisation, so*/
      /*manually modify parameters for optimisation*/
      /*first discover their indices*/
      ipc1=igc1=isg=ifk=icx[0]=icx[1]=icx[2]=icx[3]=0;
      for (ii=0;ii<300;ii++) {
	if (strncmp((p_all_par+ii)->name,"penc_co1",8)==0) {
	  ipc1=ii;
	}
	if (strncmp((p_all_par+ii)->name,"genc_co1",8)==0) {
	  igc1=ii;
	}
	if (strncmp((p_all_par+ii)->name,"sg",2)==0) {
	  isg=ii;
	}
	if (strncmp((p_all_par+ii)->name,"fk",2)==0) {
	  ifk=ii;
	}
	if (strncmp((p_all_par+ii)->name,"chipx[0]",8)==0) {
	  icx[0]=ii;
	}
	if (strncmp((p_all_par+ii)->name,"chipx[1]",8)==0) {
	  icx[1]=ii;
	}
	if (strncmp((p_all_par+ii)->name,"chipx[2]",8)==0) {
	  icx[2]=ii;
	}
	if (strncmp((p_all_par+ii)->name,"chipx[3]",8)==0) {
	  icx[3]=ii;
	}
      }

      /*Check how many lines in slit[5] (which is usually well populated)*/
      s5_size=0;
      for (ii=0;ii<msp_coord_size;ii++) {
	/*slit_pos is 0 iff in crires_model_fix recipe*/
	if (msp_coord[ii].slit_pos==5 || msp_coord[ii].slit_pos==0) {
	  s5_size+=1;
	  onchip[msp_coord[ii].chip]+=1;
	}
      }

      /*Must be at least one line, so set the genc_co1 params*/
      (p_all_par+igc1)->flag=1;
      abest[0]=p_crires->genc_co1/DEG2RAD;
      aname[0]=igc1;
      /*If not more than 2 lines then genc_co1 needs to have a large range*/
      if (s5_size<=2) {
	(p_all_par+igc1)->min=p_crires->genc_co1-0.1*DEG2RAD;
	(p_all_par+igc1)->max=p_crires->genc_co1+0.1*DEG2RAD;
	amin[0]=p_crires->genc_co1/DEG2RAD-0.1;
	amax[0]=p_crires->genc_co1/DEG2RAD+0.1;
	adim=1;
      }
      /*If 3 or more then genc_co1 can be restricted to a realistic range
	(realistic encoder uncertainty). Large range for fk and sg can
	accomodate bigger shifts*/
      else if (s5_size>=3) {
	(p_all_par+igc1)->min=p_crires->genc_co1-fabs(100.0*p_crires->genc_co2);
	(p_all_par+igc1)->max=p_crires->genc_co1+fabs(100.0*p_crires->genc_co2);
	amin[0]=(p_crires->genc_co1-fabs(100.0*p_crires->genc_co2))/DEG2RAD;
	amax[0]=(p_crires->genc_co1+fabs(100.0*p_crires->genc_co2))/DEG2RAD;
	(p_all_par+ifk)->min=p_crires->fk-10.0;
	(p_all_par+ifk)->max=p_crires->fk+10.0;
	(p_all_par+ifk)->flag=1;
	abest[1]=p_crires->fk;
	amin[1]=p_crires->fk-10.0;
	amax[1]=p_crires->fk+10.0;
	aname[1]=ifk;
	(p_all_par+isg)->min=p_crires->sg-0.2;
	(p_all_par+isg)->max=p_crires->sg+0.2;
	(p_all_par+isg)->flag=1;
	abest[2]=p_crires->sg;
	amin[2]=p_crires->sg-0.2;
	amax[2]=p_crires->sg+0.2;
	aname[2]=isg;
	adim=3;
      }
      /*Always set the range for penc_co1 to be realistic (encoder
	uncertainty), otherwise solutions that don't pass the int slit
	may occur*/
      (p_all_par+ipc1)->min=p_crires->penc_co1-(100.0*p_crires->penc_co2);
      (p_all_par+ipc1)->max=p_crires->penc_co1+(100.0*p_crires->penc_co2);
      (p_all_par+ipc1)->flag=1;
      /*If 2 then penc_co1 is second in the list after genc_co1*/
      if (s5_size==2) {
	abest[1]=p_crires->penc_co1/DEG2RAD;
	amin[1]=(p_crires->penc_co1/DEG2RAD-100.0*p_crires->penc_co2)/DEG2RAD;
	amax[1]=(p_crires->penc_co1/DEG2RAD+100.0*p_crires->penc_co2)/DEG2RAD;
	aname[1]=ipc1;
	adim=2;
      }
      /*If >3 then penc_co1 is 4th in the list after genc_co1,
	fk and sg*/
      else if (s5_size>3) {
	abest[3]=p_crires->penc_co1/DEG2RAD;
	amin[3]=(p_crires->penc_co1-100.0*p_crires->penc_co2)/DEG2RAD;
	amax[3]=(p_crires->penc_co1+100.0*p_crires->penc_co2)/DEG2RAD;
	aname[3]=ipc1;
	adim=4;
      }
      /*If there are a lot of matched features then anneal chipx as well*/
      if (s5_size>=5) {
	jj=0;
	for (ii=0 ; ii<CRIRES_NB_DETECTORS ; ii++) {
	  if (onchip[ii]>1) {
	    (p_all_par+icx[ii])->min=p_crires->chipx[ii]-0.5;
	    (p_all_par+icx[ii])->max=p_crires->chipx[ii]+0.5;
	    (p_all_par+icx[ii])->flag=1;
	    abest[4+jj]=p_crires->chipx[ii];
	    amin[4+jj]=p_crires->chipx[ii]-0.5;
	    amax[4+jj]=p_crires->chipx[ii]+0.5;
	    aname[4+jj]=icx[ii];
	    jj+=1;
	  }
	}
	adim=4+jj;
      }
    }

    p_crires->slit_on=0;
    for (ii=0;ii<msp_coord_size;ii++) {
      p_crires->es_x=p_crires->slit[msp_coord[ii].slit_pos];
      morder=msp_coord[ii].mode;
      p_crires->slitw=msp_coord[ii].slitw;
      p_crires->mup1=p_crires->penc_co1+
	p_crires->penc_co2*(msp_coord[ii].penc-36000)+
	p_crires->penc_co3*(msp_coord[ii].penc-36000)*(msp_coord[ii].penc-36000);
      p_crires->mug=p_crires->genc_co1+
	p_crires->genc_co2*(float)(msp_coord[ii].genc-200000)+
	p_crires->genc_co3*(float)(msp_coord[ii].genc-200000)*(float)(msp_coord[ii].genc-200000);
      if (crires_model_init(p_crires)!=0) {
	cpl_msg_error(__func__, "Failed to initialise model" );
	crires_model_free2Darray(znse_ref,8);
	return NULL;
      }
      /*Call the model kernel with the initial values*/
      crires_model_eval(msp_coord[ii].wave,morder,znse_ref,p_crires);
      crires_model_detpix(p_crires);
      printf("check %d %lf %d %lf %d %d %lf %lf %lf %lf %lf %lf %d %d %d %lf\n",
	     ii,msp_coord[ii].wave*1000000.0,msp_coord[ii].slit_pos,p_crires->es_x,p_crires->chippix[0],
	     msp_coord[ii].chip,p_crires->xpospix, p_crires->ypospix,
	     msp_coord[ii].x,msp_coord[ii].y,
	     msp_coord[ii].x-p_crires->xpospix,
	     msp_coord[ii].y-p_crires->ypospix,
	     msp_coord[ii].mode, msp_coord[ii].penc, msp_coord[ii].genc,
             msp_coord[ii].slitw);
    }

    /*Call the annealing*/
    if ((tab = crires_model_qanneal_comp(p_all_par,
					 adim,
					 abest,
					 amin,
					 amax,
					 aname,
					 p_crires,
					 msp_coord_size,
					 msp_coord,
					 znse_ref)) == NULL) {
      cpl_msg_error(__func__, "error calling crires_model_qanneal_comp");
      crires_model_free2Darray(znse_ref,8);
      return NULL ;
    }
    if ((crires_model_free2Darray(znse_ref,8))!=0) {
      cpl_msg_error(__func__, "Cannot free 2D array");
      return NULL;
    }

    /* Return */
    return tab ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Use annealing to fix the model
  @param    pos_x   X positions provided by the user
  @param    pos_y   Y positions provided by the user
  @param    chip    1 to CRIRES_NB_DETECTORS
  @param    wls     wave length provided by the user
  @param    file    Input file to get infos from (order ...)
  @param    cfg     Input model configuration file
  @return   The table containing the new model parameters
 */
/*----------------------------------------------------------------------------*/
cpl_table * crires_model_refining_fix(
        const cpl_vector    *   pos_x,
        const cpl_vector    *   pos_y,
        const int           *   chip,
        const cpl_vector    *   wls,
        const char          *   file,
        const char          *   cfg)
{
  cpl_table       *   new_conf ;
  int ii,j,morder,grat_enc,prism_enc,msp_coord_size;
  cpl_vector      **  traces ;
  /*Assume not more than 20 wavelengths supplied*/
  coord msp_coord[20];
  cpl_propertylist    *   plist ;
  double y_pos_pix=-1.0;
  double pos_y_min=520.0;
  double pos_y_max=-1.0;
  double y_pred, slitw;

  msp_coord_size=cpl_vector_get_size(pos_x);

  for (ii=0;ii<msp_coord_size;ii++) {
    if (cpl_vector_get(pos_y,ii)<pos_y_min) {
      pos_y_min=cpl_vector_get(pos_y,ii);
    }
    if (cpl_vector_get(pos_y,ii)>pos_y_max) {
      pos_y_max=cpl_vector_get(pos_y,ii);
    }
    if (chip[ii]-1==2) {
      y_pos_pix=cpl_vector_get(pos_y,ii);
    }
  }
  if (y_pos_pix<0.0) {
    cpl_msg_error(__func__, "No data points on chip 3. Please provide in the CRIRES_CAPLRO_MODEL_FIX_TAB either a line centroid for chip 3, or the row number of the fibre trace on chip 3 with the column centroid (pos_x) as -1");
    return NULL;
  }
    if (pos_y_max-pos_y_min>40.0) {
      cpl_msg_error(__func__, "The centroids do not all come from the same fibre/entrance slit position");
    return NULL;
  }

  if ((plist=cpl_propertylist_load(file, 0)) == NULL) {
    cpl_msg_error(__func__, "getting header from file %s", file);
    cpl_propertylist_delete(plist) ;
    return NULL ;
  }
  morder=crires_pfits_get_order(plist);
  grat_enc=crires_pfits_get_grat_enc(plist);
  prism_enc=crires_pfits_get_prism_enc(plist);
  slitw=crires_pfits_get_slit2_pos(plist);
  cpl_propertylist_delete(plist) ;
  if (cpl_error_get_code() != CPL_ERROR_NONE) {
    cpl_msg_error(__func__, "Cannot read keywords from the header") ;
    return NULL ;  
  }  

  /* Get the shape of the spectrum at the detector y posn specified */
  cpl_msg_info(__func__, "Get the shape of the spectra") ;
  if ((traces = crires_model_profile(file, cfg, -1,y_pos_pix)) == NULL) {
    cpl_msg_error(__func__, "Cannot get the spectrum trace") ;
    return NULL ;
  }

  for (ii=0;ii<msp_coord_size;ii++) {
    if (cpl_vector_get(wls,ii)>0.0) {
      y_pred=cpl_vector_get(traces[chip[ii]-1],(int)(cpl_vector_get(pos_x,ii)));
      if (fabs(y_pred-cpl_vector_get(pos_y,ii))>2.0) {
	cpl_msg_warning(__func__, "large discrepancy between expected y (row) centroid and given value for point %d", ii+1);
      }
      msp_coord[ii].wave=cpl_vector_get(wls,ii)/1000000.0;
      msp_coord[ii].x=cpl_vector_get(pos_x,ii);
      msp_coord[ii].flux=100.0;
      msp_coord[ii].chip=chip[ii]-1;
      msp_coord[ii].mode=morder;
      msp_coord[ii].y=y_pred;
      msp_coord[ii].penc=prism_enc;
      msp_coord[ii].genc=grat_enc;
      msp_coord[ii].slit_pos=0;
      msp_coord[ii].counter=ii;
      msp_coord[ii].slitw=slitw;
    }
  }

  /*traces[2],1024 holds the y posn in mm*/
  if ((new_conf = crires_model_refining_anneal(msp_coord,
					       msp_coord_size,
					       cpl_vector_get(traces[2],1024),
					       file,
					       cfg)) == NULL) {
    cpl_msg_error(__func__, "Cannot compute the new configuration file");
    for (j=0 ;j<CRIRES_NB_DETECTORS; j++) 
      if (traces[j] != NULL) cpl_vector_delete(traces[j]) ;
      cpl_free(traces) ;
    return NULL; 
  }

  for (j=0 ; j<CRIRES_NB_DETECTORS; j++) {
    if (traces[j] != NULL) cpl_vector_delete(traces[j]);
  }
  cpl_free(traces) ;

  return new_conf ;
}

/**@}*/
