/* $Id: crires_parameter.c,v 1.13 2012-09-20 12:06:34 yjung Exp $
 *
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2012-09-20 12:06:34 $
 * $Revision: 1.13 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <math.h>
#include <float.h>
#include <assert.h>
#include <cpl.h>

#include "irplib_plugin.h"

#include "crires_parameter.h"

/*-----------------------------------------------------------------------------
                                   Define
 -----------------------------------------------------------------------------*/

#define CRIRES_PARAMETER_SET(MASK, VARNAME, TYPE, MAN, DEFAULT, SHORT)         \
if (bitmask & MASK) {                                                          \
    char * paramname = cpl_sprintf(PACKAGE ".%s.%s", recipe, VARNAME);         \
    assert( paramname != NULL );                                               \
                                                                               \
    bitmask ^= MASK; /* Reset bit. At the end bitmask must be zero */          \
                                                                               \
    p = cpl_parameter_new_value(paramname, TYPE, MAN, context, DEFAULT);       \
                                                                               \
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, SHORT);                 \
                                                                               \
    cpl_free(paramname);                                                       \
    /* Environment variable names may not use the .-character. */              \
    paramname = cpl_sprintf("%s_%s", recipe, VARNAME);                         \
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_ENV, paramname);             \
    cpl_free(paramname);                                                       \
                                                                               \
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV) ;                         \
    cpl_parameterlist_append(self, p);                                         \
}

#define CRIRES_PARAMETER_GET_BOOL(MASK, VARNAME)                               \
if (bitmask & MASK) {                                                          \
    value = irplib_parameterlist_get_bool(self, PACKAGE, recipe, VARNAME);     \
                                                                               \
    cpl_ensure(!cpl_error_get_code(), cpl_error_get_code(), CPL_FALSE);        \
                                                                               \
    nbits++;         /* Total number of bits must be one  */                   \
    bitmask ^= MASK; /* - bitmask must be zero at the end */                   \
}

#define CRIRES_PARAMETER_GET_INT(MASK, VARNAME)                                \
if (bitmask & MASK) {                                                          \
    value = irplib_parameterlist_get_int(self, PACKAGE, recipe, VARNAME);      \
                                                                               \
    cpl_ensure(!cpl_error_get_code(), cpl_error_get_code(), 0);                \
                                                                               \
    nbits++;         /* Total number of bits must be one  */                   \
    bitmask ^= MASK; /* - bitmask must be zero at the end */                   \
}

#define CRIRES_PARAMETER_GET_DOUBLE(MASK, VARNAME)                             \
if (bitmask & MASK) {                                                          \
    value = irplib_parameterlist_get_double(self, PACKAGE, recipe, VARNAME);   \
                                                                               \
    cpl_ensure(!cpl_error_get_code(), cpl_error_get_code(), 0.0);              \
                                                                               \
    nbits++;         /* Total number of bits must be one  */                   \
    bitmask ^= MASK; /* - bitmask must be zero at the end */                   \
}

#define CRIRES_PARAMETER_GET_STRING(MASK, VARNAME)                             \
if (bitmask & MASK) {                                                          \
    value = irplib_parameterlist_get_string(self, PACKAGE, recipe, VARNAME);   \
                                                                               \
    cpl_ensure(value != NULL, cpl_error_get_code(), NULL);                     \
                                                                               \
    nbits++;         /* Total number of bits must be one  */                   \
    bitmask ^= MASK; /* - bitmask must be zero at the end */                   \
}

/*----------------------------------------------------------------------------*/
/**
 * @defgroup crires_parameter CRIRES recipe parameters
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Define the specified parameters
  @param    self     The parameterlist to set
  @param    recipe   The recipe name
  @param    bitmask  parameter specification
  @return   CPL_ERROR_NONE or the relevant CPL error code on error

 */
/*----------------------------------------------------------------------------*/
cpl_error_code crires_parameter_set(cpl_parameterlist * self,
                                  const char * recipe,
                                  crires_parameter bitmask)
{

    cpl_parameter * p;
    char * context;

    cpl_ensure_code(self,   CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(recipe, CPL_ERROR_NULL_INPUT);

    context = cpl_sprintf(PACKAGE ".%s", recipe);

    /* CPL_TYPE_INT */
    CRIRES_PARAMETER_SET(CRIRES_PARAM_DISPLAY, "display", CPL_TYPE_INT,
            "The chip to display (1 -> 4)", 0, "display");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_FWHM, "fwhm", CPL_TYPE_INT,
            "The FWHM for the convolution", 3, "fwhm");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_SEARCH_BOX, "search_box_sz", CPL_TYPE_INT,
            "The size of the search box for lines ident.", 10, "search_box_sz");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_MIN_MATCH, "min_matches_nb", CPL_TYPE_INT,
            "The minimum number of match requested", 3, "min_matches_nb");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_ORDER, "order", CPL_TYPE_INT,
            "Order", -1, "order");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_HOR_SIZE, "hor_size", CPL_TYPE_INT,
            "box size for profile computation", 31, "hor_size");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_SPEC_HSIZE, "spec_hsize", CPL_TYPE_INT,
            "Spectrum half size", 20, "spec_hsize");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_CLOSING_HSIZE, "closing_hs", CPL_TYPE_INT,
            "Closing kernel half size for BGD cleaning", 2, "closing_hs");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_WL_NBSAMPLES, "wl_samples", CPL_TYPE_INT,
            "Number of samples for the wavelength cal.", 100, "wl_samples");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_Y_WIDTH, "y_width", CPL_TYPE_INT,
            "Width in pix along Y of the zone", 10, "y_width");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_DEGREE, "degree", CPL_TYPE_INT,
            "Degree of the polynomial for the solution", 2, "degree");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_RON_SAMPLES, "nsamples", CPL_TYPE_INT,
            "Number of samples for RON computation", 100, "nsamples");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_RON_SZ, "hsize", CPL_TYPE_INT,
            "Half size of the window for RON computation", 6, "hsize");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_NORM_STARTY, "starty", CPL_TYPE_INT,
            "Y start of the zone used for median normalisation", -1, "starty");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_NORM_STOPY, "stopy", CPL_TYPE_INT,
            "Y stop of the zone used for median normalisation", -1, "stopy");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_EXTR_MODE, "mode", CPL_TYPE_INT,
            "1-spectrum, 2-lines", 1, "mode");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_CONFIG_MODE, "mode", CPL_TYPE_INT,
            "1 - REFINE_CONFIG / 2 - CONFIG", 0, "mode");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_LINES_MODE, "mode", CPL_TYPE_INT,
            "1-Hitran, 2-OH, 3-ThAr, 4-N2O, 5-Model", 1, "mode");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_PMIN, "pmin", CPL_TYPE_INT,
            "Minimum pixel to plot", -1, "pmin");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_PMAX, "pmax", CPL_TYPE_INT,
            "Maximum pixel to plot", -1, "pmax");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_FIRST_PLOTTED, "fplotted", CPL_TYPE_INT,
            "First Line to plot", 0, "fplotted");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_LAST_PLOTTED, "lplotted", CPL_TYPE_INT,
            "Last Line to plot", 0, "lplotted");
    
    /* CPL_TYPE_DOUBLE */
    CRIRES_PARAMETER_SET(CRIRES_PARAM_SIGMA, "sigma", CPL_TYPE_DOUBLE,
            "The sigma for the detection", 3.0, "sigma");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_KAPPA_SIGCLIP, "kappa_sigclip", 
            CPL_TYPE_DOUBLE, "The kappa for the sigma clipping collapse", 3.0,
            "kappa_sigclip");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_YPOS, "y_pos", CPL_TYPE_DOUBLE,
            "Approx. y co-ord of centre of arc lamp lines", -1.0, "y_pos");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_KAPPA, "kappa", CPL_TYPE_DOUBLE,
            "kappa for background removal", 1.0, "kappa");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_CLEAN_RATE, "clean_rate", CPL_TYPE_DOUBLE,
            "Min rate of bgd pixels per line allowed", 0.5, "clean_rate");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_WL_ERROR, "wl_err", CPL_TYPE_DOUBLE,
            "The wavelength error (in pixels)", 60.0, "wl_err");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_XC_LIMIT, "wl_xclimit", CPL_TYPE_DOUBLE,
            "The lowest allowed value for XC factor", 0.2, "wl_xclimit");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_BPM_RATE, "lines_ratio", CPL_TYPE_DOUBLE,
            "Maximum ratio of bad pixels per line", 0.5, "lines_ratio");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_TOT_NDIT, "tot_ndit", CPL_TYPE_DOUBLE,
            "Total number of ndit*dit", -1.0, "tot_ndit");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_WL_FACTOR, "wl_factor", CPL_TYPE_DOUBLE,
            "The factor used to multiply the wl", 1.0, "wl_factor");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_EXPTIME, "exptime", CPL_TYPE_DOUBLE,
            "Exposure time in seconds", -1.0, "exptime");

    /* CPL_TYPE_STRING */
    CRIRES_PARAMETER_SET(CRIRES_PARAM_Y_POS_CHIP1, "y_pos_c1", 
            CPL_TYPE_STRING, "Y positions on CHIP 1", "", "y_pos_c1");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_Y_POS_CHIP2, "y_pos_c2", 
            CPL_TYPE_STRING, "Y positions on CHIP 2", "", "y_pos_c2");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_Y_POS_CHIP3, "y_pos_c3", 
            CPL_TYPE_STRING, "Y positions on CHIP 3", "", "y_pos_c3");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_Y_POS_CHIP4, "y_pos_c4", 
            CPL_TYPE_STRING, "Y positions on CHIP 4", "", "y_pos_c4");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_WAVES, "waves", CPL_TYPE_STRING,
            "wavelength boundaries", 
            "-1.0,-1.0,-1.0,-1.0,-1.0,-1.0,-1.0,-1.0", "waves");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_REJECT, "rej", CPL_TYPE_STRING,
            "Left and right rejections", "0,0", "rej");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_Y_SPEC_ZONE_CHIP1, "spec_zone_c1", 
            CPL_TYPE_STRING, "Spectrum search zone in CHIP 1", "-1,-1", 
            "spec_zone_c1");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_Y_SPEC_ZONE_CHIP2, "spec_zone_c2", 
            CPL_TYPE_STRING, "Spectrum search zone in CHIP 2", "-1,-1", 
            "spec_zone_c2");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_Y_SPEC_ZONE_CHIP3, "spec_zone_c3", 
            CPL_TYPE_STRING, "Spectrum search zone in CHIP 3", "-1,-1", 
            "spec_zone_c3");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_Y_SPEC_ZONE_CHIP4, "spec_zone_c4", 
            CPL_TYPE_STRING, "Spectrum search zone in CHIP 4", "-1,-1", 
            "spec_zone_c4");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_THRESHOLDS, "thresholds", CPL_TYPE_STRING,
            "Low and high thresholds for the BPM", "0.5,2.0", "thresholds");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_LINES, "lines", CPL_TYPE_STRING,
            "lines file", NULL, "lines");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_COLLAPSE_METH, "coll_meth", 
            CPL_TYPE_STRING, "avg / med / sig", "sig", "coll_meth");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_COMB_USED, "comb_used", 
            CPL_TYPE_STRING, "Combined image to be used : NODA / NODB / COMB", 
            "COMB", "comb_used");

    /* CPL_TYPE_BOOL */
    CRIRES_PARAMETER_SET(CRIRES_PARAM_REFINE, "refine", CPL_TYPE_BOOL,
            "refine offsets", TRUE, "refine");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_ONLYA, "onlyA", CPL_TYPE_BOOL,
            "Also combine using only nodding A", FALSE, "onlyA");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_ONLYB, "onlyB", CPL_TYPE_BOOL,
            "Also combine using only nodding B", FALSE, "onlyB");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_BLIND, "blind", CPL_TYPE_BOOL,
            "Flag to get blindly the offsets", FALSE, "blind");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_WL_LOG, "wl_log", CPL_TYPE_BOOL,
            "Flag to use the logarithm of the catalog", FALSE, "wl_log");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_WL_NOLIMIT, "wl_nolimit", CPL_TYPE_BOOL,
            "Flag to ignore the WMIN and WMAX values from the header", FALSE, 
            "wl_nolimit");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_WL_CLEAN, "wl_clean", CPL_TYPE_BOOL,
            "Flag to remove the low freq. part of the spectrum", FALSE, 
            "wl_clean");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_FORCE, "force", CPL_TYPE_BOOL,
            "flag to force the computation", FALSE, "force");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_REPLACE, "replace", CPL_TYPE_BOOL,
            "Replace by 1 in the flat the pixels outside the bounds", FALSE, 
            "replace");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_CLEAN_FLAT, "clean_flat", CPL_TYPE_BOOL,
            "Clean the flat using the created BPM", FALSE, 
            "clean_flat");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_FILL_BLANKS, "fill_blanks", CPL_TYPE_BOOL,
            "flag to fill blanks", FALSE, "fill_blanks");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_PLOT, "display", CPL_TYPE_BOOL,
            "flag to plot", FALSE, "display");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_WL_MODEL, "model", CPL_TYPE_BOOL,
            "To use the Wavelength from the model", FALSE, "model");
    CRIRES_PARAMETER_SET(CRIRES_PARAM_ADJUST, "adjust", CPL_TYPE_BOOL,
            "To adjust the level of 2 signals before printing", FALSE,"adjust");
    cpl_free(context);

    cpl_ensure_code(bitmask == 0, CPL_ERROR_UNSUPPORTED_MODE);

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Retrieve the value of a CRIRES boolean parameter 
  @param    self     The parameterlist to get from
  @param    recipe   The recipe name
  @param    bitmask  parameter specification
  @return   The boolean value, or CPL_FALSE on error
  @note     If called with a pre-existing error, the function will do nothing.

 */
/*----------------------------------------------------------------------------*/
cpl_boolean crires_parameterlist_get_bool(const cpl_parameterlist * self,
                                        const char * recipe,
                                        crires_parameter bitmask)
{

    int nbits = 0;
    cpl_boolean value = CPL_FALSE; /* Avoid (false) uninit warning */


    cpl_ensure(!cpl_error_get_code(), cpl_error_get_code(), CPL_FALSE);
    cpl_ensure(self,   CPL_ERROR_NULL_INPUT, CPL_FALSE);
    cpl_ensure(recipe, CPL_ERROR_NULL_INPUT, CPL_FALSE);

    CRIRES_PARAMETER_GET_BOOL(CRIRES_PARAM_REFINE, "refine");
    CRIRES_PARAMETER_GET_BOOL(CRIRES_PARAM_ONLYA, "onlyA");
    CRIRES_PARAMETER_GET_BOOL(CRIRES_PARAM_ONLYB, "onlyB");
    CRIRES_PARAMETER_GET_BOOL(CRIRES_PARAM_BLIND, "blind");
    CRIRES_PARAMETER_GET_BOOL(CRIRES_PARAM_WL_LOG, "wl_log");
    CRIRES_PARAMETER_GET_BOOL(CRIRES_PARAM_WL_NOLIMIT, "wl_nolimit");
    CRIRES_PARAMETER_GET_BOOL(CRIRES_PARAM_WL_CLEAN, "wl_clean");
    CRIRES_PARAMETER_GET_BOOL(CRIRES_PARAM_FORCE, "force");
    CRIRES_PARAMETER_GET_BOOL(CRIRES_PARAM_REPLACE, "replace");
    CRIRES_PARAMETER_GET_BOOL(CRIRES_PARAM_CLEAN_FLAT, "clean_flat");
    CRIRES_PARAMETER_GET_BOOL(CRIRES_PARAM_FILL_BLANKS, "fill_blanks");
    CRIRES_PARAMETER_GET_BOOL(CRIRES_PARAM_PLOT, "display");
    CRIRES_PARAMETER_GET_BOOL(CRIRES_PARAM_WL_MODEL, "model");
    CRIRES_PARAMETER_GET_BOOL(CRIRES_PARAM_ADJUST, "adjust");

    cpl_ensure(bitmask == 0, CPL_ERROR_UNSUPPORTED_MODE, CPL_FALSE);
    cpl_ensure(nbits == 1,   CPL_ERROR_ILLEGAL_INPUT, CPL_FALSE);

    return value;

}

/*----------------------------------------------------------------------------*/
/**
  @brief    Retrieve the value of a CRIRES integer parameter 
  @param    self     The parameterlist to get from
  @param    recipe   The recipe name
  @param    bitmask  parameter specification
  @return   The integer value, or zero on error
  @note     If called with a pre-existing error, the function will do nothing.

 */
/*----------------------------------------------------------------------------*/
int crires_parameterlist_get_int(const cpl_parameterlist * self,
                               const char * recipe,
                               crires_parameter bitmask)
{

    int nbits = 0;
    int value = 0; /* Avoid (false) uninit warning */


    cpl_ensure(!cpl_error_get_code(), cpl_error_get_code(), 0);
    cpl_ensure(self,   CPL_ERROR_NULL_INPUT, 0);
    cpl_ensure(recipe, CPL_ERROR_NULL_INPUT, 0);

    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_DISPLAY, "display");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_FWHM, "fwhm");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_SEARCH_BOX, "search_box_sz");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_MIN_MATCH, "min_matches_nb");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_ORDER, "order");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_HOR_SIZE, "hor_size");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_SPEC_HSIZE, "spec_hsize");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_CLOSING_HSIZE, "closing_hs");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_WL_NBSAMPLES, "wl_samples");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_Y_WIDTH, "y_width");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_DEGREE, "degree");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_RON_SAMPLES, "nsamples");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_RON_SZ, "hsize");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_NORM_STARTY, "starty");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_NORM_STOPY, "stopy");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_EXTR_MODE, "mode");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_CONFIG_MODE, "mode");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_LINES_MODE, "mode");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_PMIN, "pmin");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_PMAX, "pmax");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_FIRST_PLOTTED, "fplotted");
    CRIRES_PARAMETER_GET_INT(CRIRES_PARAM_LAST_PLOTTED, "lplotted");

    cpl_ensure(bitmask == 0, CPL_ERROR_UNSUPPORTED_MODE, 0);
    cpl_ensure(nbits == 1,   CPL_ERROR_ILLEGAL_INPUT, 0);

    return value;

}

/*----------------------------------------------------------------------------*/
/**
  @brief    Retrieve the value of a CRIRES parameter of type double
  @param    self     The parameterlist to get from
  @param    recipe   The recipe name
  @param    bitmask  parameter specification
  @return   The value, or undefined on error
  @note     If called with a pre-existing error, the function will do nothing.

 */
/*----------------------------------------------------------------------------*/
double crires_parameterlist_get_double(const cpl_parameterlist * self,
                                     const char * recipe,
                                     crires_parameter bitmask)
{

    int nbits = 0;
    double value = DBL_MAX; /* Avoid (false) uninit warning */


    cpl_ensure(!cpl_error_get_code(), cpl_error_get_code(), 0.0);
    cpl_ensure(self,   CPL_ERROR_NULL_INPUT, 0.0);
    cpl_ensure(recipe, CPL_ERROR_NULL_INPUT, 0.0);

    CRIRES_PARAMETER_GET_DOUBLE(CRIRES_PARAM_SIGMA, "sigma");
    CRIRES_PARAMETER_GET_DOUBLE(CRIRES_PARAM_KAPPA_SIGCLIP, "kappa_sigclip");
    CRIRES_PARAMETER_GET_DOUBLE(CRIRES_PARAM_YPOS, "y_pos");
    CRIRES_PARAMETER_GET_DOUBLE(CRIRES_PARAM_KAPPA, "kappa");
    CRIRES_PARAMETER_GET_DOUBLE(CRIRES_PARAM_CLEAN_RATE, "clean_rate");
    CRIRES_PARAMETER_GET_DOUBLE(CRIRES_PARAM_WL_ERROR, "wl_err");
    CRIRES_PARAMETER_GET_DOUBLE(CRIRES_PARAM_XC_LIMIT, "wl_xclimit");
    CRIRES_PARAMETER_GET_DOUBLE(CRIRES_PARAM_BPM_RATE, "lines_ratio");
    CRIRES_PARAMETER_GET_DOUBLE(CRIRES_PARAM_TOT_NDIT, "tot_ndit");
    CRIRES_PARAMETER_GET_DOUBLE(CRIRES_PARAM_WL_FACTOR, "wl_factor");
    CRIRES_PARAMETER_GET_DOUBLE(CRIRES_PARAM_EXPTIME, "exptime");

    cpl_ensure(bitmask == 0, CPL_ERROR_UNSUPPORTED_MODE, 0.0);
    cpl_ensure(nbits == 1,   CPL_ERROR_ILLEGAL_INPUT, 0.0);

    return value;

}

/*----------------------------------------------------------------------------*/
/**
  @brief    Retrieve the value of a CRIRES string parameter 
  @param    self     The parameterlist to get from
  @param    recipe   The recipe name
  @param    bitmask  parameter specification
  @return   The string, or NULL on error

 */
/*----------------------------------------------------------------------------*/
const char * crires_parameterlist_get_string(const cpl_parameterlist * self,
                                           const char * recipe,
                                           crires_parameter bitmask)
{

    int nbits = 0;
    const char * value = NULL; /* Avoid (false) uninit warning */


    cpl_ensure(self,   CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(recipe, CPL_ERROR_NULL_INPUT, NULL);

    CRIRES_PARAMETER_GET_STRING(CRIRES_PARAM_Y_POS_CHIP1, "y_pos_c1");
    CRIRES_PARAMETER_GET_STRING(CRIRES_PARAM_Y_POS_CHIP2, "y_pos_c2");
    CRIRES_PARAMETER_GET_STRING(CRIRES_PARAM_Y_POS_CHIP3, "y_pos_c3");
    CRIRES_PARAMETER_GET_STRING(CRIRES_PARAM_Y_POS_CHIP4, "y_pos_c4");
    CRIRES_PARAMETER_GET_STRING(CRIRES_PARAM_WAVES, "waves");
    CRIRES_PARAMETER_GET_STRING(CRIRES_PARAM_REJECT, "rej");
    CRIRES_PARAMETER_GET_STRING(CRIRES_PARAM_Y_SPEC_ZONE_CHIP1, "spec_zone_c1");
    CRIRES_PARAMETER_GET_STRING(CRIRES_PARAM_Y_SPEC_ZONE_CHIP2, "spec_zone_c2");
    CRIRES_PARAMETER_GET_STRING(CRIRES_PARAM_Y_SPEC_ZONE_CHIP3, "spec_zone_c3");
    CRIRES_PARAMETER_GET_STRING(CRIRES_PARAM_Y_SPEC_ZONE_CHIP4, "spec_zone_c4");
    CRIRES_PARAMETER_GET_STRING(CRIRES_PARAM_THRESHOLDS, "thresholds");
    CRIRES_PARAMETER_GET_STRING(CRIRES_PARAM_LINES, "lines");
    CRIRES_PARAMETER_GET_STRING(CRIRES_PARAM_COLLAPSE_METH, "coll_meth");
    CRIRES_PARAMETER_GET_STRING(CRIRES_PARAM_COMB_USED, "comb_used");

    cpl_ensure(bitmask == 0, CPL_ERROR_UNSUPPORTED_MODE, NULL);
    cpl_ensure(nbits == 1,   CPL_ERROR_ILLEGAL_INPUT, NULL);

    assert(value != NULL);

    return value;

}

/**@}*/
