/* $Id: crires_utils.c,v 1.26 2012-10-09 08:18:01 yjung Exp $
 *
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2012-10-09 08:18:01 $
 * $Revision: 1.26 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include <math.h>
#include <string.h>

#include "crires_dfs.h"
#include "crires_utils.h"
#include "crires_load.h"
#include "crires_pfits.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup crires_utils     Miscellaneous Utilities
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the pipeline copyright and license
  @return   The copyright and license string

  The function returns a pointer to the statically allocated license string.
  This string should not be modified using the returned pointer.
 */
/*----------------------------------------------------------------------------*/
const char * crires_get_license(void)
{
    const char  *   crires_license = 
        "This file is part of the CRIRES Instrument Pipeline\n"
        "Copyright (C) 2002,2003 European Southern Observatory\n"
        "\n"
        "This program is free software; you can redistribute it and/or modify\n"
        "it under the terms of the GNU General Public License as published by\n"
        "the Free Software Foundation; either version 2 of the License, or\n"
        "(at your option) any later version.\n"
        "\n"
        "This program is distributed in the hope that it will be useful,\n"
        "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
        "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
        "GNU General Public License for more details.\n"
        "\n"
        "You should have received a copy of the GNU General Public License\n"
        "along with this program; if not, write to the Free Software\n"
        "Foundation, Inc., 59 Temple Place, Suite 330, Boston, \n"
        "MA  02111-1307  USA" ;
    return crires_license ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Merge upper and lower images to an image
  @param    ima_upper       the upper image
  @param    ima_lower       the lower image 
  @return   the merged image, NULL in error case
 */
/*----------------------------------------------------------------------------*/
cpl_image * crires_util_ima_merge(
        const cpl_image     *   ima_upper,
        const cpl_image     *   ima_lower)
{   
    cpl_image       *   out ;
    int                 nx, ny ;

    /* Check entries */
    if (ima_upper == NULL) return NULL ;
    if (ima_lower == NULL) return NULL ;
    nx = cpl_image_get_size_x(ima_upper) ;
    ny = cpl_image_get_size_y(ima_upper) ;

    if (nx != cpl_image_get_size_x(ima_lower)) return NULL ;
    if (ny != cpl_image_get_size_x(ima_lower)) return NULL ;

    /* Create the big image */
    out = cpl_image_new(nx, 2*ny, cpl_image_get_type(ima_upper)) ;

    /* Copy the lower part */
    if (cpl_image_copy(out, ima_lower, 1, 1) != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot copy the lower part") ;
        cpl_image_delete(out) ;
        return NULL ;
    }

    /* Copy the upper part */
    if (cpl_image_copy(out, ima_upper, 1, ny+1) != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot copy the upper part") ;
        cpl_image_delete(out) ;
        return NULL ;
    }
    return out ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Extract a window from an image list
  @param    ilist   the image list
  @param    lx
  @param    ly
  @param    ux
  @param    uy
  @return   the extracted image list, NULL in error case
 */
/*----------------------------------------------------------------------------*/
cpl_imagelist * crires_util_ilist_extract(
        const cpl_imagelist *   ilist,
        int                     lx,
        int                     ly,
        int                     ux,
        int                     uy)
{   
    cpl_imagelist       *   out ;
    int                     nima ;
    cpl_image           *   tmp ;
    int                     i ;

    /* Check entries */
    if (ilist == NULL) return NULL ;
    nima = cpl_imagelist_get_size(ilist) ;

    /* Create the imagelist */
    out = cpl_imagelist_new() ;

    /* Loop on the input images */
    for (i=0 ; i<nima ; i++) {
        /* Extract the current image */
        if ((tmp = cpl_image_extract(cpl_imagelist_get_const(ilist, i), 
                        lx, ly, ux, uy)) == NULL) {
            cpl_msg_error(__func__, "Cannot extract image %d", i+1) ;
            cpl_imagelist_delete(out) ;
            return NULL ;
        }
        /* Add the current image in the image list */
        if (cpl_imagelist_set(out, tmp, i) != CPL_ERROR_NONE) {
            cpl_msg_error(__func__, "Cannot add to the imagelist") ;
            cpl_image_delete(tmp) ;
            cpl_imagelist_delete(out) ;
            return NULL ;
        }
    }
    return out ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Remove the central part from an image
  @param    ima         the image with a central part to remove
  @param    stripe_y    the stripe y size
  @return   the new image is made from the lower part and the upper parts of 
            the input image (both parts of height stripe_y pixels).
 */
/*----------------------------------------------------------------------------*/
cpl_image * crires_util_remove_gap(
        const cpl_image     *       ima,
        int                         stripe_y)
{   
    cpl_image       *   out ;
    cpl_image       *   extracted ;
    int                 nx, ny ;

    /* Check entries */
    if (ima == NULL) return NULL ;
    nx = cpl_image_get_size_x(ima) ;
    ny = cpl_image_get_size_y(ima) ;
       
    /* Create the output image */
    out = cpl_image_new(nx, 2*stripe_y, cpl_image_get_type(ima)) ;
    
    /* Extract the lower part */
    extracted = cpl_image_extract(ima, 1, 1, nx, stripe_y) ;

    /* Copy the lower part */
    if (cpl_image_copy(out, extracted, 1, 1) != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot copy the lower part") ;
        cpl_image_delete(extracted) ;
        cpl_image_delete(out) ;
        return NULL ;
    }
    cpl_image_delete(extracted) ;

    /* Extract the upper part */
    extracted = cpl_image_extract(ima, 1, ny-stripe_y, nx, ny) ;

    /* Copy the upper part */
    if (cpl_image_copy(out, extracted, 1, stripe_y+1) != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot copy the upper part") ;
        cpl_image_delete(extracted) ;
        cpl_image_delete(out) ;
        return NULL ;
    }
    cpl_image_delete(extracted) ;

    return out ;
}
/*----------------------------------------------------------------------------*/
/**
  @brief    Extend an imagelist with a gap in the middle of the images
  @param    ilist       the image list
  @param    offset_y    Distance between the center of stripes
  @return   the extended image list, NULL in error case
  The extended image is the input image lower and upper halves with a gap of 
  pixels in the middle. The upper and lower parts centers are separated by 
  offset_y pixels.
 */
/*----------------------------------------------------------------------------*/
cpl_imagelist * crires_util_ilist_extend(
        const cpl_imagelist *   ilist,
        int                     offset_y)
{   
    cpl_imagelist       *   out ;
    int                     nima, nx, ny ;
    cpl_image           *   tmp ;
    cpl_image           *   extracted ;
    const cpl_image     *   cur_ima ;
    int                     i ;

    /* Check entries */
    if (ilist == NULL) return NULL ;
    nima = cpl_imagelist_get_size(ilist) ;
    nx = cpl_image_get_size_x(cpl_imagelist_get_const(ilist, 0)) ;
    ny = cpl_image_get_size_y(cpl_imagelist_get_const(ilist, 0)) ;

    /* Create the imagelist */
    out = cpl_imagelist_new() ;

    /* Loop on the input images */
    for (i=0 ; i<nima ; i++) {
        cur_ima = cpl_imagelist_get_const(ilist, i) ;

        /* Create the new image */
        tmp = cpl_image_new(nx, (int)(ny/2)+offset_y, cpl_image_get_type(cur_ima)) ;

        /* Extract the lower part */
        extracted = cpl_image_extract(cur_ima, 1, 1, nx, (int)(ny/2)) ;

        /* Copy the lower part */
        if (cpl_image_copy(tmp, extracted, 1, 1) != CPL_ERROR_NONE) {
            cpl_msg_error(__func__, "Cannot copy the lower part") ;
            cpl_image_delete(extracted) ;
            cpl_image_delete(tmp) ;
            cpl_imagelist_delete(out) ;
            return NULL ;
        }
        cpl_image_delete(extracted) ;

        /* Extract the upper part */
        extracted = cpl_image_extract(cur_ima, 1, (int)(ny/2)+1, nx, ny) ;

        /* Copy the upper part */
        if (cpl_image_copy(tmp, extracted, 1, offset_y+1) 
                != CPL_ERROR_NONE) {
            cpl_msg_error(__func__, "Cannot copy the upper part") ;
            cpl_image_delete(extracted) ;
            cpl_image_delete(tmp) ;
            cpl_imagelist_delete(out) ;
            return NULL ;
        }
        cpl_image_delete(extracted) ;

        /* Add the current image in the image list */
        if (cpl_imagelist_set(out, tmp, i) != CPL_ERROR_NONE) {
            cpl_msg_error(__func__, "Cannot add to the imagelist") ;
            cpl_image_delete(tmp) ;
            cpl_imagelist_delete(out) ;
            return NULL ;
        }
    }
    return out ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Check if the STRIP keywords match in two files
  @param    fname1  first file
  @param    fname2  second file
  @return   0 if they match, 1 if they are different, -1 in error case
 */
/*----------------------------------------------------------------------------*/
int crire_stripe_keys_mismatch(
        const char  *   fname1,
        const char  *   fname2)
{   
    cpl_propertylist    *   plist ;
    int                     offsety1, centery1, sizey1 ;
    int                     offsety2, centery2, sizey2 ;

    /* Switch off the stripe keys check */
    if (crires_stripe_keys_check_off()) return 0 ;

    /* Check entries */
    if ((fname1 == NULL) || (fname2 == NULL)) return -1 ;

    /* Get the first file infos */
    if ((plist=cpl_propertylist_load(fname1, 0)) == NULL) return -1 ;
    offsety1 = crires_pfits_get_stripe_offsety(plist) ;
    centery1 = crires_pfits_get_stripe_centery(plist) ;
    cpl_propertylist_delete(plist) ;
    if (cpl_error_get_code()) return -1 ;

    /* Get the first file infos */
    if ((plist=cpl_propertylist_load(fname1, 1)) == NULL) return -1 ;
    sizey1 = crires_pfits_get_naxis2(plist) ;
    cpl_propertylist_delete(plist) ;
    if (cpl_error_get_code()) return -1 ;

    /* Get the second file infos */
    if ((plist=cpl_propertylist_load(fname2, 0)) == NULL) return -1 ;
    offsety2 = crires_pfits_get_stripe_offsety(plist) ;
    centery2 = crires_pfits_get_stripe_centery(plist) ;
    cpl_propertylist_delete(plist) ;
    if (cpl_error_get_code()) return -1 ;

    /* Get the second file infos */
    if ((plist=cpl_propertylist_load(fname2, 1)) == NULL) return -1 ;
    sizey2 = crires_pfits_get_naxis2(plist) ;
    cpl_propertylist_delete(plist) ;
    if (cpl_error_get_code()) return -1 ;

    /* Compare */
    if ((sizey1 == sizey2) && 
            (offsety1 == offsety2) &&
            (centery1 == centery2)) {
        return 0 ;
    } else {
        return 1 ;
    }
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Find out the Low Y of the detector 1 illuminated zone
  @param    Period
  @return   the value or -1 in error case
 */
/*----------------------------------------------------------------------------*/
int crires_get_detector_ly1(crires_illum_period period)
{
    int                     res = -1 ;

    /* Check entries */
    if (period == CRIRES_ILLUM_UNKNOWN) return -1 ;
    if (period == CRIRES_ILLUM_LOWER_HALF) return -1 ;
    if (period == CRIRES_ILLUM_UPPER_HALF) return -1 ;
    if (period == CRIRES_ILLUM_FULL_DETECTOR) return -1 ;

    /* Get the position */
    if (period == CRIRES_ILLUM_PERIOD1)         res = CRIRES_ILLUM_P1_LY1 ;
    else if (period == CRIRES_ILLUM_PERIOD2)    res = CRIRES_ILLUM_P2_LY1 ;
    else if (period == CRIRES_ILLUM_PERIOD3)    res = CRIRES_ILLUM_P3_LY1 ;

    /* Return  */
    return res ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Find out the High Y of the detector 1 illuminated zone
  @param    Period
  @return   the value or -1 in error case
 */
/*----------------------------------------------------------------------------*/
int crires_get_detector_uy1(crires_illum_period period)
{
    int                     res = -1 ;

    /* Check entries */
    if (period == CRIRES_ILLUM_UNKNOWN) return -1 ;
    if (period == CRIRES_ILLUM_LOWER_HALF) return -1 ;
    if (period == CRIRES_ILLUM_UPPER_HALF) return -1 ;
    if (period == CRIRES_ILLUM_FULL_DETECTOR) return -1 ;

    /* Get the position */
    if (period == CRIRES_ILLUM_PERIOD1)         res = CRIRES_ILLUM_P1_UY1 ;
    else if (period == CRIRES_ILLUM_PERIOD2)    res = CRIRES_ILLUM_P2_UY1 ;
    else if (period == CRIRES_ILLUM_PERIOD3)    res = CRIRES_ILLUM_P3_UY1 ;

    /* Return  */
    return res ;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    Find out the Low Y of the detector 2 illuminated zone
  @param    Period
  @return   the value or -1 in error case
 */
/*----------------------------------------------------------------------------*/
int crires_get_detector_ly2(crires_illum_period period)
{
    int                     res = -1 ;

    /* Check entries */
    if (period == CRIRES_ILLUM_UNKNOWN) return -1 ;
    if (period == CRIRES_ILLUM_LOWER_HALF) return -1 ;
    if (period == CRIRES_ILLUM_UPPER_HALF) return -1 ;
    if (period == CRIRES_ILLUM_FULL_DETECTOR) return -1 ;

    /* Get the position */
    if (period == CRIRES_ILLUM_PERIOD1)         res = CRIRES_ILLUM_P1_LY2 ;
    else if (period == CRIRES_ILLUM_PERIOD2)    res = CRIRES_ILLUM_P2_LY2 ;
    else if (period == CRIRES_ILLUM_PERIOD3)    res = CRIRES_ILLUM_P3_LY2 ;

    /* Return  */
    return res ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Find out the High Y of the detector 2 illuminated zone
  @param    Period
  @return   the value or -1 in error case
 */
/*----------------------------------------------------------------------------*/
int crires_get_detector_uy2(crires_illum_period period)
{
    int                     res = -1 ;

    /* Check entries */
    if (period == CRIRES_ILLUM_UNKNOWN) return -1 ;
    if (period == CRIRES_ILLUM_LOWER_HALF) return -1 ;
    if (period == CRIRES_ILLUM_UPPER_HALF) return -1 ;
    if (period == CRIRES_ILLUM_FULL_DETECTOR) return -1 ;

    /* Get the position */
    if (period == CRIRES_ILLUM_PERIOD1)         res = CRIRES_ILLUM_P1_UY2 ;
    else if (period == CRIRES_ILLUM_PERIOD2)    res = CRIRES_ILLUM_P2_UY2 ;
    else if (period == CRIRES_ILLUM_PERIOD3)    res = CRIRES_ILLUM_P3_UY2 ;

    /* Return  */
    return res ;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    Find out the Low Y of the detector 3 illuminated zone
  @param    Period
  @return   the value or -1 in error case
 */
/*----------------------------------------------------------------------------*/
int crires_get_detector_ly3(crires_illum_period period)
{
    int                     res = -1 ;

    /* Check entries */
    if (period == CRIRES_ILLUM_UNKNOWN) return -1 ;
    if (period == CRIRES_ILLUM_LOWER_HALF) return -1 ;
    if (period == CRIRES_ILLUM_UPPER_HALF) return -1 ;
    if (period == CRIRES_ILLUM_FULL_DETECTOR) return -1 ;

    /* Get the position */
    if (period == CRIRES_ILLUM_PERIOD1)         res = CRIRES_ILLUM_P1_LY3 ;
    else if (period == CRIRES_ILLUM_PERIOD2)    res = CRIRES_ILLUM_P2_LY3 ;
    else if (period == CRIRES_ILLUM_PERIOD3)    res = CRIRES_ILLUM_P3_LY3 ;

    /* Return  */
    return res ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Find out the High Y of the detector 3 illuminated zone
  @param    Period
  @return   the value or -1 in error case
 */
/*----------------------------------------------------------------------------*/
int crires_get_detector_uy3(crires_illum_period period)
{
    int                     res = -1 ;

    /* Check entries */
    if (period == CRIRES_ILLUM_UNKNOWN) return -1 ;
    if (period == CRIRES_ILLUM_LOWER_HALF) return -1 ;
    if (period == CRIRES_ILLUM_UPPER_HALF) return -1 ;
    if (period == CRIRES_ILLUM_FULL_DETECTOR) return -1 ;

    /* Get the position */
    if (period == CRIRES_ILLUM_PERIOD1)         res = CRIRES_ILLUM_P1_UY3 ;
    else if (period == CRIRES_ILLUM_PERIOD2)    res = CRIRES_ILLUM_P2_UY3 ;
    else if (period == CRIRES_ILLUM_PERIOD3)    res = CRIRES_ILLUM_P3_UY3 ;

    /* Return  */
    return res ;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    Find out the Low Y of the detector 4 illuminated zone
  @param    Period
  @return   the value or -1 in error case
 */
/*----------------------------------------------------------------------------*/
int crires_get_detector_ly4(crires_illum_period period)
{
    int                     res = -1 ;

    /* Check entries */
    if (period == CRIRES_ILLUM_UNKNOWN) return -1 ;
    if (period == CRIRES_ILLUM_LOWER_HALF) return -1 ;
    if (period == CRIRES_ILLUM_UPPER_HALF) return -1 ;
    if (period == CRIRES_ILLUM_FULL_DETECTOR) return -1 ;

    /* Get the position */
    if (period == CRIRES_ILLUM_PERIOD1)         res = CRIRES_ILLUM_P1_LY4 ;
    else if (period == CRIRES_ILLUM_PERIOD2)    res = CRIRES_ILLUM_P2_LY4 ;
    else if (period == CRIRES_ILLUM_PERIOD3)    res = CRIRES_ILLUM_P3_LY4 ;

    /* Return  */
    return res ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Find out the High Y of the detector 4 illuminated zone
  @param    Period
  @return   the value or -1 in error case
 */
/*----------------------------------------------------------------------------*/
int crires_get_detector_uy4(crires_illum_period period)
{
    int                     res = -1 ;

    /* Check entries */
    if (period == CRIRES_ILLUM_UNKNOWN) return -1 ;
    if (period == CRIRES_ILLUM_LOWER_HALF) return -1 ;
    if (period == CRIRES_ILLUM_UPPER_HALF) return -1 ;
    if (period == CRIRES_ILLUM_FULL_DETECTOR) return -1 ;

    /* Get the position */
    if (period == CRIRES_ILLUM_PERIOD1)         res = CRIRES_ILLUM_P1_UY4 ;
    else if (period == CRIRES_ILLUM_PERIOD2)    res = CRIRES_ILLUM_P2_UY4 ;
    else if (period == CRIRES_ILLUM_PERIOD3)    res = CRIRES_ILLUM_P3_UY4 ;

    /* Return  */
    return res ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Convert the RA from hh mm ss to degrees 
  @param    hh      hours
  @param    mm      minutes 
  @param    ss      seconds 
  @return   RA in degrees 

  An arc-hour is 15 degrees,
  60 arc-minutes is one arc-hour and
  60 arc-seconds is one arc-minute.
  
 */
/*----------------------------------------------------------------------------*/
double crires_ra_hms2deg(int hh, int mm, double ss)
{
    return 15.0 * crires_dec_hms2deg(hh, mm, ss);
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Convert the DEC from dd mm ss to degrees
  @param    dd      degrees
  @param    mm      minutes 
  @param    ss      seconds  
  @return   DEC in degrees

  60 arc-minutes is one degree and
  60 arc-seconds is one arc-minute.

 */
/*----------------------------------------------------------------------------*/
double crires_dec_hms2deg(int dd, int mm, double ss)
{
    return ((double)ss/60.0 + (double)mm)/60.0 + dd;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Find out the base name of a file (i.e. without prefix path)
  @param    filename    Full path name to scan.
  @return   Pointer to char within the input string.
 */
/*----------------------------------------------------------------------------*/
char * crires_get_base_name(const char *filename)
{
    char *p ;
    p = strrchr (filename, '/');
    return p ? p + 1 : (char *) filename;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Find out the root part of a basename (name without extension).
  @param    filename    File name to scan.
  @return   Pointer to statically allocated string.
 */
/*----------------------------------------------------------------------------*/
char * crires_get_root_name(const char * filename)
{
    static char path[4096+1];
    char * lastdot ;

    if (strlen(filename)>4096) return NULL ;
    memset(path, 0, 4096);
    strcpy(path, filename);
    lastdot = strrchr(path, '.');
    if (lastdot == NULL) return path ;
    if ((!strcmp(lastdot, ".fits")) || (!strcmp(lastdot, ".FITS")) ||
        (!strcmp(lastdot, ".dat")) || (!strcmp(lastdot, ".DAT")) ||
        (!strcmp(lastdot, ".paf")) || (!strcmp(lastdot, ".PAF")) ||
        (!strcmp(lastdot, ".txt")) || (!strcmp(lastdot, ".TXT")) ||
        (!strcmp(lastdot, ".ascii")) || (!strcmp(lastdot, ".ASCII")))
    {
        lastdot[0] = (char)0;
    }
    return path ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Apply the BPM correction to an image
  @param    in          the input image
  @param    period
  @param    bpm         the BPM
  @param    chip        the chip number (1 to CRIRES_NB_DETECTORS)
  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
int crires_bpm_correct_image(
        cpl_image           *   in,
        crires_illum_period     period,
        const char          *   bpm,
        int                     chip)
{
    cpl_imagelist       *   callist ;
    cpl_mask            *   bpm_im_bin ;

    /* Check entries */
    if (in == NULL) return -1 ;
    if (period == CRIRES_ILLUM_UNKNOWN) return -1 ;
    if (bpm == NULL) return -1 ;
    if (chip < 1 || chip > CRIRES_NB_DETECTORS) return -1 ;

    /* Load the bpm */
    if ((callist = crires_load_file(bpm, period, CPL_TYPE_FLOAT)) == NULL) {
        cpl_msg_error(__func__, "Cannot load the bpm") ;
        return -1 ;
    }
    /* Convert the map to binary */
    bpm_im_bin = cpl_mask_threshold_image_create(
            cpl_imagelist_get(callist, chip-1), -0.5, 0.5) ;
    cpl_mask_not(bpm_im_bin) ;
    cpl_imagelist_delete(callist) ;
    /* Apply the bad pixels cleaning */
    cpl_image_reject_from_mask(in, bpm_im_bin);
    if (cpl_detector_interpolate_rejected(in) != CPL_ERROR_NONE) {
        cpl_mask_delete(bpm_im_bin) ;
        return -1 ;
    }
    cpl_mask_delete(bpm_im_bin) ;

    return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Apply the detector linearity correction
  @param    ilist   the input image list
  @param    period
  @param    detlin      the detlin coeffs
  @param    chip        the chip number (1 to CRIRES_NB_DETECTORS)
  @param    stripe_center_y     the y center position of the lower stripe
  @param    stripe_offset_y     the y offset between the 2 stripes centers
  @param    stripe_size_y       the y size of a stripe
  The passed detlin data are only taken in normal mode. If the data to 
  calibrate are WIN data, the stripe_* parameters tell where the detlin areas
  must be extracted from the passed full 1024x512 detlin data. 
  If stripe_* parameters are negative (normal mode), the full detlin data are 
  taken.
  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
int crires_detlin_correct(
        cpl_imagelist       *   ilist,
        crires_illum_period     period,
        const char          *   detlin,
        int                     chip,
        int                     stripe_center_y,
        int                     stripe_offset_y,
        int                     stripe_size_y)
{
    cpl_image       *   ima ;
    cpl_image       *   imb ;
    cpl_image       *   imc ;
    float           *   pima ;
    float           *   pimb ;
    float           *   pimc ;
    float           *   pdata ;
    int                 nx, ny, ni ;
    float               val, val2, val3 ;
    int                 i, j ;

    /* Test entries */
    if (!ilist || !detlin) return -1 ;
    if (period == CRIRES_ILLUM_UNKNOWN) return -1 ;
    if (period == CRIRES_ILLUM_LOWER_HALF) return -1 ;
    if (period == CRIRES_ILLUM_UPPER_HALF) return -1 ;
    if (chip < 1 || chip > CRIRES_NB_DETECTORS) return -1 ;

    /* Special treatment for WIN data */
    if (stripe_center_y > 0 && stripe_offset_y > 0 && stripe_size_y > 0) {
        /* Test entries */
        if (period != CRIRES_ILLUM_FULL_DETECTOR) return -1 ;
        if (chip == 1 || chip == CRIRES_NB_DETECTORS) return 0 ;

        /* Load the 3 coeffs images */
        ima = crires_load_chip_plane_stripes(detlin, chip, 0, CPL_TYPE_FLOAT, 
                stripe_center_y, stripe_offset_y, stripe_size_y) ; 
        imb = crires_load_chip_plane_stripes(detlin, chip, 1, CPL_TYPE_FLOAT, 
                stripe_center_y, stripe_offset_y, stripe_size_y) ; 
        imc = crires_load_chip_plane_stripes(detlin, chip, 2, CPL_TYPE_FLOAT, 
                stripe_center_y, stripe_offset_y, stripe_size_y) ; 
    } else {
        /* Test entries */
        if (period == CRIRES_ILLUM_FULL_DETECTOR) return -1 ;

        /* Load the 3 coeffs images */
        ima = crires_load_chip_plane(detlin, period, chip, 0, CPL_TYPE_FLOAT) ;
        imb = crires_load_chip_plane(detlin, period, chip, 1, CPL_TYPE_FLOAT) ;
        imc = crires_load_chip_plane(detlin, period, chip, 2, CPL_TYPE_FLOAT) ;
    }
    if (!ima || !imb || !imc) {
        cpl_msg_error(cpl_func, "Cannot load the detlin images") ;
        if (ima) cpl_image_delete(ima) ;
        if (imb) cpl_image_delete(imb) ;
        if (imc) cpl_image_delete(imc) ;
        return -1 ;
    }
    pima = cpl_image_get_data_float(ima) ;
    pimb = cpl_image_get_data_float(imb) ;
    pimc = cpl_image_get_data_float(imc) ;
    
    /* Test sizes */
    nx = cpl_image_get_size_x(cpl_imagelist_get(ilist, 0)) ;
    ny = cpl_image_get_size_y(cpl_imagelist_get(ilist, 0)) ;
    ni = cpl_imagelist_get_size(ilist) ;
    if ((cpl_image_get_size_x(ima) != nx) ||
            (cpl_image_get_size_x(imb) != nx) ||
            (cpl_image_get_size_x(imc) != nx) ||
            (cpl_image_get_size_y(ima) != ny) ||
            (cpl_image_get_size_y(imb) != ny) ||
            (cpl_image_get_size_y(imc) != ny)) {
        cpl_msg_error(cpl_func, "Incompatible sizes") ;
        cpl_image_delete(ima) ;
        cpl_image_delete(imb) ;
        cpl_image_delete(imc) ;
        return -1 ;
    }
    
    /* Loop on pixels */
    for (i=0 ; i<nx*ny ; i++) {
        if (fabs(pimc[i]) < 1e-5) {
            /* Correct this pixel in each plane */
            for (j=0 ; j<ni ; j++) {
                pdata = cpl_image_get_data_float(cpl_imagelist_get(ilist, j)) ;
                val = pdata[i] ;
                pdata[i] = val-pima[i] ;
            }
        } else if (fabs(pimb[i]) < 1e-3) {
            for (j=0 ; j<ni ; j++) {
                pdata = cpl_image_get_data_float(cpl_imagelist_get(ilist, j)) ;
                pdata[i] = 0.0 ;
            }
        } else {
            val2 = 2 * pimc[i] / (pimb[i] * pimb[i]) ;
            /* Correct this pixel in each plane */
            for (j=0 ; j<ni ; j++) {
                pdata = cpl_image_get_data_float(cpl_imagelist_get(ilist, j)) ;
                val = pdata[i] ;
                val3 = 1-2*val2*(pima[i]-val) ;
                if (val3 < 0.0) {
                    pdata[i] = val-pima[i] ;
                } else {
                    pdata[i]=((float)sqrt(val3)-1) / val2 ;
                }
            }
        }
    }
    /* Free and return */
    cpl_image_delete(ima) ;
    cpl_image_delete(imb) ;
    cpl_image_delete(imc) ;
    return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Test if the model is off
  @return   1 if the model is off, 0 if it is on
 */
/*----------------------------------------------------------------------------*/
int crires_model_off(void)
{
    if (getenv("CRIRES_NO_MODEL")==NULL) return 0 ;
    else return 1 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Test if the stripe keys check is off
  @return   1 if the check is off, 0 if it is on
 */
/*----------------------------------------------------------------------------*/
int crires_stripe_keys_check_off(void)
{
    if (getenv("CRIRES_NO_STRIPE_KEYS_CHECK")==NULL) return 0 ;
    else return 1 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Display the illumination information
  @param    period
  @return   void
 */
/*----------------------------------------------------------------------------*/
void crires_display_detector_illum(crires_illum_period period)
{
    int         ly1, ly2, ly3, ly4, uy1, uy2, uy3, uy4 ;

    /* Test entries */
    if (period == CRIRES_ILLUM_UNKNOWN) return ;
    if (period == CRIRES_ILLUM_LOWER_HALF) return ;
    if (period == CRIRES_ILLUM_UPPER_HALF) return ;
    if (period == CRIRES_ILLUM_FULL_DETECTOR) return ;

    if (period == CRIRES_ILLUM_PERIOD1) {
        ly1 = CRIRES_ILLUM_P1_LY1 ;
        uy1 = CRIRES_ILLUM_P1_UY1 ;
        ly2 = CRIRES_ILLUM_P1_LY2 ;
        uy2 = CRIRES_ILLUM_P1_UY2 ;
        ly3 = CRIRES_ILLUM_P1_LY3 ;
        uy3 = CRIRES_ILLUM_P1_UY3 ;
        ly4 = CRIRES_ILLUM_P1_LY4 ;
        uy4 = CRIRES_ILLUM_P1_UY4 ;
    } else if (period == CRIRES_ILLUM_PERIOD2) {
        ly1 = CRIRES_ILLUM_P2_LY1 ;
        uy1 = CRIRES_ILLUM_P2_UY1 ;
        ly2 = CRIRES_ILLUM_P2_LY2 ;
        uy2 = CRIRES_ILLUM_P2_UY2 ;
        ly3 = CRIRES_ILLUM_P2_LY3 ;
        uy3 = CRIRES_ILLUM_P2_UY3 ;
        ly4 = CRIRES_ILLUM_P2_LY4 ;
        uy4 = CRIRES_ILLUM_P2_UY4 ;
    } else if (period == CRIRES_ILLUM_PERIOD3) {
        ly1 = CRIRES_ILLUM_P3_LY1 ;
        uy1 = CRIRES_ILLUM_P3_UY1 ;
        ly2 = CRIRES_ILLUM_P3_LY2 ;
        uy2 = CRIRES_ILLUM_P3_UY2 ;
        ly3 = CRIRES_ILLUM_P3_LY3 ;
        uy3 = CRIRES_ILLUM_P3_UY3 ;
        ly4 = CRIRES_ILLUM_P3_LY4 ;
        uy4 = CRIRES_ILLUM_P3_UY4 ;
    } else return ;

    cpl_msg_info(__func__, "*********************************") ;
    cpl_msg_info(__func__, "Detector 1 : Y=%d => Y=%d pixels", ly1, uy1) ;
    cpl_msg_info(__func__, "Detector 2 : Y=%d => Y=%d pixels", ly2, uy2) ;
    cpl_msg_info(__func__, "Detector 3 : Y=%d => Y=%d pixels", ly3, uy3) ;
    cpl_msg_info(__func__, "Detector 4 : Y=%d => Y=%d pixels", ly4, uy4) ;
    cpl_msg_info(__func__, "*********************************") ;
    return ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Find out the period for the detector illumination position
  @param    filename
  @return   the period or CRIRES_ILLUM_UNKNOWN in error case
 */
/*----------------------------------------------------------------------------*/
crires_illum_period crires_get_detector_illum_period(const char * filename)
{
    cpl_propertylist     *  plist ;
    const char           *  arcfile ;
    const char           *  date_obs ;
    const char           *  pipefile ;
    int                     month, year, day ;
    crires_illum_period     res ;

    /* Load the header */
    if ((plist=cpl_propertylist_load(filename, 0)) == NULL) {
        cpl_msg_error(__func__, "Cannot get header");
        return CRIRES_ILLUM_UNKNOWN ;
    }

    /* Try with ARCFILE */
    arcfile = crires_pfits_get_arcfile(plist) ;
    if (arcfile != NULL) {
        /* Parse to get the date */
        if (sscanf(arcfile, "CRIRE.%d-%d-%dT%*s", &year, &month, &day) != 3) {
            cpl_msg_error(__func__, "Cannot parse the ARCFILE");
            cpl_propertylist_delete(plist) ;
            return CRIRES_ILLUM_UNKNOWN ;
        }
        if (year < 2009 || (year == 2009 && month < 2) || 
                (year == 2009 && month == 2 && day < 5)) {
            res = CRIRES_ILLUM_PERIOD1 ;
        } else if (year < 2009 || (year == 2009 && month < 5) || 
                (year == 2009 && month == 5 && day < 29)) {
            res = CRIRES_ILLUM_PERIOD2 ;
        } else {
            res = CRIRES_ILLUM_PERIOD3 ;
        }
    } else {
        /* Try with DATE-OBS */
        cpl_error_reset() ;
        date_obs = crires_pfits_get_date_obs(plist) ;
        if (date_obs != NULL) {
            /* Parse to get the date */
            if (sscanf(date_obs, "%d-%d-%dT%*s", &year, &month, &day) != 3) {
                cpl_msg_error(__func__, "Cannot parse the DATE-OBS");
                cpl_propertylist_delete(plist) ;
                return CRIRES_ILLUM_UNKNOWN ;
            }
            if (year < 2009 || (year == 2009 && month < 2) || 
                    (year == 2009 && month == 2 && day < 5)) {
                res = CRIRES_ILLUM_PERIOD1 ;
            } else if (year < 2009 || (year == 2009 && month < 5) || 
                    (year == 2009 && month == 5 && day < 29)) {
                res = CRIRES_ILLUM_PERIOD2 ;
            } else {
                res = CRIRES_ILLUM_PERIOD3 ;
            }
        } else {
            /* Try with PIPEFILE */
            cpl_error_reset() ;
            pipefile = crires_pfits_get_pipefile(plist) ;
            if (pipefile == NULL) {
                cpl_msg_warning(__func__,"Cannot find the file creation date") ;
                cpl_msg_warning(__func__,"--> Use the latest") ;
                cpl_error_reset() ;
                res = CRIRES_ILLUM_PERIOD3 ;
            } else {
                /* Parse to get the date */
                if (sscanf(pipefile, 
                            "r.CRIRE.%d-%d-%dT%*s", &year, &month, &day) == 3) {
                    if (year < 2009 || (year == 2009 && month < 2) || 
                            (year == 2009 && month == 2 && day < 5)) {
                        res = CRIRES_ILLUM_PERIOD1 ;
                    } else if (year < 2009 || (year == 2009 && month < 5) || 
                            (year == 2009 && month == 5 && day < 29)) {
                        res = CRIRES_ILLUM_PERIOD2 ;
                    } else {
                        res = CRIRES_ILLUM_PERIOD3 ;
                    }
                } else {
                    cpl_msg_warning(__func__, "Cannot parse the PIPEFILE");
                    cpl_msg_warning(__func__, "--> Use the latest") ;
                    res = CRIRES_ILLUM_PERIOD3 ;
                }
            }
        }
    }

    cpl_propertylist_delete(plist) ;
    return res ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Read the list of Y positions from a table
  @param    fname   The file containing the positions
  @param    chip    The requested chip
  @return   the positions in a vector or NULL
 */
/*----------------------------------------------------------------------------*/
cpl_vector * crires_read_y_positions(const char * fname, int chip)
{
    cpl_table   *   tab ;
    int             nb_entries ;
    cpl_vector  *   positions ;
    int             chip_nb ;
    int             i ;

    /* Check entries */
    if (fname == NULL) return NULL ;
    if (chip < 1 || chip > CRIRES_NB_DETECTORS) return NULL ;

    /* Initialise */
    nb_entries = 0 ;

    /* Load the table */
    if ((tab = cpl_table_load(fname, 1, 0)) == NULL) {
        cpl_msg_error(__func__, "Cannot load %s as a table", fname) ;
        return NULL ;
    }

    /* Count the number of postions */
    for (i=0 ; i<cpl_table_get_nrow(tab) ; i++) {
        chip_nb = (int)cpl_table_get_double(tab, CRIRES_CHIP_NB, (cpl_size)i,
                NULL) ;
        if (chip_nb == chip) nb_entries++ ;
    }

    /* No entry found */
    if (nb_entries == 0) {
        cpl_table_delete(tab) ;
        return NULL ;
    }

    /* Create the output vector */
    positions = cpl_vector_new(nb_entries) ;

    /* Re-initialise */
    nb_entries = 0 ;
    for (i=0 ; i<cpl_table_get_nrow(tab) ; i++) {
        chip_nb = (int)cpl_table_get_double(tab, CRIRES_CHIP_NB, (cpl_size)i,
                NULL) ;
        if (chip_nb == chip) {
            cpl_vector_set(positions, nb_entries,
                    cpl_table_get_double(tab, CRIRES_COL_DETPOSY, (cpl_size)i,
                         NULL)) ;
            nb_entries ++ ;
        }
    }
    cpl_table_delete(tab) ;
    return positions ;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    Parse the list of Y positions 
  @param    pos     The string containing the positions
  @return   the positions in a vector or NULL
  The input string contains comma separated positions.
  The string is parsed, and the positions are stored in a vector.
 */
/*----------------------------------------------------------------------------*/
cpl_vector * crires_parse_y_positions(const char * pos)
{
    const char  *   ptr = pos ;
    char            field[32];
    int             nb_entries ;
    cpl_vector  *   positions ;
    int             n;

    /* Initialise */
    nb_entries = 0 ;

    /* Count the number of postions */
    while ( sscanf(ptr, "%31[^,]%n", field, &n) == 1 ) {
        /* Advance the pointer by the number of characters read */
        ptr += n;
        if (atof(field) == 0) break ;
        nb_entries ++ ;
        /* Didn't find an expected delimiter, done? */
        if ( *ptr != ',' ) {
            break ;
        }
        /* Skip the delimiter */
        ++ptr;
    }

    /* No entry found */
    if (nb_entries == 0) return NULL ;

    /* Create the output vector */
    positions = cpl_vector_new(nb_entries) ;

    /* Re-initialise */
    nb_entries = 0 ;
    ptr = pos ;
    while ( sscanf(ptr, "%31[^,]%n", field, &n) == 1 ) {
        /* Advance the pointer by the number of characters read */
        ptr += n;
        if (atof(field) == 0) break ;
        cpl_vector_set(positions, nb_entries, atof(field)) ;
        nb_entries ++ ;
        /* Didn't find an expected delimiter, done? */
        if ( *ptr != ',' ) {
            break ;
        }
        /* Skip the delimiter */
        ++ptr;
    }
    return positions ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the Spectrum position form the header
  @param    fname   Name of the extracted spectrum file
  @param    ext     The extension number
  @return   the spectrum Y position (1->..)
 */
/*----------------------------------------------------------------------------*/
int crires_get_y_spec_position(const char * fname, int ext)
{
    cpl_propertylist    *   plist ;
    int                     specpos ;

    /* Test Entries */
    if (fname == NULL) return -1 ;
    if (ext < 1 || ext > CRIRES_NB_DETECTORS) return -1 ;

    /* Load the propertylist */
    if ((plist=cpl_propertylist_load(fname, ext)) == NULL) {
        return -1 ;
    }

    specpos = crires_pfits_get_specpos(plist) ;

    cpl_propertylist_delete(plist) ;
    return specpos ;
}

/**@}*/
