/* $Id: crires_spec_wavecal.c,v 1.75 2012-10-09 08:18:01 yjung Exp $
 *
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2012-10-09 08:18:01 $
 * $Revision: 1.75 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <locale.h>
#include "crires_recipe.h"

#include "crires_combine.h"
#include "crires_wlcalib.h"
#include "crires_extract.h"
#include "crires_model_kernel.h"
#include "irplib_utils.h"

/*-----------------------------------------------------------------------------
                                Define
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "crires_spec_wavecal"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int crires_spec_wavecal_save(const cpl_imagelist *, 
        const cpl_imagelist *, const cpl_table **, const cpl_parameterlist *, 
        cpl_frameset *) ;

static char crires_spec_wavecal_description[] =
"crires_spec_wavecal -- Wavelength calibration\n"
"The files listed in the Set Of Frames (sof-file) must be tagged:\n"
"raw-file.fits "CRIRES_SPEC_WAVECAL_SKY_RAW" or\n"
"raw-file.fits "CRIRES_SPEC_WAVECAL_LAMP_RAW" or\n" 
"raw-file.fits "CRIRES_SPEC_WAVECAL_ABS_RAW" or\n" 
"flat-file.fits "CRIRES_CALPRO_FLAT" or\n"
"bpm-file.fits "CRIRES_CALPRO_BPM" or\n"
"dark-file.fits "CRIRES_CALPRO_DARK" or\n"
"detlin-file.fits "CRIRES_CALPRO_COEFFS_CUBE" or\n"
"catalog-file.fits "CRIRES_CALPRO_THAR_CAT" or\n" 
"catalog-file.fits "CRIRES_CALPRO_OH_CAT" or\n" 
"catalog-file.fits "CRIRES_CALPRO_HITRAN_CAT" or\n" 
"ypos-file.fits "CRIRES_CALPRO_THAR_POS" or\n" 
"model-config-file.fits "CRIRES_CALPRO_MODEL_CONFIG".\n"
"\n"
"This recipe produces 3 files:\n"
"First product:     the image with the wavelength values.\n"
"                   (PRO TYPE = "CRIRES_PROTYPE_WL_MAP")\n"
"Second product:    the image with the wavelength values from the model.\n"
"                   (PRO TYPE = "CRIRES_PROTYPE_WL_MAP")\n"
"Third product:     the table with the wavelength polynomial coefficients.\n"
"                   (PRO TYPE = "CRIRES_PROTYPE_WL_POLY")\n" ;

CRIRES_RECIPE_DEFINE(crires_spec_wavecal,
        CRIRES_PARAM_WAVES          |
        CRIRES_PARAM_DISPLAY        |
        CRIRES_PARAM_WL_LOG         |
        CRIRES_PARAM_WL_NOLIMIT     |
        CRIRES_PARAM_WL_ERROR       |
        CRIRES_PARAM_XC_LIMIT       |
        CRIRES_PARAM_WL_NBSAMPLES   |
        CRIRES_PARAM_Y_POS_CHIP1    |
        CRIRES_PARAM_Y_POS_CHIP2    |
        CRIRES_PARAM_Y_POS_CHIP3    |
        CRIRES_PARAM_Y_POS_CHIP4    |
        CRIRES_PARAM_Y_WIDTH        |
        CRIRES_PARAM_DEGREE         |
        CRIRES_PARAM_WL_CLEAN,
        "Wavelength calibration",
        crires_spec_wavecal_description) ;

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    double              wstart[CRIRES_NB_DETECTORS] ;
    double              wstop[CRIRES_NB_DETECTORS] ;
    int                 wl_nolimit ;
    int                 wl_log ;
    const char      *   wl_ypos_c1 ;
    const char      *   wl_ypos_c2 ;
    const char      *   wl_ypos_c3 ;
    const char      *   wl_ypos_c4 ;
    int                 wl_width ;
    double              wl_fwhm ;
    double              wl_slitw ;
    int                 wl_degree ;
    double              wl_err ;
    int                 wl_samples ;
    int                 wl_clean ;
    double              wl_xclimit ;
    int                 wl_ppm ;
    int                 display ;
    /* Outputs */
    crires_illum_period period ;
    int                 mode ;
    double              qc_wlxc[CRIRES_NB_DETECTORS] ;
    double              qc_wlcent[CRIRES_NB_DETECTORS] ;
    double              qc_wldisp[CRIRES_NB_DETECTORS] ;
    double              qc_lines_flux[CRIRES_NB_DETECTORS] ;
    double              qc_fwhm[CRIRES_NB_DETECTORS] ;
    double              qc_rpower[CRIRES_NB_DETECTORS] ;
} crires_spec_wavecal_config ;

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    parlist     the parameters list
  @param    frameset   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int crires_spec_wavecal(
        cpl_frameset            *   frameset,
        const cpl_parameterlist *   parlist)
{
    const char          *   sval ;
    const char          *   wl_ypos ;
    cpl_frameset        *   rawframes ;
    const char          *   fname ;
    const char          *   flat ;
    const char          *   dark ;
    const char          *   bpm ;
    const char          *   detlin ;
    const char          *   thar_cat ;
    const char          *   n2o_cat ;
    const char          *   oh_cat ;
    const char          *   hitran_cat ;
    const char          *   cfg_model ;
    const char          *   thar_positions ;
    cpl_propertylist    *   plist ;
    double                  wmin, wmax ;
    cpl_table           *   wave_tab[CRIRES_NB_DETECTORS] ;
    cpl_imagelist       *   wl_map ;
    cpl_imagelist       *   wl_map_model ;
    cpl_vector          *   wave_ypos ;
    int                     pix ;
    int                     i, j ;

    /* Needed for sscanf() */
    setlocale(LC_NUMERIC, "C");

    /* Initialise */
    rawframes = NULL ;
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        crires_spec_wavecal_config.qc_wlxc[i] = -1.0 ;
        crires_spec_wavecal_config.qc_wlcent[i] = -1.0 ;
        crires_spec_wavecal_config.qc_wldisp[i] = -1.0 ;
        crires_spec_wavecal_config.qc_lines_flux[i] = -1.0 ;
        crires_spec_wavecal_config.qc_fwhm[i] = -1.0 ;
        crires_spec_wavecal_config.qc_rpower[i] = -1.0 ;
    }
    crires_spec_wavecal_config.wl_ppm = 0 ;
    crires_spec_wavecal_config.wl_slitw = 2.0 ;
    crires_spec_wavecal_config.wl_fwhm = 2.0 ;
    
    /* Retrieve input parameters */

    sval = crires_parameterlist_get_string(parlist, RECIPE_STRING,
            CRIRES_PARAM_WAVES) ;
    if (sscanf(sval, "%lg,%lg,%lg,%lg,%lg,%lg,%lg,%lg",
                    &crires_spec_wavecal_config.wstart[0],
                    &crires_spec_wavecal_config.wstop[0],
                    &crires_spec_wavecal_config.wstart[1],
                    &crires_spec_wavecal_config.wstop[1],
                    &crires_spec_wavecal_config.wstart[2],
                    &crires_spec_wavecal_config.wstop[2],
                    &crires_spec_wavecal_config.wstart[3],
                    &crires_spec_wavecal_config.wstop[3])!=2*CRIRES_NB_DETECTORS){
        return -1 ;
    }
    crires_spec_wavecal_config.display = crires_parameterlist_get_int(parlist,
            RECIPE_STRING, CRIRES_PARAM_DISPLAY) ;
    crires_spec_wavecal_config.wl_log = crires_parameterlist_get_bool(parlist,
            RECIPE_STRING, CRIRES_PARAM_WL_LOG) ;
    crires_spec_wavecal_config.wl_nolimit = crires_parameterlist_get_bool(
            parlist, RECIPE_STRING, CRIRES_PARAM_WL_NOLIMIT) ;
    crires_spec_wavecal_config.wl_degree = crires_parameterlist_get_int(parlist,
            RECIPE_STRING, CRIRES_PARAM_DEGREE) ;
    crires_spec_wavecal_config.wl_err = crires_parameterlist_get_double(parlist,
            RECIPE_STRING, CRIRES_PARAM_WL_ERROR) ;
    crires_spec_wavecal_config.wl_xclimit = crires_parameterlist_get_double(
            parlist, RECIPE_STRING, CRIRES_PARAM_XC_LIMIT) ;
    crires_spec_wavecal_config.wl_ypos_c1=crires_parameterlist_get_string(
            parlist, RECIPE_STRING, CRIRES_PARAM_Y_POS_CHIP1) ;
    crires_spec_wavecal_config.wl_ypos_c2=crires_parameterlist_get_string(
            parlist, RECIPE_STRING, CRIRES_PARAM_Y_POS_CHIP2) ;
    crires_spec_wavecal_config.wl_ypos_c3=crires_parameterlist_get_string(
            parlist, RECIPE_STRING, CRIRES_PARAM_Y_POS_CHIP3) ;
    crires_spec_wavecal_config.wl_ypos_c4=crires_parameterlist_get_string(
            parlist, RECIPE_STRING, CRIRES_PARAM_Y_POS_CHIP4) ;
    crires_spec_wavecal_config.wl_width= crires_parameterlist_get_int(parlist,
            RECIPE_STRING, CRIRES_PARAM_Y_WIDTH) ;
    crires_spec_wavecal_config.wl_samples = crires_parameterlist_get_int(
            parlist, RECIPE_STRING, CRIRES_PARAM_WL_NBSAMPLES) ;
    crires_spec_wavecal_config.wl_clean = crires_parameterlist_get_bool(parlist,
            RECIPE_STRING, CRIRES_PARAM_WL_CLEAN) ;

    /* Identify the RAW and CALIB frames in the input frameset */
    if (crires_dfs_set_groups(frameset, "crires_spec_wavecal")) {
        cpl_msg_error(__func__, "Cannot identify RAW and CALIB frames") ;
        return -1 ;
    }

    /* Retrieve calibration data */
    flat        = crires_extract_filename(frameset, CRIRES_CALPRO_FLAT) ;
    dark        = crires_extract_filename(frameset, CRIRES_CALPRO_DARK) ;
    bpm         = crires_extract_filename(frameset, CRIRES_CALPRO_BPM) ;
    detlin      = crires_extract_filename(frameset, CRIRES_CALPRO_COEFFS_CUBE) ;
    thar_cat = crires_extract_filename(frameset, CRIRES_CALPRO_THAR_CAT) ;
    n2o_cat = crires_extract_filename(frameset, CRIRES_CALPRO_N2O_CAT) ;
    oh_cat = crires_extract_filename(frameset, CRIRES_CALPRO_OH_CAT) ;
    hitran_cat = crires_extract_filename(frameset, CRIRES_CALPRO_HITRAN_CAT) ;
    cfg_model = crires_extract_filename(frameset, CRIRES_CALPRO_MODEL_CONFIG) ;
    thar_positions = crires_extract_filename(frameset, CRIRES_CALPRO_THAR_POS) ;
    
    /* Retrieve raw frames */
    if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_SPEC_WAVECAL_SKY_RAW)) != NULL) {
        crires_spec_wavecal_config.mode = 1 ;
    } else if ((rawframes = crires_extract_frameset(frameset, 
                    CRIRES_SPEC_WAVECAL_LAMP_RAW)) != NULL) {
        crires_spec_wavecal_config.mode = 2 ;
    } else if ((rawframes = crires_extract_frameset(frameset, 
                    CRIRES_SPEC_WAVECAL_ABS_RAW)) != NULL) {
        crires_spec_wavecal_config.mode = 3 ;
    } else {
        cpl_msg_error(__func__, "No raw frame in input") ;
        return -1 ;
    }

    /* Get the detector illumination period */
    crires_spec_wavecal_config.period =
        crires_get_detector_illum_period(
            cpl_frame_get_filename(cpl_frameset_get_position(rawframes, 0))) ;
    if (crires_spec_wavecal_config.period == CRIRES_ILLUM_UNKNOWN) {
        cpl_msg_error(__func__,
                "Cannot determine the detector illumination period") ;
        cpl_frameset_delete(rawframes) ;
        return -1 ;
    } else {
        crires_display_detector_illum(crires_spec_wavecal_config.period) ;
    }

    /* Reduce the first raw frame */
    fname = cpl_frame_get_filename(cpl_frameset_get_position(rawframes,0)) ;

    /* Get the Minimum and Maximum wavelengths */
    if (crires_spec_wavecal_config.wl_nolimit == 0) {
        plist = cpl_propertylist_load(fname, 0) ;
        wmin = crires_pfits_get_wlen_min(plist) ;
        wmax = crires_pfits_get_wlen_max(plist) ;
        cpl_propertylist_delete(plist) ;
        if (cpl_error_get_code()) {
            wmin = wmax = -1.0 ;
            cpl_error_reset() ;
        }
    } else {
        wmin = wmax = -1.0 ;
    }
   
    /* Wavelength calibration */
    cpl_msg_info(__func__, "Apply the Wavelength Calibration") ;
    cpl_msg_indent_more() ;
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        cpl_msg_info(__func__, "Calibrate chip number %d", i+1) ;
        cpl_msg_indent_more() ;

        /* Where Compute the wavelength from ? */
        wl_ypos = "" ;
        if (i+1 == 1) wl_ypos = crires_spec_wavecal_config.wl_ypos_c1 ;
        if (i+1 == 2) wl_ypos = crires_spec_wavecal_config.wl_ypos_c2 ;
        if (i+1 == 3) wl_ypos = crires_spec_wavecal_config.wl_ypos_c3 ;
        if (i+1 == 4) wl_ypos = crires_spec_wavecal_config.wl_ypos_c4 ;

        /* Try to parse the user specified positions */
        if (!strcmp(wl_ypos, "")) {
            wave_ypos = NULL ;
        } else {
            cpl_msg_info(__func__, 
                    "Use the Y positions provided on the command line") ;
            if ((wave_ypos = crires_parse_y_positions(wl_ypos)) == NULL) {
                cpl_msg_warning(__func__,
                        "Cannot parse the y_positions value : %s", wl_ypos) ;
            }
        }
        
        /* Otherwise use passed calibration file */
        if (wave_ypos == NULL && thar_positions != NULL) {
            cpl_msg_info(__func__, 
                    "Use the Y positions provided in the FITS file") ;
            if ((wave_ypos = crires_read_y_positions(thar_positions, 
                            i+1)) == NULL) {
                cpl_msg_warning(__func__,
                        "Cannot read the Y positions from file %s", 
                        thar_positions) ;
            }
        }

        /* Try to detect the positions */
        if (wave_ypos == NULL) {
            cpl_msg_info(__func__, 
                    "Try an automatic detection of the Y positions") ;
            wave_ypos = crires_wlcalib_detect_wave_ypos(fname, 
                    crires_spec_wavecal_config.period, i+1);
        }

        /* Check the Mode */
        if (crires_spec_wavecal_config.mode == 1) {
            /* Calibrate from the sky */
            cpl_msg_info(__func__,"Get the calibration from the oh/hitran sky");
            wave_tab[i] = crires_wlcalib_sky(fname,
                    crires_spec_wavecal_config.period,
                    oh_cat, hitran_cat, crires_spec_wavecal_config.wl_log,
                    flat, dark, bpm, detlin,
                    crires_spec_wavecal_config.wstart[i],
                    crires_spec_wavecal_config.wstop[i],
                    wmin, wmax, i+1, 
                    wave_ypos,
                    crires_spec_wavecal_config.wl_width,
                    crires_spec_wavecal_config.wl_degree,
                    crires_spec_wavecal_config.wl_slitw,
                    crires_spec_wavecal_config.wl_fwhm,
                    crires_spec_wavecal_config.wl_err,
                    crires_spec_wavecal_config.wl_samples,
                    crires_spec_wavecal_config.wl_clean,
                    crires_spec_wavecal_config.wl_xclimit,
                    crires_spec_wavecal_config.wl_ppm,
                    (i+1==crires_spec_wavecal_config.display)) ;
        } else if (crires_spec_wavecal_config.mode == 2) {
            /* Calibrate from the lamp */
            cpl_msg_info(__func__, "Get the calibration from the thar lamp") ;
            wave_tab[i] = crires_wlcalib_lamp(fname,
                    crires_spec_wavecal_config.period,
                    thar_cat, crires_spec_wavecal_config.wl_log,
                    flat, dark, bpm, detlin,
                    crires_spec_wavecal_config.wstart[i],
                    crires_spec_wavecal_config.wstop[i],
                    wmin, wmax, i+1,
                    wave_ypos,
                    crires_spec_wavecal_config.wl_width,
                    crires_spec_wavecal_config.wl_degree,
                    crires_spec_wavecal_config.wl_slitw,
                    crires_spec_wavecal_config.wl_fwhm,
                    crires_spec_wavecal_config.wl_err,
                    crires_spec_wavecal_config.wl_samples,
                    crires_spec_wavecal_config.wl_clean,
                    crires_spec_wavecal_config.wl_xclimit,
                    crires_spec_wavecal_config.wl_ppm,
                    (i+1==crires_spec_wavecal_config.display),
                    &(crires_spec_wavecal_config.qc_lines_flux[i]),
                    &(crires_spec_wavecal_config.qc_fwhm[i]),
                    &(crires_spec_wavecal_config.qc_rpower[i])) ;
        } else if (crires_spec_wavecal_config.mode == 3) {
            /* Calibrate from the gas cell */
            cpl_msg_info(__func__, "Get the calibration from the n2o gas cell");
            wave_tab[i] = crires_wlcalib_lamp(fname,
                    crires_spec_wavecal_config.period,
                    n2o_cat, crires_spec_wavecal_config.wl_log, 
                    flat, dark, bpm, detlin,
                    crires_spec_wavecal_config.wstart[i],
                    crires_spec_wavecal_config.wstop[i],
                    wmin, wmax, i+1,
                    wave_ypos,
                    crires_spec_wavecal_config.wl_width,
                    crires_spec_wavecal_config.wl_degree,
                    crires_spec_wavecal_config.wl_slitw,
                    crires_spec_wavecal_config.wl_fwhm,
                    crires_spec_wavecal_config.wl_err,
                    crires_spec_wavecal_config.wl_samples,
                    crires_spec_wavecal_config.wl_clean,
                    crires_spec_wavecal_config.wl_xclimit,
                    crires_spec_wavecal_config.wl_ppm,
                    (i+1==crires_spec_wavecal_config.display),
                    &(crires_spec_wavecal_config.qc_lines_flux[i]), 
                    &(crires_spec_wavecal_config.qc_fwhm[i]),
                    &(crires_spec_wavecal_config.qc_rpower[i])) ;
        }
        cpl_msg_indent_less() ;
    cpl_vector_delete(wave_ypos) ;
    }
    cpl_msg_indent_less() ;

    /* Create the wave map */
    if ((wl_map = crires_wlcalib_gen_wlmap((const cpl_table **)wave_tab))
            == NULL) {
        cpl_msg_error(__func__, "Cannot compute the Wavelength Map") ;
        cpl_frameset_delete(rawframes) ;
        for (j=0 ; j<CRIRES_NB_DETECTORS ; j++) {
            if (wave_tab[j] != NULL) cpl_table_delete(wave_tab[j]);
        }
        return -1 ;
    }

    /* Compute the QC parameters */
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        crires_spec_wavecal_config.qc_wlcent[i] =
            cpl_image_get(cpl_imagelist_get(wl_map, i),
                    512, 256, &pix) ;
        crires_spec_wavecal_config.qc_wldisp[i] =
            ((cpl_image_get(cpl_imagelist_get(wl_map, i), 1024,
                            256, &pix)) -
             (cpl_image_get(cpl_imagelist_get(wl_map, i), 1,
                            256, &pix)))
            / 1023 ;
        crires_spec_wavecal_config.qc_wlxc[i] =
            crires_wlcalib_get_better_xc(wave_tab[i]) ;
    }

    /* Get the wl map from the model */
    if ((cfg_model != NULL) && (!crires_model_off()) && 
            (crires_model_config_check(cfg_model, fname) == 0)) {
        cpl_msg_info(__func__, "Call the model to get the wavelength map") ;
        cpl_msg_indent_more() ;
        wl_map_model = crires_model_wavpix(fname, cfg_model, -1) ;
        if (wl_map_model == NULL) {
            cpl_msg_warning(__func__, "Model function returns NULL") ;
            cpl_error_reset() ;
        }
        cpl_msg_indent_less() ;
    } else {
        wl_map_model = NULL ;
    }
    cpl_frameset_delete(rawframes) ;
    
    /* Save the product */
    cpl_msg_info(__func__, "Save the product") ;
    cpl_msg_indent_more() ;
    if (crires_spec_wavecal_save(wl_map, wl_map_model, 
                (const cpl_table **)wave_tab, parlist, frameset)) {
        cpl_msg_error(__func__, "Cannot save the product") ;
        for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) 
            if (wave_tab[i] != NULL) cpl_table_delete(wave_tab[i]);
        cpl_imagelist_delete(wl_map) ;
        if (wl_map_model) cpl_imagelist_delete(wl_map_model) ;
        cpl_msg_indent_less() ;
        return -1 ;
    }
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++)
        if (wave_tab[i] != NULL) cpl_table_delete(wave_tab[i]);
    cpl_imagelist_delete(wl_map) ;
    if (wl_map_model) cpl_imagelist_delete(wl_map_model) ;
    cpl_msg_indent_less() ;

    /* Return */
    if (cpl_error_get_code()) return -1 ;
    else return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the products
  @param    ilist       the wavelength images
  @param    ilist_model the wavelength images from the model
  @param    wl_tab      the wavelength tables
  @param    parlist     the parameters list
  @param    set         the frames set
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int crires_spec_wavecal_save(
        const cpl_imagelist     *   ilist,
        const cpl_imagelist     *   ilist_model,
        const cpl_table         **  wl_tab,
        const cpl_parameterlist *   parlist,
        cpl_frameset            *   set)
{
    cpl_propertylist    **  qclists ;
    const cpl_frame     *   ref_frame ;
    cpl_propertylist    *   inputlist ;
    const char          *   recipe_name = "crires_spec_wavecal" ;
    int                     i ;

    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW) ;

    /* Create the QC lists */
    qclists = cpl_malloc(CRIRES_NB_DETECTORS * sizeof(cpl_propertylist*)) ;
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        qclists[i] = cpl_propertylist_new() ;
        cpl_propertylist_append_double(qclists[i], "ESO QC CENTWL",
                crires_spec_wavecal_config.qc_wlcent[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC DISPWL",
                crires_spec_wavecal_config.qc_wldisp[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC XCORR",
                crires_spec_wavecal_config.qc_wlxc[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC LINES FLUX",
                crires_spec_wavecal_config.qc_lines_flux[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC FWHM MED",
                crires_spec_wavecal_config.qc_fwhm[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC RESOL MED",
                crires_spec_wavecal_config.qc_rpower[i]) ;

        /* Propagate some keywords from input raw frame extensions */
        inputlist = cpl_propertylist_load_regexp(
                cpl_frame_get_filename(ref_frame), i+1,
                CRIRES_HEADER_EXT_FORWARD, 0) ;
        cpl_propertylist_copy_property_regexp(qclists[i], inputlist,
                CRIRES_HEADER_EXT_FORWARD, 0) ;
        cpl_propertylist_delete(inputlist) ;
    }

    /* Write the image */
    crires_image_save(set,
            parlist,
            set,
            ilist,
            recipe_name,
            CRIRES_WL_MAP_IMA,
            CRIRES_PROTYPE_WL_MAP,
            crires_spec_wavecal_config.period,
            NULL,
            (const cpl_propertylist **)qclists,
            PACKAGE "/" PACKAGE_VERSION,
            "crires_spec_wavecal_ima.fits") ;

    if (ilist_model != NULL) {
        /* Write the image */
        crires_image_save(set,
                parlist,
                set,
                ilist_model,
                recipe_name,
                CRIRES_WL_MAP_MODEL_IMA,
                CRIRES_PROTYPE_WL_MAP,
                crires_spec_wavecal_config.period,
                NULL,
                (const cpl_propertylist **)qclists,
                PACKAGE "/" PACKAGE_VERSION,
                "crires_spec_wavecal_ima_model.fits") ;
    }

    /* Write the table */
    crires_table_save(set,
            parlist,
            set, 
            wl_tab, 
            recipe_name,
            CRIRES_CALPRO_WAVE, 
            CRIRES_PROTYPE_WL_POLY,
            NULL,
            (const cpl_propertylist **)qclists, 
            PACKAGE "/" PACKAGE_VERSION,
            "crires_spec_wavecal_tab.fits") ;

    /* Free and return */
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        cpl_propertylist_delete(qclists[i]) ;
    }
    cpl_free(qclists) ;
    return  0;
}
