/* $Id: crires_util_combine.c,v 1.41 2012-09-19 14:10:27 yjung Exp $
 *
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2012-09-19 14:10:27 $
 * $Revision: 1.41 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include "crires_recipe.h"

#include "crires_combine.h"
#include "crires_wlcalib.h"
#include "crires_extract.h"
#include "crires_photom.h"

/*-----------------------------------------------------------------------------
                                Define
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "crires_util_combine"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int crires_util_combine_save(const cpl_imagelist **, 
        const cpl_parameterlist *, cpl_frameset *) ;

static char crires_util_combine_description[] =
"crires_util_combine -- Images Combination tool\n"
"The files listed in the Set Of Frames (sof-file) must be tagged:\n"
"raw-file.fits "CRIRES_SPEC_JITTER_RAW" or\n"
"raw-file.fits "CRIRES_SPEC_JITTER_J_RAW" or\n"
"raw-file.fits "CRIRES_SPEC_JITTER_STD_RAW" or\n"
"raw-file.fits "CRIRES_SPEC_JITTER_J_STD_RAW" or\n"
"raw-file.fits "CRIRES_SPEC_NODDING_OBJECT_RAW" or\n"
"raw-file.fits "CRIRES_SPEC_NODDING_SKY_RAW" or\n"
"raw-file.fits "CRIRES_SPEC_GENERIC_OBJECT_RAW" or\n"
"raw-file.fits "CRIRES_SPEC_GENERIC_SKY_RAW" or\n"
"raw-file.fits "CRIRES_SPEC_NODDING_RAW" or\n"
"raw-file.fits "CRIRES_SPEC_NODDING_J_RAW" or\n"
"raw-file.fits "CRIRES_SPEC_NODDING_STD_RAW" or\n"
"raw-file.fits "CRIRES_SPEC_NODDING_J_STD_RAW" or\n"
"raw-file.fits "CRIRES_WIN_NODDING_OBJECT_RAW" or\n"
"raw-file.fits "CRIRES_WIN_NODDING_SKY_RAW" or\n"
"raw-file.fits "CRIRES_WIN_NODDING_RAW" or\n"
"raw-file.fits "CRIRES_WIN_NODDING_J_RAW" or\n"
"raw-file.fits "CRIRES_WIN_NODDING_STD_RAW" or\n"
"raw-file.fits "CRIRES_WIN_NODDING_J_STD_RAW" or\n"
"flat-file.fits "CRIRES_CALPRO_FLAT" or\n" 
"flat-file.fits "CRIRES_CALPRO_FLAT_WIN" or\n" 
"bpm-file.fits "CRIRES_CALPRO_BPM" or\n" 
"bpm-file.fits "CRIRES_CALPRO_BPM_WIN" or\n" 
"dark-file.fits "CRIRES_CALPRO_DARK" or\n" 
"dark-file.fits "CRIRES_CALPRO_DARK_WIN" or\n" 
"detlin-file.fits "CRIRES_CALPRO_COEFFS_CUBE" or\n"   
"\n"
"In the case of a nodding observation, in order not to degrade the \n"
"   instrument high resolution, the combined images using only NODA\n"
"   or NODB nodding positions can be produced on request. (see --onlyA/B)\n"
"\n"
"This recipe produces 2 files:\n"
"First product:     the combined image\n"
"                   (PRO TYPE = "CRIRES_PROTYPE_COMBINED")\n"
"Second product:    the contribution map\n"
"                   (PRO TYPE = "CRIRES_PROTYPE_CONTRIB")\n" ;

CRIRES_RECIPE_DEFINE(crires_util_combine,
        CRIRES_PARAM_REFINE         |
        CRIRES_PARAM_ONLYA          |
        CRIRES_PARAM_ONLYB          |
        CRIRES_PARAM_BLIND,
        "Images Combination tool",
        crires_util_combine_description) ;

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    int                 blind ;
    int                 refine ;
    int                 onlyA ;
    int                 onlyB ;
    int                 nodding ;
    /* Outputs */
    int                 win_mode ;
    crires_illum_period period ;
} crires_util_combine_config ;

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    parlist     the parameters list
  @param    frameset   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int crires_util_combine(
        cpl_frameset            *   frameset,
        const cpl_parameterlist *   parlist)
{
    cpl_frameset    *   rawframes ;
    cpl_frameset    *   skyframes ;
    const char      *   fname ;
    const char      *   flat ;
    const char      *   dark ;
    const char      *   bpm ;
    const char      *   flat_win ;
    const char      *   dark_win ;
    const char      *   bpm_win ;
    const char      *   detlin ;
    cpl_imagelist   **  comblist ;
    cpl_propertylist    *   plist ;
    const char          *   sval ;

    /* Initialise */
    rawframes = NULL ;
    skyframes = NULL ;
    
    /* Retrieve input parameters */
    crires_util_combine_config.refine = crires_parameterlist_get_bool(
            parlist, RECIPE_STRING, CRIRES_PARAM_REFINE) ;
    crires_util_combine_config.onlyA = crires_parameterlist_get_bool(
            parlist, RECIPE_STRING, CRIRES_PARAM_ONLYA) ;
    crires_util_combine_config.onlyB = crires_parameterlist_get_bool(
            parlist, RECIPE_STRING, CRIRES_PARAM_ONLYB) ;
    crires_util_combine_config.blind = crires_parameterlist_get_bool(
            parlist, RECIPE_STRING, CRIRES_PARAM_BLIND) ;
 
    /* Identify the RAW and CALIB frames in the input frameset */
    if (crires_dfs_set_groups(frameset, NULL)) {
        cpl_msg_error(__func__, "Cannot identify RAW and CALIB frames") ;
        return -1 ;
    }

    /* Retrieve calibration data */
    flat        = crires_extract_filename(frameset, CRIRES_CALPRO_FLAT) ;
    dark        = crires_extract_filename(frameset, CRIRES_CALPRO_DARK) ;
    bpm         = crires_extract_filename(frameset, CRIRES_CALPRO_BPM) ;
    flat_win    = crires_extract_filename(frameset, CRIRES_CALPRO_FLAT_WIN) ;
    dark_win    = crires_extract_filename(frameset, CRIRES_CALPRO_DARK_WIN) ;
    bpm_win     = crires_extract_filename(frameset, CRIRES_CALPRO_BPM_WIN) ;
    detlin      = crires_extract_filename(frameset, CRIRES_CALPRO_COEFFS_CUBE) ;
 
    /* Retrieve raw frames */
    if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_SPEC_JITTER_RAW)) != NULL) {
        crires_util_combine_config.nodding = 0 ;
    } else if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_SPEC_JITTER_J_RAW)) != NULL) {
        crires_util_combine_config.nodding = 0 ;
    } else if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_SPEC_NODDING_RAW)) != NULL) {
        crires_util_combine_config.nodding = 1 ;
    } else if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_SPEC_NODDING_OBJECT_RAW)) != NULL) {
        crires_util_combine_config.nodding = 1 ;
    } else if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_SPEC_NODDING_J_RAW)) != NULL) {
        crires_util_combine_config.nodding = 1 ;
    } else if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_SPEC_GENERIC_OBJECT_RAW)) != NULL) {
        crires_util_combine_config.nodding = 0 ;
    } else if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_SPEC_JITTER_STD_RAW)) != NULL) {
        crires_util_combine_config.nodding = 0 ;
    } else if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_SPEC_JITTER_J_STD_RAW)) != NULL) {
        crires_util_combine_config.nodding = 0 ;
    } else if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_SPEC_NODDING_STD_RAW)) != NULL) {
        crires_util_combine_config.nodding = 1 ;
    } else if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_SPEC_NODDING_J_STD_RAW)) != NULL) {
        crires_util_combine_config.nodding = 1 ;
    } else if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_WIN_NODDING_OBJECT_RAW)) != NULL) {
        crires_util_combine_config.nodding = 1 ;
    } else if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_WIN_NODDING_RAW)) != NULL) {
        crires_util_combine_config.nodding = 1 ;
    } else if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_WIN_NODDING_J_RAW)) != NULL) {
        crires_util_combine_config.nodding = 1 ;
    } else if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_WIN_NODDING_STD_RAW)) != NULL) {
        crires_util_combine_config.nodding = 1 ;
    } else if ((rawframes = crires_extract_frameset(frameset,
                    CRIRES_WIN_NODDING_J_STD_RAW)) != NULL) {
        crires_util_combine_config.nodding = 1 ;
    } else {
        cpl_msg_error(__func__, "No raw frame in input") ;
        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
        return -1 ;
    }

    /* OnlyA and OnlyB are only possible in nodding mode */
    if (crires_util_combine_config.nodding == 0) {
        if (crires_util_combine_config.onlyA) {
            cpl_msg_warning(__func__, "onlyA only possible in nodding mode") ;
            crires_util_combine_config.onlyA = 0 ;
        }
        if (crires_util_combine_config.onlyB) {
            cpl_msg_warning(__func__, "onlyB only possible in nodding mode") ;
            crires_util_combine_config.onlyB = 0 ;
        }
    }

    /* Retrieve sky frames if any */
    skyframes = crires_extract_frameset(frameset, CRIRES_SPEC_NODDING_SKY_RAW) ;
    if (skyframes == NULL) {
        skyframes = crires_extract_frameset(frameset,
                CRIRES_WIN_NODDING_SKY_RAW) ;
    }
    if (skyframes == NULL) {
        skyframes = crires_extract_frameset(frameset,
                CRIRES_SPEC_GENERIC_SKY_RAW) ;
    }

    /* Get the detector illumination period */
    crires_util_combine_config.period =
        crires_get_detector_illum_period(cpl_frame_get_filename(
                    cpl_frameset_get_position(rawframes, 0))) ;
    if (crires_util_combine_config.period == CRIRES_ILLUM_UNKNOWN) {
        cpl_msg_error(__func__,
                "Cannot determine the detector illumination period") ;
        cpl_frameset_delete(rawframes) ;
        if (skyframes != NULL) cpl_frameset_delete(skyframes) ;
        return -1 ;
    }

    /* Windowing mode ? */
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(
                        cpl_frameset_get_position(rawframes, 0)), 0)) == NULL)
        return -1 ;
    sval = crires_pfits_get_ncorrs(plist) ;
    if (!strcmp(sval, "FowlerNsampGRstWin")) {
        crires_util_combine_config.period = CRIRES_ILLUM_FULL_DETECTOR ;
        crires_util_combine_config.win_mode = 1 ;
    } else {
        crires_util_combine_config.win_mode = 0 ;
    }
    cpl_propertylist_delete(plist) ;
    
    /* Display the Detector illumination */
    crires_display_detector_illum(crires_util_combine_config.period) ;

    /* Verify the STRIPE keys conformity */
    if (crires_util_combine_config.win_mode == 1) {
        fname = cpl_frame_get_filename(cpl_frameset_get_position(rawframes,0)) ;
        if (flat_win != NULL) {
            if (crire_stripe_keys_mismatch(fname, flat_win)) {
                cpl_msg_error(__func__,
                        "Mismatch of STRIPE keys with the flat frame") ;
                cpl_frameset_delete(rawframes) ;
                if (skyframes != NULL) cpl_frameset_delete(skyframes) ;
                cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
                return -1 ;
            }
        }
        if (bpm_win != NULL) {
            if (crire_stripe_keys_mismatch(fname, bpm_win)) {
                cpl_msg_error(__func__,
                        "Mismatch of STRIPE keys with the bpm frame") ;
                cpl_frameset_delete(rawframes) ;
                if (skyframes != NULL) cpl_frameset_delete(skyframes) ;
                cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
                return -1 ;
            }
        }
        if (dark_win != NULL) {
            if (crire_stripe_keys_mismatch(fname, dark_win)) {
                cpl_msg_error(__func__,
                        "Mismatch of STRIPE keys with the dark frame") ;
                cpl_frameset_delete(rawframes) ;
                if (skyframes != NULL) cpl_frameset_delete(skyframes) ;
                cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
                return -1 ;
            }
        }
    }

    /* Images recombination */
    cpl_msg_info(__func__, "Images combination") ;
    cpl_msg_indent_more() ;
    if (crires_util_combine_config.win_mode == 0) {
        comblist = crires_combine_imagelist(rawframes, skyframes,
                    crires_util_combine_config.period,
                    flat, dark, bpm, detlin,
                    crires_util_combine_config.nodding,
                    crires_util_combine_config.blind,
                    crires_util_combine_config.refine,
                    crires_util_combine_config.onlyA,
                    crires_util_combine_config.onlyB) ;
    } else {
        comblist = crires_combine_imagelist_win(rawframes,
                flat_win, dark_win, bpm_win, detlin,
                crires_util_combine_config.refine,
                crires_util_combine_config.onlyA,
                crires_util_combine_config.onlyB) ;
    }
    if (comblist == NULL) {
        cpl_msg_error(__func__, "Cannot combine the images") ;
        cpl_frameset_delete(rawframes) ;
        if (skyframes != NULL) cpl_frameset_delete(skyframes) ;
        cpl_msg_indent_less() ;
        return -1 ;
    }
    cpl_frameset_delete(rawframes) ;
    if (skyframes != NULL) cpl_frameset_delete(skyframes) ;
    cpl_msg_indent_less() ;
    
    /* Save the product */
    cpl_msg_info(__func__, "Save the product") ;
    cpl_msg_indent_more() ;
    if (crires_util_combine_save((const cpl_imagelist **)comblist, parlist, 
                frameset)) {
        cpl_msg_error(__func__, "Cannot save the product") ;
        cpl_imagelist_delete(comblist[0]) ;
        cpl_imagelist_delete(comblist[1]) ;
        if (crires_util_combine_config.onlyA) {
            cpl_imagelist_delete(comblist[2]) ;
            cpl_imagelist_delete(comblist[3]) ;
        }
        if (crires_util_combine_config.onlyB) {
            cpl_imagelist_delete(comblist[4]) ;
            cpl_imagelist_delete(comblist[5]) ;
        }
        cpl_free(comblist) ;
        cpl_msg_indent_less() ;
        return -1 ;
    }
    cpl_imagelist_delete(comblist[0]) ;
    cpl_imagelist_delete(comblist[1]) ;
    if (crires_util_combine_config.onlyA) {
        cpl_imagelist_delete(comblist[2]) ;
        cpl_imagelist_delete(comblist[3]) ;
    }
    if (crires_util_combine_config.onlyB) {
        cpl_imagelist_delete(comblist[4]) ;
        cpl_imagelist_delete(comblist[5]) ;
    }
    cpl_free(comblist) ;
    cpl_msg_indent_less() ;

    /* Return */
    if (cpl_error_get_code()) return -1 ;
    else return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the products
  @param    images      6 image lists : combined, contrib, combined
            nodding A, contrib nodding A, combined nodding B, contrib nodding B
  @param    parlist     the parameters list
  @param    set         the frames set
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int crires_util_combine_save(
        const cpl_imagelist     **  images,
        const cpl_parameterlist *   parlist,
        cpl_frameset            *   set)
{
    const char          *   recipe_name = "crires_util_combine" ;

    /* Write the combined image */
    crires_image_save(set,
            parlist,
            set, 
            images[0], 
            recipe_name,
            CRIRES_OBS_COMBINED_IMA, 
            CRIRES_PROTYPE_COMBINED,
            crires_util_combine_config.period,
            NULL,
            NULL,
            PACKAGE "/" PACKAGE_VERSION,
            "crires_util_combine_comb.fits") ;

    /* Write the contribution map */
    crires_image_save(set,
            parlist,
            set, 
            images[1], 
            recipe_name,
            CRIRES_OBS_CONTRIBUTION_IMA, 
            CRIRES_PROTYPE_CONTRIB,
            crires_util_combine_config.period,
            NULL,
            NULL,
            PACKAGE "/" PACKAGE_VERSION,
            "crires_util_combine_contrib.fits") ;

    /* Nodded A support */
    if (crires_util_combine_config.onlyA) {
        /* Write the combined Nodded A image */
        crires_image_save(set,
                parlist,
                set,
                images[2],
                recipe_name,
                CRIRES_OBS_COMBINED_NA_IMA,
                CRIRES_PROTYPE_COMBINED,
                crires_util_combine_config.period,
                NULL,
                NULL,
                PACKAGE "/" PACKAGE_VERSION,
                "crires_util_combine_comb_noddedA.fits") ;

        /* Write the contribution Nodded A image */
        crires_image_save(set,
                parlist,
                set,
                images[3],
                recipe_name,
                CRIRES_OBS_CONTRIBUTION_NA_IMA,
                CRIRES_PROTYPE_CONTRIB,
                crires_util_combine_config.period,
                NULL,
                NULL,
                PACKAGE "/" PACKAGE_VERSION,
                "crires_util_combine_contrib_noddedA.fits") ;
    }

    /* Nodded B support */
    if (crires_util_combine_config.onlyB) {
        /* Write the combined Nodded B image */
        crires_image_save(set,
                parlist,
                set,
                images[4],
                recipe_name,
                CRIRES_OBS_COMBINED_NB_IMA,
                CRIRES_PROTYPE_COMBINED,
                crires_util_combine_config.period,
                NULL,
                NULL,
                PACKAGE "/" PACKAGE_VERSION,
                "crires_util_combine_comb_noddedB.fits") ;

        /* Write the contribution Nodded B image */
        crires_image_save(set,
                parlist,
                set,
                images[5],
                recipe_name,
                CRIRES_OBS_CONTRIBUTION_NB_IMA,
                CRIRES_PROTYPE_CONTRIB,
                crires_util_combine_config.period,
                NULL,
                NULL,
                PACKAGE "/" PACKAGE_VERSION,
                "crires_util_combine_contrib_noddedB.fits") ;
    }
    return  0;
}

