/* $Id: crires_util_genypos.c,v 1.2 2012-10-09 08:18:01 yjung Exp $
 *
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2012-10-09 08:18:01 $
 * $Revision: 1.2 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include "crires_recipe.h"

/*-----------------------------------------------------------------------------
                                Define
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "crires_util_genypos"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int crires_util_genypos_save(cpl_table *, const cpl_parameterlist *, 
        cpl_frameset *);

static char crires_util_genypos_description[] =
"This recipe is used to generate the FITS table with the Y positions.\n"
"The sof file contains the names of the input ASCII file\n"
"tagged with "CRIRES_UTIL_GENYPOS_RAW".\n"
"The ASCII file must contain two columns:\n"
"1st: Chip Number\n"
"2nd: Y Position\n"
"The ASCII file is in the catalogs/ directory of the CRIRES distribution.\n"
"This recipe produces 1 file:\n"
"First product:     the table with the positions.\n"
"                   (PRO TYPE = "CRIRES_PROTYPE_YPOS")\n" ;

CRIRES_RECIPE_DEFINE(crires_util_genypos, 0,
        "Generate Y Positions FITS table",
        crires_util_genypos_description) ;

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    The FITS file creation occurs here 
  @param    framelist   the frames list
  @return   0 iff everything is ok

  The recipe expects a text file with two columns, and will create a FITS file
  out of it. 
 */
/*----------------------------------------------------------------------------*/
static int crires_util_genypos(
        cpl_frameset            *   framelist,
        const cpl_parameterlist *   parlist)
{
    cpl_bivector    *   bivec ;
    double          *   pbivec_x ;
    double          *   pbivec_y ;
    cpl_bivector    *   bivec_fill ;
    double          *   pbivec_fill_x ;
    double          *   pbivec_fill_y ;
    cpl_frame       *   cur_frame ;
    int                 nvals, nb_new_vals ;
    double              wavel ;
    cpl_table       *   tab ;
    int                 i ;

    /* Identify the RAW and CALIB frames in the input frameset */
    if (crires_dfs_set_groups(framelist, NULL)) {
        cpl_msg_error(__func__, "Cannot identify RAW and CALIB frames") ;
        return -1 ;
    }
   
    /* Load the file */
    cur_frame = cpl_frameset_get_position(framelist, 0) ;
    if ((bivec=cpl_bivector_read(cpl_frame_get_filename(cur_frame)))==NULL) {
        cpl_msg_error(__func__, "Cannot load the file in the bivector") ;
        return -1 ;
    }
    nvals = cpl_bivector_get_size(bivec) ;
    
    /* Allocate the data container */
    tab = cpl_table_new(nvals) ;
    cpl_table_wrap_double(tab, cpl_bivector_get_x_data(bivec),
            CRIRES_CHIP_NB) ;
    cpl_table_wrap_double(tab, cpl_bivector_get_y_data(bivec), 
            CRIRES_COL_DETPOSY) ;
   
    /* Save the table */
    cpl_msg_info(__func__, "Saving the table with %d rows", nvals) ;
    if (crires_util_genypos_save(tab, parlist, framelist) == -1) {
        cpl_msg_error(__func__, "Cannot write the table") ;
        cpl_bivector_delete(bivec) ;
        cpl_table_unwrap(tab, CRIRES_CHIP_NB) ;
        cpl_table_unwrap(tab, CRIRES_COL_DETPOSY) ;
        cpl_table_delete(tab) ;
        return -1 ;
    }
    cpl_bivector_delete(bivec) ;
    cpl_table_unwrap(tab, CRIRES_CHIP_NB) ;
    cpl_table_unwrap(tab, CRIRES_COL_DETPOSY) ;
    cpl_table_delete(tab) ;
    return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the product of the recipe
  @param    out_table   the table 
  @param    parlist     the input list of parameters
  @param    set         the input frame set
  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
static int crires_util_genypos_save(
        cpl_table               *   out_table,
        const cpl_parameterlist *   parlist,
        cpl_frameset            *   set)
{
    cpl_propertylist    *   plist ;

    plist = cpl_propertylist_new();
    cpl_propertylist_append_string(plist, "INSTRUME", "CRIRES") ;
    cpl_propertylist_append_string(plist, CPL_DFS_PRO_CATG, 
            CRIRES_CALPRO_THAR_POS) ;
    cpl_propertylist_append_string(plist, CPL_DFS_PRO_TYPE, 
            CRIRES_PROTYPE_YPOS) ;

    if (cpl_dfs_save_table(set, NULL, parlist, set, NULL, out_table,
                NULL, "crires_util_genypos", plist, NULL,
                PACKAGE "/" PACKAGE_VERSION,
                "crires_util_genypos.fits") != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot save the table") ;
        return -1 ;
    }
    cpl_propertylist_delete(plist) ;

    /* Return */
    return 0 ;
}
