/* $Id: crires_util_sensitivity.c,v 1.32 2012-09-20 12:15:03 yjung Exp $
 *
 * This file is part of the crires Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2012-09-20 12:15:03 $
 * $Revision: 1.32 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include "crires_recipe.h"

#include "crires_photom.h"
#include "irplib_utils.h"

/*-----------------------------------------------------------------------------
                                Define
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "crires_util_sensitivity"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int crires_util_sensitivity_save(const cpl_table **, 
        const cpl_parameterlist *, cpl_frameset *) ;

static char crires_util_sensitivity_description[] = 
"This recipe accepts 1 parameter:\n"
"First parameter:   the table with the conversion factor.\n"
"                   (PRO TYPE = "CRIRES_PROTYPE_CONVERS")\n"
"\n"
"This recipe produces 1 file:\n"
"First product:     the table with the sensitivity signal.\n"
"                   (PRO TYPE = "CRIRES_PROTYPE_SENSIT")\n" ;

CRIRES_RECIPE_DEFINE(crires_util_sensitivity,
        CRIRES_PARAM_DISPLAY        |
        CRIRES_PARAM_EXPTIME,
        "Compute the sensitivity",
        crires_util_sensitivity_description) ;

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    int             display ;
    double          exptime ;
    /* Outputs */
    int             win_mode ;
    double          sens_med[CRIRES_NB_DETECTORS] ;
} crires_util_sensitivity_config ;

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    frameset   the frames list
  @param    parlist     the parameters list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int crires_util_sensitivity(
        cpl_frameset            *   frameset,
        const cpl_parameterlist *   parlist)
{
    const char          *   sval ;
    cpl_propertylist    *   plist ;
    cpl_frame           *   fr ;
    cpl_table           *   conv_tab[CRIRES_NB_DETECTORS] ;
    double                  exptime_comp ;
    int                     i ;

    /* Initialise */
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        crires_util_sensitivity_config.sens_med[i] = -1.0 ;
        conv_tab[i] = NULL ;
    } 

    /* Retrieve input parameters */
    crires_util_sensitivity_config.display = crires_parameterlist_get_int(
            parlist, RECIPE_STRING, CRIRES_PARAM_DISPLAY) ;
    crires_util_sensitivity_config.exptime = crires_parameterlist_get_double(
            parlist, RECIPE_STRING, CRIRES_PARAM_EXPTIME) ;

    /* Identify the RAW and CALIB frames in the input frameset */
    if (crires_dfs_set_groups(frameset, NULL)) {
        cpl_msg_error(__func__, "Cannot identify RAW and CALIB frames") ;
        return -1 ;
    }

    /* Windowing mode ? */
    fr = cpl_frameset_get_position(frameset, 0);
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(fr), 0)) == NULL)
        return -1 ;
    sval = crires_pfits_get_ncorrs(plist) ;
    if (!strcmp(sval, "FowlerNsampGRstWin")) {
        crires_util_sensitivity_config.win_mode = 1 ;
    } else {
        crires_util_sensitivity_config.win_mode = 0 ;
    }
    cpl_propertylist_delete(plist) ;

    /* Loop on the CRIRES_NB_DETECTORS detectors */
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        /* Skip some detectors in windowing mode */
        if ((i==0 || i==CRIRES_NB_DETECTORS-1) &&
                (crires_util_sensitivity_config.win_mode == 1)) {
            continue ;
        }

        cpl_msg_info(__func__, "Sensitivity computation for chip %d", i+1) ;

        /* The first frame must be the conversion table */
        cpl_msg_info(__func__, "Load the extracted table") ;
        cpl_msg_indent_more() ;
        fr = cpl_frameset_get_position(frameset, 0);
        if ((conv_tab[i] = crires_load_table_check(cpl_frame_get_filename(fr),
                        i+1, CRIRES_PROTYPE_CONVERS, -1, -1, 0)) == NULL) {
            cpl_msg_indent_less() ;
            continue ;
        }
        cpl_msg_indent_less() ;

        /* Compute */
        cpl_msg_info(__func__, "Sensitivity computation") ;
        cpl_msg_indent_more() ;
        if ((exptime_comp = crires_photom_sens_engine(conv_tab[i], 
                cpl_frame_get_filename(fr),
                crires_util_sensitivity_config.exptime,
                crires_util_sensitivity_config.display==i+1)) < 0.0 ) {
            cpl_msg_error(__func__, "Cannot compute sensitivity") ;
            cpl_msg_indent_less() ;
            cpl_table_delete(conv_tab[i]) ;
            conv_tab[i] = NULL ;
            continue ;
        }
        cpl_msg_indent_less() ;
        crires_util_sensitivity_config.exptime = exptime_comp ;

        /* Compute the Median */
        crires_util_sensitivity_config.sens_med[i] = 
            cpl_table_get_column_median(conv_tab[i], CRIRES_COL_SENSITIVITY) ;
    }
   
    /* Reconstruct chips in windowing mode using detector 2 */
    if (crires_util_sensitivity_config.win_mode == 1) {
        if (conv_tab[1] != NULL) {
            conv_tab[0] = cpl_table_duplicate(conv_tab[1]) ;
            cpl_table_set_size(conv_tab[0], 0) ;
            conv_tab[CRIRES_NB_DETECTORS-1] = cpl_table_duplicate(conv_tab[0]) ;
        }
    }

    /* Save the product */
    cpl_msg_info(__func__, "Save the product") ;
    cpl_msg_indent_more() ;
    if (crires_util_sensitivity_save((const cpl_table **)conv_tab, parlist, 
                frameset)) {
        cpl_msg_error(__func__, "Cannot save the product") ;
        for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
            if (conv_tab[i] != NULL) cpl_table_delete(conv_tab[i]) ;
        }
        cpl_msg_indent_less() ;
        return -1 ;
    }
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        if (conv_tab[i] != NULL) cpl_table_delete(conv_tab[i]) ;
    }
    cpl_msg_indent_less() ;

    /* Return */
    if (cpl_error_get_code()) return -1 ;
    else return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the products
  @param    out         the tables to save
  @param    parlist     the parameters list
  @param    set         the frames set
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int crires_util_sensitivity_save(
        const cpl_table         **  out,
        const cpl_parameterlist *   parlist,
        cpl_frameset            *   set)
{
    cpl_propertylist    **  qclists ;
    const cpl_frame     *   ref_frame ;
    cpl_propertylist    *   inputlist ;
    const char          *   recipe_name = "crires_util_sensitivity" ;
    int                     i ;

    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW) ;

    /* Create the QC lists */
    qclists = cpl_malloc(CRIRES_NB_DETECTORS * sizeof(cpl_propertylist*)) ;
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        qclists[i] = cpl_propertylist_new() ;
        cpl_propertylist_append_double(qclists[i], "ESO QC SENSMED",
                crires_util_sensitivity_config.sens_med[i]) ;
        cpl_propertylist_append_double(qclists[i], "ESO QC EXPTIME",
                crires_util_sensitivity_config.exptime) ;

        /* Propagate some keywords from input raw frame extensions */
        inputlist = cpl_propertylist_load_regexp(
                cpl_frame_get_filename(ref_frame), i+1,
                CRIRES_HEADER_EXT_FORWARD, 0) ;
        cpl_propertylist_copy_property_regexp(qclists[i], inputlist,
                CRIRES_HEADER_EXT_FORWARD, 0) ;
        cpl_propertylist_delete(inputlist) ;
    }

    /* Write the table */
    crires_table_save(set,
            parlist,
            set,
            out,
            recipe_name,
            CRIRES_EXTRACT_SENS_TAB,
            CRIRES_PROTYPE_SENSIT,
            NULL, 
            (const cpl_propertylist **)qclists,
            PACKAGE "/" PACKAGE_VERSION,
            "crires_util_sensitivity.fits") ;

    /* Free and return */
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        cpl_propertylist_delete(qclists[i]) ;
    }
    cpl_free(qclists) ;
    return  0;
}

